%{
/*
** Copyright (C) 2005-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  Tokenizer for probe configuration file.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: probeconfscan.l 13074 2008-11-20 20:28:45Z mthomas $");

#include "probeconfscan.h"
#include "probeconfparse.h"


/* TYPEDEFS AND DEFINES */

/* To avoid warnings with flex */
#define YY_NO_UNPUT 1

typedef struct pcscan_file_st {
    YY_BUFFER_STATE state;
    char           *name;
    FILE           *fp;
    int             line;
} pcscan_file_t;


/* EXPORTED VARIABLES */

/* Global error count for return status of skpcParse */
int pcscan_errors;

/* keyword we read; for printing messages in the parser. */
char pcscan_clause[128];

int (*extra_sensor_verify_fn)(skpc_sensor_t *sensor) = NULL;


/* LOCAL FUNCTION PROTOTYPES */

static int stash(int tok);
static void err_bad_line(int is_id);
static void err_include_arg(void);
static void err_old_style(void);
static char *get_name(const char *s, int c);

/* support for "include" */
static pcscan_file_t pcscan_stack[PCSCAN_MAX_INCLUDE_DEPTH];

/* next stack position */
static int pcscan_stack_depth = 0;

/* current pointer into pcscan_stack[] */
static pcscan_file_t *pcscan_file = NULL;

/* local buffer for reading quoted strings */
#define BUF_SIZE 2048
static char pcscan_buf[BUF_SIZE];
static char *pcscan_buf_end = pcscan_buf + BUF_SIZE - 1;
static char *pcscan_buf_ptr;

/* Check that we're not overflowing the buffer */
#define STRING_CHECK                            \
    if ( pcscan_buf_ptr >= pcscan_buf_end ) {   \
        skpcParseErr("String too long");        \
        BEGIN(ST_ERR);                          \
        return ERR_STR_TOO_LONG;                \
    }

%}

%option never-interactive nounput
%option prefix="probeconfscan_" outfile="lex.yy.c"


    /* create a state that says we have read a keyword and are parsing
     * the arguments */

%s ST_ARGS

    /* create a state for an unrecognized keyword at start of line;
     * reads to EOL */

%s ST_ERR

    /* create a state for an unrecognized keyword within a statement;
     * reads to EOL */

%s ST_ARG_ERR

    /* create a state for processing quoted strings */

%s ST_STRING

    /* ignore the sensor statements */

%s ST_IGNORE_SENSOR


    /* basic character classes */

idchar                                  [-._/A-Za-z0-9:]

notidchar                               [^-._/A-Za-z0-9:\n]

    /* an ID must have at least one non-number character; note that this
     * must also match hostnames and IP addresses */

id                                      [-._/A-Za-z0-9:]+

    /* comment; ignore, and increment line count */

comment                                 [ \t\r]*#.*\n

    /* whitespace; we ignore it */

wschar                                  [ \t\r]

ws                                      [ \t\r]+

    /* end of line; increment line count */

nl                                      [ \t\r]*\n

%%

{ws}                                    ;

<INITIAL,ST_ERR>{comment}               { ++pcscan_file->line;BEGIN(INITIAL); }

<ST_ARGS>{comment}                      { ++pcscan_file->line;
                                          BEGIN(INITIAL);
                                          return EOL; }

<INITIAL,ST_ERR>{nl}                    { ++pcscan_file->line;BEGIN(INITIAL); }

<ST_ARGS,ST_ARG_ERR>{nl}                { ++pcscan_file->line;
                                          BEGIN(INITIAL);
                                          return EOL; }

<INITIAL>{id}-interfaces?               { yylval.string = get_name(yytext,'-');
                                          return stash(INTERFACE); }

<INITIAL>{id}-ipblocks?                 { yylval.string = get_name(yytext,'-');
                                          return stash(IPBLOCK); }

<INITIAL>{id}-probes                    { yylval.string = get_name(yytext,'-');
                                          return stash(PROBES); }

<INITIAL>accept-from-host               { return stash(ACCEPT_FROM_HOST_T); }
<INITIAL>destination-network            { yylval.net_dir = SKPC_DIR_DST;
                                          return stash(NET_DIRECTION);}
<INITIAL>end{ws}probe                   { return stash(END_PROBE_T); }
<INITIAL>end{ws}sensor                  { return stash(END_SENSOR_T); }
<INITIAL>include                        { return stash(INCLUDE_T); }
<INITIAL>include{ws}[^\"].*{nl}         { err_include_arg();
                                          ++pcscan_file->line; }
<INITIAL>isp-ips?                       { return stash(ISP_IP_T); }
<INITIAL>listen-as-host                 { return stash(LISTEN_AS_HOST_T); }
<INITIAL>listen-on-port                 { return stash(LISTEN_ON_PORT_T); }
<INITIAL>listen-on-unix-domain-socket   { return stash(LISTEN_ON_USOCKET_T);}
<INITIAL>listen-on-unix-socket          { return stash(LISTEN_ON_USOCKET_T);}
<INITIAL>log-flags                      { return stash(LOG_FLAGS_T); }
<INITIAL>poll-directory                 { return stash(POLL_DIRECTORY_T); }
<INITIAL>priority                       { return stash(PRIORITY_T); }
<INITIAL>probe                          { return stash(PROBE_T); }
<INITIAL>protocol                       { return stash(PROTOCOL_T); }
<INITIAL>read-from-file                 { return stash(READ_FROM_FILE_T); }
<INITIAL>source-network                 { yylval.net_dir = SKPC_DIR_SRC;
                                          return stash(NET_DIRECTION); }
<INITIAL>sensor                         { if (extra_sensor_verify_fn) {
                                              return stash(SENSOR_T);
                                          } else {
                                              BEGIN(ST_IGNORE_SENSOR);
                                        } }
<INITIAL>sensor-probe                   { /* error when old sensor-probe */
                                          err_old_style(); }

<INITIAL>.                              { BEGIN(ST_ERR); yyless(0); }

<ST_ARGS>remainder                      { return REMAINDER_T; }

<ST_ARGS>{id}                           { yylval.string = strdup(yytext);
                                          return ID; }

<ST_ARGS>\"                             { BEGIN(ST_STRING);
                                          pcscan_buf_ptr = pcscan_buf; }

<ST_ARGS>","                            { return COMMA; }

<ST_ARGS>.                              { BEGIN(ST_ARG_ERR); yyless(0); }

<ST_ERR,ST_ARG_ERR>{idchar}.*           { err_bad_line(1); }

<ST_ERR,ST_ARG_ERR>{notidchar}.*        { err_bad_line(0); }



    /* ST_IGNORE_SENSOR state: ignore everything until we reach the
     * "end sensor" line. */

<ST_IGNORE_SENSOR>end{ws}sensor{comment} |
<ST_IGNORE_SENSOR>end{ws}sensor{nl}     { ++pcscan_file->line;
                                          BEGIN(INITIAL); }
<ST_IGNORE_SENSOR>.*{nl}                { ++pcscan_file->line; }


    /* ST_STRING state: Accepts the remainder of a quoted string
     * (after the initial quote) and returns the value.  This is
     * pretty much equivalent to quoted strings in C. */

<ST_STRING>\"           { BEGIN(ST_ARGS);
                          STRING_CHECK;
                          *pcscan_buf_ptr = '\0';
                          yylval.string = strdup(pcscan_buf);
                          return QUOTED_STRING; }

<ST_STRING>\n           { BEGIN(INITIAL);

                          ++pcscan_file->line;
                          return EOL; }

<ST_STRING>\\n          { STRING_CHECK; *pcscan_buf_ptr++ = '\n'; }
<ST_STRING>\\t          { STRING_CHECK; *pcscan_buf_ptr++ = '\t'; }
<ST_STRING>\\r          { STRING_CHECK; *pcscan_buf_ptr++ = '\r'; }
<ST_STRING>\\b          { STRING_CHECK; *pcscan_buf_ptr++ = '\b'; }
<ST_STRING>\\f          { STRING_CHECK; *pcscan_buf_ptr++ = '\f'; }
<ST_STRING>\\\n         { STRING_CHECK; *pcscan_buf_ptr++ = '\n'; }
<ST_STRING>\\.          { STRING_CHECK; *pcscan_buf_ptr++ = yytext[1]; }

<ST_STRING>[^\\\n\"]+   { if ((pcscan_buf_ptr + yyleng) > pcscan_buf_end ) {
                              BEGIN(ST_ERR);
                              return ERR_STR_TOO_LONG;
                          }
                          memcpy(pcscan_buf_ptr, yytext, yyleng);
                          pcscan_buf_ptr += yyleng; }

%%

int yywrap(void)
{
    return skpcParseIncludePop();
}


/*
 *  stash(int)
 *
 *      Stash the current keyword (using the yyleng and yytext
 *      globals), put us into a ST_ARGS start-state, and return the
 *      value we were passed.
 */
static int stash(int tok)
{
    /* make certain the keyword fits in our buffer */
    assert(yyleng < ((int)sizeof(pcscan_clause)));

    /* copy it in */
    strncpy(pcscan_clause, yytext, sizeof(pcscan_clause));

    /* we're processing the arguments */
    BEGIN(ST_ARGS);

    /* return whatever we were passed */
    return tok;
}


static char *get_name(const char *s, int c)
{
    const char *cp = strrchr(s, c);
    char *sp;

    if (cp == NULL) {
        return strdup(s);
    }
    sp = malloc(1 + (cp - s));
    if (sp) {
        strncpy(sp, s, (cp - s));
        sp[(cp - s)] = '\0';
    }
    return sp;
}


/*
 *  err_bad_line(is_id);
 *
 *    Called when the scanner reads a bad line of input.  This
 *    function prints an error and increases the global error count.
 *
 *    If 'is_id' is 0, we have a single uknown character; otherwise we
 *    have a bad keyword.  At this point yytext holds the entire bad
 *    line.
 */
static void err_bad_line(int is_id)
{
    size_t i;

    /* Make sure the parser lets the application know that things failed. */
    ++pcscan_errors;

    /* Handle a bad character; this is easy */
    if (is_id == 0) {
        skpcParseErr("Unexpected character '%c'", yytext[0]);
        return;
    }

    /* Handle a bad keyword.  Copy the first keyword into
     * pcscan_clause[] so we can print it. */
    for (i = 0; i < sizeof(pcscan_clause)-1; ++i) {
        /* This switch has the same character classes as our {idchar}. */
        switch (yytext[i]) {
          case '\0':
            goto END_WORD;
          case '-':
          case '.':
          case '_':
          case '/':
            pcscan_clause[i] = yytext[i];
            break;
          default:
            if (isalnum((int)yytext[i])) {
                pcscan_clause[i] = yytext[i];
            } else {
                goto END_WORD;
            }
            break;
        }
    }
  END_WORD:
    pcscan_clause[i] = '\0';

    skpcParseErr("Unrecognized keyword '%s'", pcscan_clause);
}


/*
 *  err_include_arg();
 *
 *    Print a message about the argument to include statement.
 */
static void err_include_arg(void)
{
    /* Make sure the parser lets the application know that things failed. */
    ++pcscan_errors;

    skpcParseErr("Argument to include must be in \"double quotes\"");
}

/*
 *  err_old_style();
 *
 *    Print a message that sensor-probe is no longer supported and
 *    exit the program.
 */
static void err_old_style(void)
{
    skpcParseErr("The sensor-probe keyword is no longer supported.\n"
                 "\tVerify your file follows the new sensor.conf syntax");
    exit(EXIT_FAILURE);
}


#ifndef TEST_PRINTF_FORMATS
int skpcParseErr(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    if (pcscan_file != NULL && pcscan_file->name != NULL) {
        fprintf(stderr, "%s: Error while parsing file %s at line %d:\n\t",
                skAppName(), pcscan_file->name, pcscan_file->line);
    } else {
        fprintf(stderr, "%s: Error during parsing\n\t",
                skAppName());
    }
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
    return 0;
}
#endif  /* TEST_PRINTF_FORMATS */


static int skpcParseOpenFile(char *filename)
{
    pcscan_file_t *new_file;
    const char *action_name =
        ((pcscan_stack_depth == 0) ? "read sensor-config" : "include");

    if (filename == NULL) {
        goto ERROR;
    }

    if (pcscan_stack_depth >= PCSCAN_MAX_INCLUDE_DEPTH) {
        skpcParseErr("Cannot include file '%s': includes nested too deeply",
                     filename);
        goto ERROR;
    }

    new_file = &pcscan_stack[pcscan_stack_depth];
    new_file->name = filename;
    new_file->line = 1;

    /* open the file */
    new_file->fp = fopen(new_file->name, "r");
    if (!new_file->fp) {
        int errnum = errno;
        if (!skFileExists(new_file->name)) {
            skpcParseErr(("Cannot %s '%s': "
                          "File does not exist or is not a regular file"),
                         action_name, filename);
        } else {
            skpcParseErr("Cannot %s '%s': %s",
                         action_name, filename, strerror(errnum));
        }
        goto ERROR;
    }

    /* set up scanner */
    new_file->state = yy_create_buffer(new_file->fp, YY_BUF_SIZE);
    yy_switch_to_buffer(new_file->state);

    /* success */
    pcscan_file = new_file;
    ++pcscan_stack_depth;
    BEGIN(INITIAL);

    return 0;

  ERROR:
    if (filename) {
        free(filename);
    }
    return -1;
}


int skpcParseIncludePop(void)
{
    if (pcscan_stack_depth == 0) {
        return 1;
    }

    /* free current state */
    yy_delete_buffer(YY_CURRENT_BUFFER);
    fclose(pcscan_file->fp);
    free(pcscan_file->name);

    /* pop to previous state; if no more states, we're done */
    --pcscan_stack_depth;
    if (pcscan_stack_depth == 0) {
        pcscan_file = NULL;
        return 1;
    }

    pcscan_file = &(pcscan_stack[pcscan_stack_depth-1]);
    yy_switch_to_buffer(pcscan_file->state);

    /* let processing continue */
    return 0;
}


int skpcParseIncludePush(char *filename)
{
    return skpcParseOpenFile(filename);
}


int skpcParse(
    const char         *filename,
    int               (*site_verify_sensor_fn)(skpc_sensor_t *sensor))
{
    /* store the site-specific sensor-verify function */
    extra_sensor_verify_fn = site_verify_sensor_fn;

    /* clear the error count */
    pcscan_errors = 0;

    /* initialize the include stack */
    pcscan_stack_depth = 0;

    /* send in a copy so we can free it */
    if (skpcParseOpenFile(strdup(filename))) {
        return -1;
    }

    /* parse */
    (void)yyparse();

    /* return non-zero if we had errors */
    if (pcscan_errors > 0) {
        return -1;
    }

    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
