/*
** Copyright (C) 2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  cutmatch.c
**
**  A plug-in to be loaded by rwcut to display the values that rwmatch
**  encodes in the NextHopIP field.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: cutmatch.c 11587 2008-05-20 17:11:40Z mthomas $");

#include "utils.h"
#include "dynlib.h"
#include "rwpack.h"

/* exported functions--called by dynlib */
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType);
void teardown(dynlibSymbolId appType);
int configure(dynlibInfoStruct_t *dlISP, void *data);
int cut(unsigned int type, char *out, size_t len_out, rwRec *rwrec);


/* Plugin's name */
static const char *pluginName = "cutmatch";

/* Fields this plug-in provides */
static sk_stringmap_entry_t fields[] = {
    {"match", 1}
};


/*
 *  setup
 *      Called by dynlib interface code to setup this plugin.  This
 *      routine should set up options handling, if required.
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plugin takes over output
 *  Side Effects:
 */
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
{
    if (DYNLIB_CUT != appType) {
        skAppPrintErr("%s is only for use by rwcut\n", pluginName);
        return DYNLIB_FAILED;
    }

    skAppContextSet(dynlibGetAppContext(dlISP));
    return DYNLIB_WONTPROCESS;
}


/*
 *  teardown
 *      Called by dynlib interface code to tear down this plugin.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    return;
}


/*
 *  int configure(dynlibInfoStruct_t *dlISP, void *data)
 *
 *    Perform additional configuration of the plug-in prior to
 *    processing the options.
 */
int configure(dynlibInfoStruct_t *dlISP, void *data)
{
    const dynlibSymbolId appType = dynlibGetAppType(dlISP);

    switch(appType) {
      case DYNLIB_CUT:
        /* for these apps, data is a pointer into which we should copy
         * the address of the fields this plug-in provides to the
         * application */
        if (data == NULL) {
            return -1;
        } else {
            sk_stringmap_entry_t **fields_ptr = (sk_stringmap_entry_t **)data;
            *fields_ptr = fields;
        }
        break;

      default:
        /* nothing to do */
        break;
    }

    return 0;
}


/*
 *  cut
 *      Called by dynlib interface code to output records
 *  Arguments:
 *      which field to output (indexed from 1)
 *      a string buffer to hold the output
 *      the length of the buffer
 *      a RW record
 *  Returns:
 *      if field is zero, the number of supported fields
 *      if out is NULL and rwrec is NULL, the size of buffer needed
 *         for a title for this field
 *      if out is NULL and rwrec is non-NULL, the size of buffer
 *         needed for the value to be output for this field
 *      if out is non-NULL and rwrec is NULL, outputs the title,
 *         returns length of buffer necessary to hold full string
 *      if out is non-NULL and rwrec is non-NULL, outputs the value,
 *         returns length of buffer necessary to hold full string
 *  Side Effects:
 *      None
 */
int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec)
{
    static const char title[] = "<->Match#";
    static const char *match_dir[] = {"->", "<-"};
    uint32_t nhip;
    uint32_t match_count;

    switch (field) {
      case 0:
        /* caller wants number of supported fields */
        return 1;

      case 1:
        /* Common case first: caller wants body */
        if (out && rwrec) {
            nhip = rwRecGetNhIPv4(rwrec);
            match_count = 0x00FFFFFF & nhip;
            if (match_count) {
                return snprintf(out, len_out, "%s%8u",
                                match_dir[(nhip & 0xFF000000) ? 1 : 0],
                                match_count);
            } else {
                return snprintf(out, len_out, "%-10s",
                                match_dir[(nhip & 0xFF000000) ? 1 : 0]);
            }
        }
        if (NULL == rwrec) {
            if (NULL == out) {
                /* caller wants title length */
                return 11;
            }
            /* caller wants title string */
            return snprintf(out, len_out, "%s", title);
        }

        if (NULL == out) {
            /* caller wants max body length */
            return 11;
        }
        break;
    }

    /* so there must be a problem somewhere. */
    return -1;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
