/*
** Copyright (C) 2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: flowrate.c 12521 2008-09-18 14:33:01Z mthomas $");

#include "rwpack.h"
#include "dynlib.h"
#include <float.h>

/*
**  flowrate.c
**
**    Plug-in to allow filtering, sorting, counting, and printing of a
**    packets-per-second, bytes-per-second, and bytes-per-packet
**    values.
**
**    Mark Thomas
**    July 2008
*/


/* EXPORTED FUNCTIONS */

/* The dynamic library glue code calls these */
int dynlib_api_version(void);
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType);
int configure(dynlibInfoStruct_t *dlISP, void *data);
void teardown(dynlibSymbolId UNUSED(appType));
void optionsUsage(dynlibSymbolId appType, FILE *fh);
int filter(rwRec *rwrec);
int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec);
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
int uniq(
    unsigned int    field,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);


/* TYPEDEFS AND DEFINES */

#define PCKTS_PER_SEC       1
#define BYTES_PER_SEC       2
#define BYTES_PER_PACKET    3

/* in rwsort and rwuniq, we encode the value as a uint64_t in
 * big-endian format.  To get the uint64_t, we multiply the double
 * value by this precision. */
#define PRECISION 1e4

#define DOUBLE_TO_UINT64(v) ((uint64_t)((v) * PRECISION))
#define UINT64_TO_DOUBLE(v) ((double)(v) / PRECISION)

/* how to calculate the packets per second.  rwRecGetElapsed() returns
 * the value in milliseconds.  If the elapsed time is 0, just return
 * the packet count. */
#define PCKT_RATE_DOUBLE(r)                                             \
    ((rwRecGetElapsed(r) == 0)                                          \
     ? rwRecGetPkts(r)                                                  \
     : ((double)(rwRecGetPkts(r)) * 1000.0 / (double)(rwRecGetElapsed(r))))

/* as above, but for bytes-per-second */
#define BYTE_RATE_DOUBLE(r)                                             \
    ((rwRecGetElapsed(r) == 0)                                          \
     ? rwRecGetBytes(r)                                                 \
     : ((double)(rwRecGetBytes(r)) * 1000.0 / (double)(rwRecGetElapsed(r))))

/* as above, for bytes-per-packet */
#define BYTES_PER_PACKET_DOUBLE(r)                              \
    ((double)rwRecGetBytes(r) / (double)rwRecGetPkts(r))

/* structure to hold min and max values */
typedef struct value_range_st {
    double      min;
    double      max;
    unsigned    is_active :1;
} value_range_t;


/* LOCAL VARIABLES */

static const char *plugin_name = "flowrate";

/* for filtering, pass records whose packets-per-second and
 * bytes-per-second values fall within these range. */
static value_range_t pckt_rate = {0, DBL_MAX, 0};
static value_range_t byte_rate = {0, DBL_MAX, 0};

typedef enum plugin_options_en {
    OPT_PACKETS_PER_SECOND,
    OPT_BYTES_PER_SECOND
} plugin_options_enum;

static struct option plugin_options[] = {
    {"packets-per-second",  REQUIRED_ARG, 0, OPT_PACKETS_PER_SECOND},
    {"bytes-per-second",    REQUIRED_ARG, 0, OPT_BYTES_PER_SECOND},
    {0, 0, 0, 0}            /* sentinel */
};

static const char *plugin_help[] = {
    "Packets-per-second count is within decimal range X-Y.",
    "Bytes-per-second count is within decimal range X-Y.",
    NULL
};

/* fields for rwcut, rwuniq, etc */
static sk_stringmap_entry_t plugin_fields[] = {
    {"pckts/sec",       PCKTS_PER_SEC},
    {"bytes/sec",       BYTES_PER_SEC},
    {"bytes/packet",    BYTES_PER_PACKET},
    {NULL,              UINT32_MAX}     /* sentinel */
};


/* LOCAL FUNCTION PROTOTYPES */

static int optionsHandler(clientData cData, int opt_index, char *opt_arg);
static int parseDecimalRange(
    value_range_t  *range,
    const char     *range_string,
    int             opt_index);


/* FUNCTION DEFINITIONS */

/*
 *  int dynlib_api_version(void);
 *
 *    Return the dynlib API version that this plugin was compiled
 *    against.
 */
int dynlib_api_version(void)
{
    return DYNLIB_API_VERSION;
}


/*
 * int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
 *
 * Called by dynlib interface code to set up the plugin.  This routine
 * should set up options handling if required.  Arguments: pointer to
 * the dynamic library interface structure, application type that is
 * using this plugin.  Returns DYNLIB_FAILED if processing fails,
 * DYNLIB_WONTPROCESS if the application should do normal output, and
 * DYNLIB_WILLPROCESS if this plugin takes over output.
 */
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
{
    /* Make the plugin use the application's context, so we can add
     * options to main app. */
    skAppContextSet(dynlibGetAppContext(dlISP));

    switch (appType) {
      case DYNLIB_SHAR_FILTER:
        if (skOptionsRegister(plugin_options, &optionsHandler,
                              (clientData)dlISP))
        {
            skAppPrintErr("Unable to register plugin options.");
            return DYNLIB_FAILED;
        }
        break;

      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        /* no options.  the fields will get added in the 'configure'
         * function. */
        break;

      default:
        skAppPrintErr("Cannot use %s plug-in with %s application",
                      plugin_name, skAppName());
        return DYNLIB_FAILED;
    }

    return DYNLIB_WONTPROCESS;
}


/*
 *  int configure(dynlibInfoStruct_t *dlISP, void *data)
 *
 *    Perform additional configuration of the plug-in prior to
 *    processing the options.
 */
int configure(dynlibInfoStruct_t *dlISP, void *data)
{
    const dynlibSymbolId appType = dynlibGetAppType(dlISP);

    switch(appType) {
      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        /* for these apps, data is a pointer into which we should copy
         * the address of the fields this plug-in provides to the
         * application */
        if (data == NULL) {
            return -1;
        } else {
            sk_stringmap_entry_t **fields_ptr = (sk_stringmap_entry_t **)data;
            *fields_ptr = plugin_fields;
        }
        break;

      default:
        /* nothing to do */
        break;
    }

    return 0;
}


/*
 *  void teardown(dynlibSymbolId appType);
 *
 *    Called by dynlib interface code to tear down this plugin.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    return;
}


/*
 * void optionsUsage(dynlibSymbolId UNUSED(appType), FILE *fh);
 *
 * Called by the dynlib interface to allow this plugin to print the
 * options it accepts.
 */
void optionsUsage(dynlibSymbolId appType, FILE *fh)
{
    int i;

    if (appType != DYNLIB_SHAR_FILTER) {
        return;
    }
    for (i = 0; plugin_options[i].name; i++) {
        fprintf(fh, "--%s %s. %s\n", plugin_options[i].name,
                SK_OPTION_HAS_ARG(plugin_options[i]), plugin_help[i]);
    }
}


/*
 * int optionsHandler(clientData cData, int opt_index, char *opt_arg);
 *
 * Handles options for the plugin.  cData is the dynlib ISP (which can
 * be used to find which application we are, among other things.)
 * opt_index is the enum passed in with the options descriptions.  opt_arg
 * is the argument, or NULL if no argument was given.  Returns 0 on
 * success, or 1 if there was a problem.
 */
static int optionsHandler(clientData cData, int opt_index, char *opt_arg)
{
    const dynlibSymbolId appType =dynlibGetAppType((dynlibInfoStruct_t*)cData);

    if (DYNLIB_SHAR_FILTER != appType) {
        skAppPrintErr("The %s plug-in does not support this application",
                      plugin_name);
        return 1;
    }

    switch ((plugin_options_enum)opt_index) {
      case OPT_PACKETS_PER_SECOND:
        if (parseDecimalRange(&pckt_rate, opt_arg, opt_index)) {
            return 1;
        }
        break;

      case OPT_BYTES_PER_SECOND:
        if (parseDecimalRange(&byte_rate, opt_arg, opt_index)) {
            return 1;
        }
        break;
    }

    return 0;
}


static int parseDecimalRange(
    value_range_t  *range,
    const char     *range_string,
    int             opt_index)
{
    double val;
    int rv;

    range->is_active = 1;

    /* parse first part of range; rv, if positive, will be location of
     * the hyphen (or other non-number). */
    rv = skStringParseDouble(&val, range_string, 0.0, 0.0);
    if (rv < 0) {
        skAppPrintErr("Invalid %s '%s': %s",
                      plugin_options[opt_index].name, range_string,
                      skStringParseStrerror(rv));
        return 1;
    }

    range->min = val;

    if (rv == 0) {
        /* single value, make range this single value */
        range->max = range->min;
        return 0;
    }

    if (range_string[rv] != '-') {
        /* stopped on non-hyphen */
        if (isspace((int)range_string[rv])) {
            skAppPrintErr("Invalid %s '%s': Embedded whitespace in range",
                          plugin_options[opt_index].name, range_string);
        } else {
            skAppPrintErr("Invalid %s '%s': Unexpected character '%c'",
                          plugin_options[opt_index].name, range_string,
                          range_string[rv]);
        }
        return 1;
    }

    /* move to start of upper bound */
    ++rv;

    /* make certain character after '-' is a digit or decimal point */
    if ((range_string[rv] != '.') && !isdigit((int)range_string[rv])) {
        /* check for open-ended range, permit whitespace between '-'
         * and end-of-string */
        const char *cp = &(range_string[rv]);
        while (isspace((int)*cp)) {
            ++cp;
        }
        if (*cp == '\0') {
            /* set upper bound to maximum allowed value */
            range->max = DBL_MAX;
            return 0;
        }
        /* got some junk char or embedded whitespace */
        if (cp != &(range_string[rv])) {
            skAppPrintErr("Invalid %s '%s': Embedded whitespace in range",
                          plugin_options[opt_index].name, range_string);
        } else {
            skAppPrintErr("Invalid %s '%s': Unexpected character '%c'",
                          plugin_options[opt_index].name, range_string,
                          range_string[rv]);
        }
        return 1;
    }

    /* parse upper part of range */
    rv = skStringParseDouble(&val, &range_string[rv], 0.0, 0.0);
    if (rv != 0) {
        skAppPrintErr("Invalid %s '%s': %s",
                      plugin_options[opt_index].name, range_string,
                      skStringParseStrerror(rv));
        return 1;
    }

    range->max = val;

    if (range->max < range->min) {
        skAppPrintErr("Invalid %s '%s': Maximum less than minimum",
                      plugin_options[opt_index].name, range_string);
        return 1;
    }

    return 0;
}


/*
 *  int filter(rwRec *rwrec);
 *
 *    The function actually used to implement filtering for the filter
 *    plugin.  Returns 0 if the record passes the filter, 1 if it
 *    fails the filter.
 */
int filter(rwRec *rwrec)
{
    double rate;

    if (pckt_rate.is_active) {
        rate = PCKT_RATE_DOUBLE(rwrec);
        if (rate < pckt_rate.min || rate > pckt_rate.max) {
            /* failed */
            return 1;
        }
    }

    if (byte_rate.is_active) {
        rate = BYTE_RATE_DOUBLE(rwrec);
        if (rate < byte_rate.min || rate > byte_rate.max) {
            /* failed */
            return 1;
        }
    }

    return 0;
}


static int getTitle(
    char       *text_value,
    size_t      text_size,
    void       *idx)
{
    assert(*((unsigned int*)(idx)) >= 1);
    assert(*((unsigned int*)(idx)) <= 3);

    return snprintf(text_value, text_size, "%s",
                    plugin_fields[(*((unsigned int*)(idx)) - 1)].name);
}


static int recToText(
    const rwRec    *rwrec,
    char           *text_value,
    size_t          text_size,
    void           *idx)
{
    switch (*((unsigned int*)(idx))) {
      case PCKTS_PER_SEC:
        return snprintf(text_value, text_size, "%.3f",
                        PCKT_RATE_DOUBLE(rwrec));
      case BYTES_PER_SEC:
        return snprintf(text_value, text_size, "%.3f",
                        BYTE_RATE_DOUBLE(rwrec));
      case BYTES_PER_PACKET:
        return snprintf(text_value, text_size, "%.3f",
                        BYTES_PER_PACKET_DOUBLE(rwrec));
    }
    return -1;
}


static int recToBin(const rwRec *rwrec, uint8_t *bin_value, void *idx)
{
    uint64_t rate_u64;
    double rate_d;

    switch (*((unsigned int*)(idx))) {
      case PCKTS_PER_SEC:
        rate_d = PCKT_RATE_DOUBLE(rwrec);
        break;
      case BYTES_PER_SEC:
        rate_d = BYTE_RATE_DOUBLE(rwrec);
        break;
      case BYTES_PER_PACKET:
        rate_d = BYTES_PER_PACKET_DOUBLE(rwrec);
        break;
      default:
        return -1;
    }

    rate_u64 = DOUBLE_TO_UINT64(rate_d);
    rate_u64 = hton64(rate_u64);
    memcpy(bin_value, &rate_u64, sizeof(uint64_t));

    /* return width of binary value */
    return (int)sizeof(uint64_t);
}


static int binToText(
    const uint8_t  *bin_value,
    char           *text_value,
    size_t          text_size,
    void    UNUSED(*idx))
{
    /* return the text value given the binary value */
    uint64_t rate_u64;
    double rate_d;

    assert(*((unsigned int*)(idx)) >= 1);
    assert(*((unsigned int*)(idx)) <= 3);

    memcpy(&rate_u64, bin_value, sizeof(uint64_t));
    rate_d = UINT64_TO_DOUBLE(ntoh64(rate_u64));

    return snprintf(text_value, text_size, "%.3f", rate_d);
}


/*
 * int cut(unsigned int field, char *text_value, size_t text_size, rwRec *rwrec);
 *
 * If "field" is 0, then returns the number of supported fields.
 *
 * If "text_value" is NULL and "rwrec" is NULL, returns the size of
 * the buffer needed to store the title for this field.
 *
 * If "text_value" is NULL and "rwrec" is non-NULL, returns the size
 * of the buffer needed to store the value for this field.
 *
 * If "text_value" is non-NULL and "rwrec" is NULL, writes the title
 * into "text_value", and returns the number of bytes needed to store
 * that string.
 *
 * If "text_value" is non-NULL and "rwrec" is non-NULL, writes the
 * value for that record into "text_value", and returns the number of
 * bytes needed to store that string.
 */
int cut(unsigned int field, char *text_value, size_t text_size, rwRec *rwrec)
{
    if ((NULL != rwrec) && (NULL != text_value)) {
        return recToText(rwrec, text_value, text_size, &field);
    }
    if (NULL != text_value) {
        return getTitle(text_value, text_size, &field);
    }
    if (NULL != rwrec) {
        /* textual column width */
        return 16;
    }

    assert(field >= 1);
    assert(field <= 3);

    return 1 + strlen(plugin_fields[(field - 1)].name);
}


/*
 *  int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with a binary value that rwsort will
 *                use for sorting; rwsort will not ask the plug-in for
 *                more data; this value should be everything rwsort
 *                needs to know.  rwsort will use memcmp() to compare
 *                values; the values should be in big-endian byte
 *                order.  THIS BUFFER MAY NOT BE WORD ALIGNED.
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' or 'rec' is NULL
 *          => the number of bytes required to hold the binary value
 *             of field 'f'
 *        when 'bin_val' and 'rec' are non-NULL
 *          => length of the binary value; uses 'rec' to compute the
 *             value of field 'f' and writes that binary value into
 *             'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 */
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec)
{
    if (bin_value && rwrec) {
        return recToBin(rwrec, bin_value, &field);
    }

    /* return width of binary value */
    return sizeof(uint64_t);
}


/*
 *  int uniq(unsigned int f, uint8_t bin_val, char *text_value, size_t text_size, rwRec *rec)
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with binary value, or buffer to pass
 *                in the binary value when converting to text.
 *      text_value: buffer to fill with the textual value
 *      text_size: length of the buffer 'text_value'
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' is NULL
 *          => the result of running cut(f, text_value, text_size, rwrec)
 *        when 'bin_val' is non-NULL
 *          when 'text_value' is non-NULL and 'rec' has any value
 *            => length of buffer necessary to hold the complete
 *               textual value; fills 'text_value' with the first
 *               'text_size'-1 bytes of converting the binary value in
 *               'bin_val', corresponding to field 'f', to text.
 *          when 'text_value' is NULL and 'text_size' has any value
 *            when 'rec' is non-NULL
 *              => length of the binary value; uses 'rec' to compute
 *                 the value of field 'f' and writes that binary value
 *                 into 'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 *             when 'rec' is NULL
 *               => the number of bytes required to hold the binary
 *                  value of field 'f'
 */
int uniq(
    unsigned int    field,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_size,
    rwRec          *rwrec)
{
    if (bin_value) {
        if (text_value) {
            return binToText(bin_value, text_value, text_size, &field);
        }

        if (rwrec) {
            /* return the binary value for the given rwrec */
            return recToBin(rwrec, bin_value, &field);
        }

        /* return width of binary value */
        return sizeof(uint64_t);
    }

    assert(field >= 1);
    assert(field <= 3);

    /* handle the title and width of textual field */
    if (NULL != text_value) {
        return getTitle(text_value, text_size, &field);
    }
    if (NULL != rwrec) {
        /* textual column width */
        return 16;
    }
    /* title width */
    return 1 + strlen(plugin_fields[(field - 1)].name);
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
