/*
** Copyright (C) 2007-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  silkpython.c
**
**  Python plugin for rwfilter, rwcut, rwsort, and rwuniq
*/

#include "silk.h"

RCSIDENT("$SiLK: silkpython.c 13074 2008-11-20 20:28:45Z mthomas $");

#include "utils.h"
#include "dynlib.h"
#include "rwpack.h"
#include <dlfcn.h>
#include <Python.h>
#include <osdefs.h>

/* The name of the Python function rwfilter will call for each record */
#define PYFILTER_NAME "rwfilter"

/* The name of the Python function rwfilter will call before exiting */
#define FINALIZER_NAME "finalize"

/* The name of the rwrec when an expression is accepted on the command
   line */
#define PYREC_NAME "rec"

/* Block size when reading python files over an skStream */
#define FILE_BLOCK_SIZE SKIOBUF_DEFAULT_BLOCKSIZE

typedef enum {
    OPT_PYTHON_FILE,
    OPT_PYTHON_EXPR
} rwp_expr_option_t;

/* exported functions--called by dynlib */
int dynlib_api_version(void);
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType);
void teardown(dynlibSymbolId appType);
int configure(dynlibInfoStruct_t *dlISP, void *data);
void optionsUsage(dynlibSymbolId appType, FILE *fh);
int supportsThreads(dynlibInfoStruct_t *dlISP);
int filter(rwRec *rwrec);
int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec);
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
int uniq(unsigned int field, uint8_t *bin_value, char *text_value,
         size_t text_len, rwRec *rwrec);

/* private function prototypes */
static int libOptionsSetup(dynlibInfoStruct_t *dlISP, dynlibSymbolId aType);
static int optionsHandler(clientData cData, int opt_index, char *opt_arg);
static int python_initialize(dynlibSymbolId appType);
static void python_uninitialize(void);

/* The python filename */
static char *python_filename = NULL;

/* The python expression */
static char *python_expr = NULL;

/* The rwfilter function */
static PyObject *pyfilter = NULL;

/* The finalize function */
static PyObject *pyfinalize = NULL;

/* The number of fields */
static ssize_t num_fields = 0;

/* The field data */
static PyObject *field_data = NULL;

/* Field entries */
static sk_stringmap_entry_t *field_map = NULL;

/* The python globals */
static PyObject *python_globals = NULL;

/* The rwrec to raw python function */
static PyObject *rwrec_to_raw_python = NULL;

/* The rwrec to python function */
static PyObject *rwrec_to_python_fn = NULL;

/* The record wrapper */
static PyObject *python_rec = NULL;

/* An empty tuple */
static PyObject *empty_tuple = NULL;

/* A keyword dictionary */
static PyObject *kwd_dict = NULL;

/* Whether to act as a real plugin or just ignore.  Set to 1 on an MPI
   master. */
static int ignore_plugin = 0;

/* Plugin's name */
static const char *pluginName = "python-plugin";

/* the options and help strings.  these vary depending on the type of
 * app that is using the shared library, and they get set to one of
 * the arrays below.
 */
static struct option *libOptions = NULL;
static const char **optionsHelp = NULL;
static int libOptionsCount = -1;

/* Options and Help for filtering */
static struct option optionsFilter[] = {
    {"python-file",       REQUIRED_ARG, 0, OPT_PYTHON_FILE},
    {"python-expr",       REQUIRED_ARG, 0, OPT_PYTHON_EXPR},
    /* include your options for rwfilter here */
    {0, 0, 0, 0}
};
static const char *optionsHelpFilter[] = {
    ("Executes the function \"" PYFILTER_NAME "\" from the\n"
     "\tgiven file over all the records"),
    ("Uses the return value of given python expression\n"
     "\tas the pass/fail determiner (flow record is called \""
     PYREC_NAME "\")"),
    (char*)NULL
};

/* Options and Help for filtering */
static struct option optionsField[] = {
    {"python-file",       REQUIRED_ARG, 0, OPT_PYTHON_FILE},
    /* include your options for rwcut here */
    {0, 0, 0, 0}
};
static const char *optionsHelpField[] = {
    ("Uses the field data registered by\n"
     "\t'register_plugin_field' in the given python file as extra fields"),
    (char*)NULL
};


/* FUNCTION DEFINITIONS */

/*
 *  int dynlib_api_version(void);
 *
 *    Return the dynlib API version that this plugin was compiled
 *    against.
 */
int dynlib_api_version(void)
{
    return DYNLIB_API_VERSION;
}


/*
 *  setup
 *      Called by dynlib interface code to setup this plugin.  This
 *      routine should set up options handling, if required.
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      DYNLIB_FAILED - if processing fails
 *      DYNLIB_WONTPROCESS - if application should do normal output
 *      DYNLIB_WILLPROCESS - if this plugin takes over output
 *  Side Effects:
 */
int setup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
{
    skAppContextSet(dynlibGetAppContext(dlISP));
    if (libOptionsSetup(dlISP, appType)) {
        return DYNLIB_FAILED;
    }
    return DYNLIB_WONTPROCESS;
}


/*
 *  teardown
 *      Called by dynlib interface code to tear down this plugin.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void teardown(dynlibSymbolId UNUSED(appType))
{
    if (Py_IsInitialized()) {
        if (appType == DYNLIB_SHAR_FILTER && pyfinalize) {
            PyObject *rv = PyObject_CallObject(pyfinalize, NULL);
            if (rv == NULL) {
                PyErr_Print();
                PyErr_Clear();
            } else {
                Py_DECREF(rv);
            }
        }

        python_uninitialize();
    }
    return;
}


/*
 *  optionsUsage
 *      Called by dynlib interface code to allow this plugin to print
 *      the options it accepts.
 *  Arguments:
 *      --Application type that is using this plugin
 *  Returns:
 *      None.
 *  Side Effects:
 *      None.
 */
void optionsUsage(dynlibSymbolId UNUSED(appType), FILE *fh)
{
    int i;

    for (i = 0; i < libOptionsCount; ++i) {
        fprintf(fh, "--%s %s. %s\n", libOptions[i].name,
                libOptions[i].has_arg ? libOptions[i].has_arg == 1
                ? "Req. Arg" : "Opt. Arg" : "No Arg", optionsHelp[i]);
    }

    return;
}


/*
 *  libOptionsSetup
 *      Sets up the options handler for this plugin
 *  Arguments:
 *      --pointer to the dynamic library interface structure
 *      --Application type that is using this plugin.
 *  Returns:
 *      0 if o.k.; nonzero otherwise.
 *  Side Effects:
 *      Options are registered.
 */
static int libOptionsSetup(dynlibInfoStruct_t *dlISP, dynlibSymbolId appType)
{
    int help_size = 0;

    if (-1 != libOptionsCount) {
        /* we've been here before */
        return 0;
    }

    switch (appType) {
      case DYNLIB_SHAR_FILTER:
        libOptions = optionsFilter;
        optionsHelp = optionsHelpFilter;
        break;
      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        libOptions = optionsField;
        optionsHelp = optionsHelpField;
        break;
      default:
        libOptionsCount = 0;
        return 0;
    }

    for (libOptionsCount=0;
         libOptions[libOptionsCount].name;
         ++libOptionsCount)
        ; /* NO BODY */
    for (help_size = 0; optionsHelp[help_size]; ++help_size)
        ; /* NO BODY */

    /* verify same number of options and help strings */
    if (help_size != libOptionsCount) {
        skAppPrintErr("%s: mismatch in option (%d) and help (%d) counts",
                      pluginName, libOptionsCount, help_size);
        return 1;
    }

    if (skOptionsRegister(libOptions, &optionsHandler, (clientData)dlISP)) {
        skAppPrintErr("Unable to register options");
        return 1;
    }

    return 0;
}


/*
 *  optionsHandler
 *      Called by options parser to handle one user option
 *  Arguments:
 *      clientData cData: the dynamic lib interface struct pointer
 *      int index: index into appOptions of the specific option
 *      char *optarg: the argument; 0 if no argument was required/given.
 * Returns:
 *      0 if OK. 1 else.
 * Side Effects:
 *      Relevant options are set.
 */
static int optionsHandler(
    clientData cData,
    int        opt_index,
    char      *opt_arg)
{
    dynlibInfoStruct_t *dlISP = (dynlibInfoStruct_t *)cData;

    switch (opt_index) {

      case OPT_PYTHON_FILE:
        if (python_filename != NULL) {
            skAppPrintErr("Cannot use --%s more than once",
                          optionsFilter[OPT_PYTHON_FILE].name);
            return 1;
        }
        if (python_expr != NULL) {
            skAppPrintErr("Cannot use --%s at the same time as --%s",
                          optionsFilter[OPT_PYTHON_FILE].name,
                          optionsFilter[OPT_PYTHON_EXPR].name);
            return 1;
        }
        python_filename = opt_arg;

        /* Initialize python to parse file and set field values */
        if (python_initialize(dynlibGetAppType(dlISP)) != 0) {
            return 1;
        }

        break;

      case OPT_PYTHON_EXPR:
        if (python_expr != NULL) {
            skAppPrintErr("Cannot use --%s more than once",
                          optionsFilter[OPT_PYTHON_FILE].name);
            return 1;
        }
        if (python_filename != NULL) {
            skAppPrintErr("Cannot use --%s at the same time as --%s",
                          optionsFilter[OPT_PYTHON_FILE].name,
                          optionsFilter[OPT_PYTHON_EXPR].name);
            return 1;
        }
        python_expr = opt_arg;

        /* Initialize python to parse expression */
        if (python_initialize(dynlibGetAppType(dlISP)) != 0) {
            return 1;
        }

        break;

      default:
        skAppPrintErr("Unrecognized option");
        return 1;
    } /* switch */

    dynlibMakeActive(dlISP);

    return 0;
}


/*
 *  threads_ok = supportsThreads(dlISP);
 *
 *    Return 1 if the plug-in can run in a threaded environment, 0 if not.
 */
int supportsThreads(
    dynlibInfoStruct_t UNUSED(*dlISP))
{
    /* make sure we run in an unthreaded environment */
    return 0;
}

/*
 *  python_initialize
 *      Initialize python
 */
static int python_initialize(
    dynlibSymbolId appType)
{
    PyObject              *pysilk_module = NULL;
    PyObject              *silk_module   = NULL;
    PyObject              *plugin_module = NULL;
    PyObject              *function      = NULL;
    PyObject              *builtin       = NULL;
    PyObject              *none          = NULL;
    PyObject              *compiled      = NULL;
    PyObject              *result        = NULL;
    PyObject              *seq           = NULL;
    skstream_t            *stream        = NULL;
    char                  *contents      = NULL;
    char                  *old_path;
    char                  *new_path;
    size_t                 path_len;
    int                    retval        = 0;
    int                    rv;

    if (python_filename) {
        /* Get a handle to the python file before starting up python.
           This is so, if we are running on an MPI master, we don't
           have to bother starting python. */
        rv = dynlibOpenDataInputStream(&stream, SK_CONTENT_TEXT,
                                       python_filename);
        if (rv == 1) {
            ignore_plugin = 1;
            return 0;
        }
        if (rv != 0) {
            skAppPrintErr("Could not access %s", python_filename);
            return -1;
        }
    }

    if (Py_IsInitialized()) {
        return 0;
    }

    if (appType != DYNLIB_SHAR_FILTER && appType != DYNLIB_CUT &&
        appType != DYNLIB_SORT && appType != DYNLIB_UNIQ)
    {
        return -1;
    }

    assert(python_filename != NULL || python_expr != NULL);

    Py_InitializeEx(0);

    /* Import the silk plugin module */
    plugin_module = PyImport_ImportModule("silk.plugin");
    if (plugin_module == NULL) {
        skAppPrintErr("Could not load the \"silk.plugin\" python module");
        goto clean_err;
    }

    /* Set the program name */
    result = PyObject_CallMethod(plugin_module, "_set_programname", "s",
                                 skAppName());
    if (result == NULL) {
        goto err;
    }

    /* Import the silk modules */
    silk_module = PyImport_ImportModule("silk");
    if (silk_module == NULL) {
        skAppPrintErr("Could not load the \"silk\" python module");
        goto err;
    }
    pysilk_module = PyImport_ImportModule("silk.pysilk_nl");
    if (pysilk_module == NULL) {
        skAppPrintErr("Could not load the \"silk.pysilk_nl\" python module");
        goto err;
    }

    /* Import the builtin module */
    builtin = PyImport_ImportModule("__builtin__");
    if (builtin == NULL) {
        skAppPrintErr("Could not load the \"__builtin__\" module");
        goto err;
    }

    /* Get conversion functions for rwrec objects  */
    rwrec_to_raw_python = PyObject_GetAttrString(pysilk_module,
                                                 "_raw_rwrec_copy");
    if (rwrec_to_raw_python == NULL) {
        skAppPrintErr("Could not find the \"silk._raw_rwrec_copy\" function");
        goto err;
    }
    rwrec_to_python_fn = PyObject_GetAttrString(silk_module, "RWRec");
    if (rwrec_to_python_fn == NULL) {
        skAppPrintErr("Could not find the \"silk.RWRec\" function");
        goto err;
    }


    /* Create the globals dictionary */
    python_globals = PyDict_New();
    if (python_globals == NULL) {
        goto err;
    }

    /* Create a keyword dictionary used in creating rwrec objects */
    kwd_dict = PyDict_New();
    if (kwd_dict == NULL) {
        goto err;
    }

    /* Create an empty tuple used in creating rwrec objects */
    empty_tuple = PyTuple_New(0);
    if (empty_tuple == NULL) {
        goto err;
    }

    /* Ensure cwd is in the python path */
    old_path = Py_GetPath();
    path_len = strlen(old_path) + 3;
    new_path = malloc(path_len);
    if (new_path == NULL) {
        goto err;
    }
    new_path[0] = '.';
    new_path[1] = DELIM;
    strcpy(&new_path[2], old_path);
    PySys_SetPath(new_path);
    free(new_path);

    /* Add register_plugin_field to the current globals */
    function = PyObject_GetAttrString(plugin_module, "_register_plugin_field");
    if (function == NULL) {
        goto err;
    }
    rv = PyDict_SetItemString(python_globals, "register_plugin_field",
                              function);
    Py_CLEAR(function);
    if (rv != 0) {
        goto err;
    }

    if (python_filename) {
        /* Load and evaluate a python file */
        size_t      size;
        size_t      remainder;
        char       *loc;
        ssize_t     num_read;

        contents = malloc(FILE_BLOCK_SIZE);
        if (contents == NULL) {
            skAppPrintErr("Memory error");
            goto err;
        }

        /* Read the stream's data into contents */
        loc = contents;
        size = FILE_BLOCK_SIZE;
        remainder = size;
        while ((num_read = skStreamRead(stream, loc, remainder)) > 0) {
            remainder -= num_read;
            loc += num_read;
            if (remainder == 0) {
                char *realloced;

                size += FILE_BLOCK_SIZE;
                realloced = realloc(contents, size);
                if (realloced == NULL) {
                    skAppPrintErr("Memory error");
                    goto err;
                }
                remainder = FILE_BLOCK_SIZE;
                contents = realloced;
                loc = contents + (size - FILE_BLOCK_SIZE);
            }
        }
        if (num_read != 0) {
            skAppPrintErr("Read error: %s",
                          strerror(skStreamGetLastErrno(stream)));
            goto err;
        }

        skStreamDestroy(&stream);

        /* Compile the code */
        compiled = PyObject_CallMethod(builtin, "compile", "s#ss",
                                       contents, loc - contents,
                                       python_filename, "exec");
        if (compiled == NULL) {
            skAppPrintErr("Could not compile %s", python_filename);
            goto err;
        }

        free(contents);
        contents = NULL;

        /* Then evaluate the code, putting the result in the
           python_globals */
        none = PyObject_CallMethod(builtin, "eval", "OO", compiled,
                                   python_globals);
        if (none == NULL || PyErr_Occurred()) {
            skAppPrintErr("Could not parse %s", python_filename);
            goto err;
        }
        Py_CLEAR(none);

    } else {
        /* Create and compile an expression as a function */
        static char *prefix = ("def " PYFILTER_NAME "(" PYREC_NAME "): "
                               "return ");
        size_t  len         = strlen(prefix) + strlen(python_expr);
        char   *command     = malloc(len + 1);

        if (command == NULL) {
            skAppPrintErr("Memory error allocating string");
            goto err;
        }

        /* Generate the function */
        strcpy(command, prefix);
        strcat(command, python_expr);

        /* Compile the function */
        compiled = PyObject_CallMethod(builtin, "compile", "sss",
                                       command, "<command-line>", "exec");
        if (compiled == NULL) {
            skAppPrintErr("Could not compile python expression");
            goto err;
        }

        /* Add the silk module to the python_globals, so the function
           can use it */
        rv = PyDict_SetItemString(python_globals, "silk", silk_module);
        if (rv != 0) {
            skAppPrintErr("Python dictionary error");
            goto err;
        }

        /* Evaluate the function, putting the result in
           python_globals */
        none = PyObject_CallMethod(builtin, "eval", "OO", compiled,
                                   python_globals);
        if (none == NULL || PyErr_Occurred()) {
            skAppPrintErr("Could not parse python expression");
            goto err;
        }
        Py_CLEAR(none);
    }


    /* Load the function data */
    switch (appType) {

      case DYNLIB_SHAR_FILTER:
        /* Set the pyfilter and pyfinalize functions for rwfilter */
        function = PyDict_GetItemString(python_globals, PYFILTER_NAME);
        if (function == NULL || !PyCallable_Check(function)) {
            skAppPrintErr(("The python function '" PYFILTER_NAME
                           "' was not defined"));
            goto err;
        }
        Py_INCREF(function);
        pyfilter = function;

        function = PyDict_GetItemString(python_globals, FINALIZER_NAME);
        if (function != NULL && PyCallable_Check(function)) {
            Py_INCREF(function);
            pyfinalize = function;
            function = NULL;
        }
        break;

      case DYNLIB_CUT:
        /* Get field data for rwcut */
        seq = PyObject_CallMethod(plugin_module, "_get_plugin_data", "s",
                                  "cut");
        if (seq == NULL) {
            goto err;
        }
        break;

      case DYNLIB_SORT:
        /* Get field data for rwsort */
        seq = PyObject_CallMethod(plugin_module, "_get_plugin_data", "s",
                                  "sort");
        if (seq == NULL) {
            goto err;
        }
        break;

      case DYNLIB_UNIQ:
        /* Get field data for rwuniq */
        seq = PyObject_CallMethod(plugin_module, "_get_plugin_data", "s",
                                  "uniq");
        if (seq == NULL) {
            goto err;
        }
        break;

      default:
        assert(0);
        abort();
    }

    /* Handle supported fields.  This includes setting field_data,
       num_fields, and creating the skstringmap field_map for passing
       back to the main program. */
    if (seq != NULL) {
        PyObject **o;
        int        i;

        field_data = PySequence_Fast(seq, "Invalid sequence");
        if (field_data == NULL) {
            goto err;
        }
        num_fields = PySequence_Fast_GET_SIZE(field_data);
        if (num_fields < 0) {
            goto err;
        }

        assert(field_map == NULL);
        field_map = calloc((num_fields + 1), sizeof(*field_map));
        if (field_map == NULL) {
            skAppPrintErr("Error creating string map");
            goto err;
        }

        for (i = 0, o = PySequence_Fast_ITEMS(field_data);
             i < num_fields;
             i++, o++)
        {
            PyObject   *name    = PyTuple_GET_ITEM(*o, 0);
            const char *strname = strdup(PyString_AS_STRING(name));
            if (strname == NULL) {
                skAppPrintErr("Memory error copying field names");
                goto err;
            }
            field_map[i].name = strname;
            field_map[i].id = i + 1;
        }
    }

    /* Create an empty CObject */
    python_rec = PyCObject_FromVoidPtr(NULL, NULL);
    if (python_rec == NULL) {
        goto err;
    }

  cleanup:
    if (contents) {
        free(contents);
    }
    skStreamDestroy(&stream);
    Py_XDECREF(pysilk_module);
    Py_XDECREF(silk_module);
    Py_XDECREF(plugin_module);
    Py_XDECREF(function);
    Py_XDECREF(builtin);
    Py_XDECREF(none);
    Py_XDECREF(compiled);
    Py_XDECREF(result);
    Py_XDECREF(seq);

    return retval;

  clean_err:
    PyErr_Clear();

  err:
    if (PyErr_Occurred()) {
        PyErr_Print();
        PyErr_Clear();
    }
    retval = -1;
    goto cleanup;

}


/*
 *  python_unintiialize
 *      Un-Initialize python
 */
static void python_uninitialize(void)
{
    if (Py_IsInitialized()) {
        Py_CLEAR(python_globals);
        Py_CLEAR(pyfilter);
        Py_CLEAR(pyfinalize);
        Py_CLEAR(field_data);
        Py_CLEAR(rwrec_to_raw_python);
        Py_CLEAR(rwrec_to_python_fn);
        Py_CLEAR(python_rec);
        Py_CLEAR(empty_tuple);
        Py_CLEAR(kwd_dict);

        Py_Finalize();
    }

    if (field_map != NULL) {
        sk_stringmap_entry_t *entry;
        for (entry = field_map; entry->name; entry++) {
            free((char *)entry->name);
        }
        free(field_map);
        field_map = NULL;
    }
}


/*
 *  int configure(dynlibInfoStruct_t *dlISP, void *data)
 *
 *    Perform additional configuration of the plug-in prior to
 *    processing the options.
 */
int configure(dynlibInfoStruct_t *dlISP, void *data)
{
    const dynlibSymbolId appType = dynlibGetAppType(dlISP);

    if (ignore_plugin) {
        return 0;
    }

    switch(appType) {
      case DYNLIB_CUT:
      case DYNLIB_SORT:
      case DYNLIB_UNIQ:
        /* for these apps, data is a pointer into which we should copy
         * the address of the fields this plug-in provides to the
         * application */
        if (data == NULL) {
            return -1;
        } else {
            sk_stringmap_entry_t **fields_ptr = (sk_stringmap_entry_t **)data;
            *fields_ptr = field_map;
        }
        break;

      default:
        /* nothing to do */
        break;
    }

    return 0;
}

/* This function creates an RWRec python object from a C rwRec pointer */
static PyObject *rwrec_to_python(rwRec *rwrec)
{
    int       rv;
    PyObject *rawrec;
    PyObject *rec;

    assert(!ignore_plugin);

    rv = PyCObject_SetVoidPtr(python_rec, rwrec);
    if (rv == 0) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    rawrec = PyObject_CallFunctionObjArgs(rwrec_to_raw_python,
                                          python_rec, NULL);
    if (rawrec == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    rv = PyDict_SetItemString(kwd_dict, "_clone", rawrec);
    if (rv != 0) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    Py_DECREF(rawrec);

    rec = PyObject_Call(rwrec_to_python_fn, empty_tuple, kwd_dict);
    if (rec == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    return rec;
}


/*
 *  filter
 *      This subroutine works in tadem with the standard filtering rules
 *      of rwfilter (and other shared filter plugins) to determine
 *      whether the rwRec should pass the filter.  This subroutine may
 *      not see all rwRec records since they may be filtered by one of
 *      the standard rules.  The subroutine should return 0 if the rwRec
 *      passes the filter; 1 otherwise.
 *
 *      The dynlib plug-in code will call this filter() routine ONCE per
 *      rw-record.  If the user has specified multiple filter-rules
 *      switches which a single plug-in handles, the plugin must process
 *      all the switches before returning.
 *  Arguments:
 *      a RW record
 *  Returns:
 *      0 to accept the record, 1 to reject the record
 *  Side Effects:
 *      None
 */
int filter(rwRec *rwrec)
{
    int rv;
    PyObject *retval;
    PyObject *rec;

    assert(!ignore_plugin);

    rec = rwrec_to_python(rwrec);

    retval = PyObject_CallFunctionObjArgs(pyfilter, rec, NULL);
    if (retval == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    rv = PyObject_IsTrue(retval);
    Py_DECREF(retval);
    Py_DECREF(rec);

    return (rv == 1) ? 0 : 1;
}

/*
 *  cut
 *      Called by dynlib interface code to output records
 *  Arguments:
 *      which field to output (indexed from 1)
 *      a string buffer to hold the output
 *      the length of the buffer
 *      a RW record
 *  Returns:
 *      if field is zero, the number of supported fields
 *      if out is NULL and rwrec is NULL, the size of buffer needed
 *         for a title for this field
 *      if out is NULL and rwrec is non-NULL, the size of buffer
 *         needed for the value to be output for this field
 *      if out is non-NULL and rwrec is NULL, outputs the title,
 *         returns length of buffer necessary to hold full string
 *      if out is non-NULL and rwrec is non-NULL, outputs the value,
 *         returns length of buffer necessary to hold full string
 *  Side Effects:
 *      None
 */
static int static_cut(
    unsigned int    field,
    char           *out,
    size_t          len_out,
    rwRec          *rwrec)
{
    PyObject *obj;
    int       len;
    PyObject *retval;
    PyObject *rec;
    PyObject *str;
    PyObject *fun;

    assert(!ignore_plugin);

    if (field == 0) {
        /* Request for number of supported fields */
        return num_fields;
    }

    if (field > (unsigned int)num_fields) {
        return -1;              /* Error: We don't support that field */
    }

    obj = PySequence_Fast_GET_ITEM(field_data, field - 1);
    assert(obj != NULL);

    if ((out == NULL) && (rwrec == NULL)) {
        /* Request for buffer size for field title */
        len = PyString_GET_SIZE(PyTuple_GET_ITEM(obj, 0)) + 1;
        return len;
    }

    if ((out != NULL) && rwrec == NULL) {
        /* Request for field title */
        len = snprintf(out, len_out,
                       "%s", PyString_AS_STRING(PyTuple_GET_ITEM(obj, 0)));
        return len;
    }

    if ((out == NULL) && (rwrec != NULL)) {
        /* Request for buffer size for field value */
        len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 1)) + 1;
        return len;
    }

    /* Request for field value */
    assert((out != NULL) && (rwrec != NULL));

    fun = PyTuple_GET_ITEM(obj, 2);
    assert(fun != NULL);
    Py_INCREF(fun);

    rec = rwrec_to_python(rwrec);

    retval = PyObject_CallFunctionObjArgs(fun, rec, NULL);
    if (retval == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    Py_DECREF(fun);
    Py_DECREF(rec);

    str = PyObject_Str(retval);
    if (str == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    len = snprintf(out, len_out, "%s", PyString_AS_STRING(str));

    Py_DECREF(str);
    Py_DECREF(retval);

    return len;
}

int cut(unsigned int field, char *out, size_t len_out, rwRec *rwrec)
{
    return static_cut(field, out, len_out, rwrec);
}



/*
 *  int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec);
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with a binary value that rwsort will
 *                use for sorting; rwsort will not ask the plug-in for
 *                more data; this value should be everything rwsort
 *                needs to know.  rwsort will use memcmp() to compare
 *                values; the values should be in big-endian byte
 *                order.  THIS BUFFER MAY NOT BE WORD ALIGNED.
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' or 'rec' is NULL
 *          => the number of bytes required to hold the binary value
 *             of field 'f'
 *        when 'bin_val' and 'rec' are non-NULL
 *          => length of the binary value; uses 'rec' to compute the
 *             value of field 'f' and writes that binary value into
 *             'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 */
int sort(unsigned int field, uint8_t *bin_value, rwRec *rwrec)
{
    PyObject *obj;
    int       len;
    PyObject *retval;
    PyObject *rec;
    PyObject *fun;
    char     *str;
    ssize_t   slen;

    assert(!ignore_plugin);

    if (field == 0) {
        /* Request for number of supported fields */
        return num_fields;
    }

    if (field > (unsigned int)num_fields) {
        return -1;              /* Error: We don't support that field */
    }

    obj = PySequence_Fast_GET_ITEM(field_data, field - 1);
    assert(obj != NULL);

    len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 1));

    if ((bin_value == NULL) || (rwrec == NULL)) {
        /* Request for buffer size for bin value */
        return len;
    }

    assert((bin_value != NULL) && (rwrec != NULL));

    fun = PyTuple_GET_ITEM(obj, 2);
    assert(fun != NULL);
    Py_INCREF(fun);

    rec = rwrec_to_python(rwrec);

    retval = PyObject_CallFunctionObjArgs(fun, rec, NULL);
    if (retval == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }

    Py_DECREF(rec);
    Py_DECREF(fun);

    str = PyString_AsString(retval);
    if (str == NULL) {
        PyErr_Print();
        PyErr_Clear();
        exit(EXIT_FAILURE);
    }
    slen = PyString_GET_SIZE(retval);
    if (slen != len) {
        skAppPrintErr("Binary sort value returned from python is the "
                      "wrong length");
        exit(EXIT_FAILURE);
    }

    memcpy(bin_value, str, len);

    Py_DECREF(retval);

    return len;
}


/*
 *  int uniq(unsigned int f, uint8_t bin_val, char *text_val, size_t
 *           text_len, rwRec *rec)
 *
 *  Arguments:
 *      f:        field to get the binary or textual value of
 *      bin_val:  buffer to fill with binary value, or buffer to pass
 *                in the binary value when converting to text.
 *      text_val: buffer to fill with the textual value
 *      text_len: length of the buffer 'out'
 *      rec:      SiLK flow record
 *  Returns:
 *      when 'f' is zero
 *        => the number of supported fields
 *      when 'f' is non-zero
 *        when 'bin_val' is NULL
 *          => the result of running cut(f, text_val, text_len, rwrec)
 *        when 'bin_val' is non-NULL
 *          when 'text_val' is non-NULL and 'rec' has any value
 *            => length of buffer necessary to hold the complete
 *               textual value; fills 'text_val' with the first
 *               'text_len'-1 bytes of converting the binary value in
 *               'bin_val', corresponding to field 'f', to text.
 *          when 'text_val' is NULL and 'text_len' has any value
 *            when 'rec' is non-NULL
 *              => length of the binary value; uses 'rec' to compute
 *                 the value of field 'f' and writes that binary value
 *                 into 'bin_val'.  ASSUMES 'bin_val' IS BIG ENOUGH!
 *             when 'rec' is NULL
 *               => the number of bytes required to hold the binary
 *                  value of field 'f'
 */
int uniq(
    unsigned int    field,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec)
{
    PyObject *obj;
    int       len;
    PyObject *retval;
    PyObject *fun;

    assert(!ignore_plugin);

    if (field == 0) {
        /* Request for number of supported fields */
        return num_fields;
    }

    if (field > (unsigned int)num_fields) {
        return -1;              /* Error: We don't support that field */
    }

    if (bin_value) {
        obj = PySequence_Fast_GET_ITEM(field_data, field - 1);
        assert(obj != NULL);

        if (text_value) {
            /* return the text value given the binary value */
            PyObject *bin;
            PyObject *pystr;

            Py_INCREF(obj);
            len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 2));
            bin = PyString_FromStringAndSize((char *)bin_value, len);
            if (bin == NULL) {
                PyErr_Print();
                PyErr_Clear();
                exit(EXIT_FAILURE);
            }
            fun = PyTuple_GET_ITEM(obj, 4);
            assert(fun != NULL);
            Py_INCREF(fun);

            len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 1));
            Py_DECREF(obj);

            retval = PyObject_CallFunctionObjArgs(fun, bin, NULL);
            if (retval == NULL) {
                PyErr_Print();
                PyErr_Clear();
                exit(EXIT_FAILURE);
            }
            Py_DECREF(fun);
            Py_DECREF(bin);
            pystr = PyObject_Str(retval);
            if (pystr == NULL) {
                PyErr_Print();
                PyErr_Clear();
                exit(EXIT_FAILURE);
            }
            len = snprintf(text_value, len, "%s", PyString_AS_STRING(pystr));
            Py_DECREF(pystr);
            Py_DECREF(retval);
            return len;
        } else if (rwrec) {
            PyObject *rec;
            char     *str;
            ssize_t   slen;

            /* return the binary value for the given rwrec */
            len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 2));
            fun = PyTuple_GET_ITEM(obj, 3);
            assert(fun != NULL);
            Py_INCREF(fun);

            rec = rwrec_to_python(rwrec);

            retval = PyObject_CallFunctionObjArgs(fun, rec, NULL);
            if (retval == NULL) {
                PyErr_Print();
                PyErr_Clear();
                exit(EXIT_FAILURE);
            }
            Py_DECREF(fun);

            str = PyString_AsString(retval);
            if (str == NULL) {
                PyErr_Print();
                PyErr_Clear();
                exit(EXIT_FAILURE);
            }
            slen = PyString_GET_SIZE(retval);
            if (slen != len) {
                skAppPrintErr("Binary uniq value returned from python is the "
                              "wrong length");
                exit(EXIT_FAILURE);
            }
            memcpy(bin_value, str, len);

            Py_DECREF(rec);
            Py_DECREF(retval);

            return len;
        } else {
            /* return the number of bytes in the binary value */
            len = PyInt_AS_LONG(PyTuple_GET_ITEM(obj, 2));
            return len;
        }
    } else {
        /* either return the title's width---when text_value and rwrec
         * are both NULL, the textual field width---when text_value is
         * NULL and rwrec is non-NULL, or the title---when text_value
         * is non-NULL and rwrec is NULL. */
        return static_cut(field, text_value, text_len, rwrec);;
    }

    return -1; /* NOTREACHED */
}

/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
