/*
** Copyright (C) 2001-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwfilterutils.c
**
**  utility routines for rwfilter.c
**
**  Suresh L. Konda
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwfilterutils.c 13074 2008-11-20 20:28:45Z mthomas $");

#include "rwfilter.h"


/* TYPEDEFS AND DEFINES */



/* LOCAL VARIABLES */

/* the compression method to use when writing the file.
 * sksiteCompmethodOptionsRegister() will set this to the default or
 * to the value the user specifies. */
static sk_compmethod_t comp_method;


/* OPTION SETUP */

/* these values used to index into following arrays.  need to keep
 * everything in sync */
typedef enum {
    OPT_DRY_RUN,
#if SK_RWFILTER_THREADED
    OPT_THREADS,
#endif
    OPT_MAX_PASS_RECORDS, OPT_PRINT_FILE, OPT_DYNLIB,
    OPT_INPUT_PIPE, OPT_XARGS,
    OPT_PASS_DEST, OPT_FAIL_DEST, OPT_ALL_DEST,
    OPT_PRINT_STAT, OPT_PRINT_VOLUME
} appOptionsEnum;

static struct option appOptions[] = {
    {"dry-run",                 NO_ARG,       0, OPT_DRY_RUN},
#if SK_RWFILTER_THREADED
    {"threads",                 REQUIRED_ARG, 0, OPT_THREADS},
#endif
    {"max-pass-records",        REQUIRED_ARG, 0, OPT_MAX_PASS_RECORDS},
    {"print-filenames",         NO_ARG,       0, OPT_PRINT_FILE},
    {OPT_DYNAMIC_LIBRARY,       REQUIRED_ARG, 0, OPT_DYNLIB},

    {"input-pipe",              REQUIRED_ARG, 0, OPT_INPUT_PIPE},
    {"xargs",                   REQUIRED_ARG, 0, OPT_XARGS},
    {"pass-destination",        REQUIRED_ARG, 0, OPT_PASS_DEST},
    {"fail-destination",        REQUIRED_ARG, 0, OPT_FAIL_DEST},
    {"all-destination",         REQUIRED_ARG, 0, OPT_ALL_DEST},
    {"print-statistics",        OPTIONAL_ARG, 0, OPT_PRINT_STAT},
    {"print-volume-statistics", OPTIONAL_ARG, 0, OPT_PRINT_VOLUME},
    {0,0,0,0}                   /* sentinel entry */
};

static const char *appHelp[] = {
    "Parse command line switches but do not process records",
#if SK_RWFILTER_THREADED
    "Use this number of threads. Def $SILK_RWFILTER_THREADS or 1",
#endif
    "Write at most this many records; 0 for all.  Def. 0",
    "Print names of input files during processing. Def. No",
    ("Augment processing with the specified dynamic\n"
     "\tlibrary. No default"),
    ("Read SiLK flow records from a pipe: 'stdin' or\n"
     "\tpath to named pipe. No default"),
    ("Read list of input file names from a file or pipe\n"
     "\tpathname or 'stdin'. No default"),
    ("Destination for records which pass the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Destination for records which fail the filter(s):\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Destination for all records regardless of pass/fail:\n"
     "\tpathname or 'stdout'. If pathname, it must not exist. No default"),
    ("Print a count of total flows read to named file.\n"
     "\tIf no pathname provided, use stderr. No default"),
    ("Print count of flows/packets/bytes read\n"
     "\tto named file. If no pathname provided, use stderr. No default"),
    (char *)NULL
};


/* LOCAL FUNCTION PROTOTYPES */

static void appUsageLong(void);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static int  appAddDynlib(const char *dlpath, int warnOnError);
static int  filterCheckInputs(int argc);
static int  filterCheckOutputs(void);
static int  filterOpenOutputs(void);
static int  filterSetCheckers(void);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); skOptionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
#define USAGE_MSG                                                           \
    ("<app-opts> <partition-opts> {<selection-opts> | <inputFiles>}\n"      \
     "\tPartitions SiLK Flow records into one or more 'pass' and/or\n"      \
     "\t'fail' output streams.  The source of the SiLK records can\n"       \
     "\tbe stdin, a named pipe, files listed on the command line, or\n"     \
     "\tfiles selected from the data-store via the selection switches.\n"   \
     "\tThere is no default input or output; these must be specified.\n")

    FILE *fh = USAGE_FH;
    int i, j;


    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);

    fprintf(fh, "\nGENERAL SWITCHES:\n\n");
    skOptionsDefaultUsage(fh);

    /* print everything before --input-pipe */
    for (i = 0; appOptions[i].name && appOptions[i].val < OPT_INPUT_PIPE; ++i){
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    skOptionsNotesUsage(fh);
    sksiteCompmethodOptionsUsage(fh);

    /* print remaining options */
    fprintf(fh, ("\nINPUT/OUTPUT SWITCHES."
                 " An input switch or a SELECTION switch (below) is\n"
                 "\trequired.  At least one output switch is required:\n\n"));
    for ( ; appOptions[i].name; ++i) {
        fprintf(fh, "--%s %s. %s\n", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]), appHelp[i]);
    }

    /* selection switches */
    fglobUsage(fh);

    /* partitioning switches */
    filterUsage(fh);
    tupleUsage(fh);

    /* switches from plug-ins */
    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibOptionsUsage(app_dynlib_list[j], fh);
    }
    if (dlISP) {
        dynlibOptionsUsage(dlISP, fh);
    }
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
void appSetup(int argc, char **argv)
{
    int j;
    int input_count;
    int output_count;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char*)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* register the application */
    /* skAppRegister(argv[0]);  -- we do this in main() */
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize variables */
    memset(&dest_type, 0, sizeof(dest_type));

    /* load fglob module */
    if (fglobSetup()) {
        skAppPrintErr("Unable to setup fglob module");
        exit(EXIT_FAILURE);
    }

    /* load filter module */
    if (filterSetup()) {
        skAppPrintErr("Unable to setup filter module");
        exit(EXIT_FAILURE);
    }

    /* load tuple module */
    if (tupleSetup()) {
        skAppPrintErr("Unable to setup tuple module");
        exit(EXIT_FAILURE);
    }

    /* register the options */
    if (skOptionsRegister(appOptions, &appOptionsHandler, NULL)
        || skOptionsNotesRegister(NULL)
        || sksiteCompmethodOptionsRegister(&comp_method))
    {
        skAppPrintErr("Unable to register options");
        exit(EXIT_FAILURE);
    }


    /* specify function plug-ins should use for opening files */
    dynlibSetOpenInputFunction(&filterOpenInputData);

    /* add "hard-coded" dynamic libraries */
    for (j = 0; app_dynlib_names[j]; ++j) {
        appAddDynlib(app_dynlib_names[j], 0);
    }

#ifdef SK_RWFILTER_THREADED
    {
        /* check the thread count envar */
        char *env;
        uint32_t tc;

        env = getenv(RWFILTER_THREADS_ENVAR);
        if (env && env[0]) {
            if (skStringParseUint32(&tc, env, 0, 0) == 0) {
                thread_count = tc;
            } else {
                thread_count = 1;
            }
        }
    }
#endif  /* SK_RWFILTER_THREADED */

    /* parse options */
    arg_index = skOptionsParse(argc, argv);
    assert(arg_index <= argc);
    if (arg_index < 0) {
        skAppUsage();
    }

    /* Check that there is one and only one source of input to process */
    input_count = filterCheckInputs(argc);
    if (input_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (input_count > 1) {
        skAppPrintErr("Multiple input sources were specified\n"
                      "\tInput must come from only one of --input-pipe,"
                      " --xargs, file names on\n"
                      "\tthe command line,"
                      " or a combination of the file selection switches");
        skAppUsage();
    } else if (input_count == 0) {
        skAppPrintErr("No input was specified.\n"
                      "\tNo file selection switches were given,"
                      " neither --input-pipe nor --xargs\n"
                      "\twas specified, and no files are present on the"
                      " command line");
        skAppUsage();
    }

    /* Check whether we have a filtering rule--either built in or from
     * a dynamic library.  If we don't, complain unless the --all-dest
     * switch was given */
    checker_count = filterSetCheckers();
    if ((checker_count <= 0) && (dest_type[DEST_ALL].dest_list == NULL)) {
        skAppPrintErr("No filtering rules given");
        skAppUsage();
    }

    /* check that the user asked for some output */
    output_count = filterCheckOutputs();
    if (output_count < 0) {
        /* fatal error. msg already printed */
        exit(EXIT_FAILURE);
    } else if (output_count == 0) {
        skAppPrintErr("No output(s) specified. Abort");
        skAppUsage();
    }


    /* open the output streams, unless this is a "dry-run" */
    if (NULL == dryrun_fp) {
        if (filterOpenOutputs()) {
            /* fatal error. msg already printed */
            exit(EXIT_FAILURE);
        }
    }

    /* Try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names.  If fglob
     * is active, it will require the configuration file. */
    sksiteConfigure(0);

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("Unable to register appTeardown() with atexit()");
        appTeardown();
        exit(EXIT_FAILURE);
    }

    return;                     /* OK */
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    This function is passed to skOptionsRegister(); it will be called
 *    by skOptionsParse() for each user-specified switch that the
 *    application has registered; it should handle the switch as
 *    required---typically by setting global variables---and return 1
 *    if the switch processing failed or 0 if it succeeded.  Returning
 *    a non-zero from from the handler causes skOptionsParse() to return
 *    a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    destination_t *dest;
    destination_t **end;
    dest_type_t *d_type;
    int dest_id;
    int rv;

    switch ((appOptionsEnum)opt_index) {
      case OPT_PASS_DEST:
      case OPT_FAIL_DEST:
      case OPT_ALL_DEST:
        /* an output stream */
        dest_id = opt_index - OPT_PASS_DEST;
        assert(dest_id >= 0 && dest_id < DESTINATION_TYPES);

        dest = calloc(1, sizeof(destination_t));
        if (dest == NULL) {
            skAppPrintErr("Out of memory");
            return 1;
        }
        if ((rv = skStreamCreate(&dest->ios, SK_IO_WRITE,SK_CONTENT_SILK_FLOW))
            || (rv = skStreamBind(dest->ios, opt_arg)))
        {
            skStreamPrintLastErr(dest->ios, rv, &skAppPrintErr);
            skStreamDestroy(&dest->ios);
            free(dest);
            return 1;
        }

        d_type = &dest_type[dest_id];
        ++d_type->count;
        end = &d_type->dest_list;
        while (*end != NULL) {
            end = &((*end)->next);
        }
        *end = dest;
        break;

#if SK_RWFILTER_THREADED
      case OPT_THREADS:
        rv = skStringParseUint32(&thread_count, opt_arg, 1, 0);
        if (rv) {
            skAppPrintErr("Invalid %s '%s': %s",
                          appOptions[opt_index].name, opt_arg,
                          skStringParseStrerror(rv));
            return 1;
        }
        break;
#endif  /* SK_RWFILTER_THREADED */

      case OPT_INPUT_PIPE:
        /*
         *  input-pipe. Delay check for multiple inputs sources till all
         *  options have been parsed
         */
        if (input_pipe) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        input_pipe = opt_arg;
        break;

      case OPT_DYNLIB:
        if (dlISP) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        dlISP = dynlibCreate(DYNLIB_SHAR_FILTER);
        if (dynlibLoad(dlISP, opt_arg)) {
            skAppPrintErr("Fatal error loading library '%s':\n  %s",
                          opt_arg, dynlibLastError(dlISP));
            return 1;
        }
        break;

      case OPT_DRY_RUN:
        dryrun_fp = DRY_RUN_FH;
        break;

      case OPT_PRINT_FILE:
        filenames_fp = PRINT_FILENAMES_FH;
        break;

      case OPT_PRINT_VOLUME:
        print_volume_stats = 1;
        /* FALLTHROUGH */
      case OPT_PRINT_STAT:
        if (print_stat) {
            skAppPrintErr("May only specified one of --%s or --%s",
                          appOptions[OPT_PRINT_STAT].name,
                          appOptions[OPT_PRINT_VOLUME].name);
            return 1;
        }
        if ((rv = skStreamCreate(&print_stat, SK_IO_WRITE, SK_CONTENT_TEXT))
            || (rv = skStreamBind(print_stat, (opt_arg ? opt_arg : "stderr"))))
        {
            skStreamPrintLastErr(print_stat, rv, &skAppPrintErr);
            skStreamDestroy(&print_stat);
            skAppPrintErr("Invalid %s '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case OPT_MAX_PASS_RECORDS:
        rv = skStringParseUint64(&(dest_type[DEST_PASS].max_records),
                                 opt_arg, 0, 0);
        if (rv) {
            skAppPrintErr("Invalid %s '%s': %s",
                          appOptions[opt_index].name, opt_arg,
                          skStringParseStrerror(rv));
            return 1;
        }
        break;

      case OPT_XARGS:
        /* a file containing a list of file names */
        if (xargs) {
            skAppPrintErr("Only one --%s option is allowed",
                          appOptions[opt_index].name);
            return 1;
        }
        if ((rv = skStreamCreate(&xargs, SK_IO_READ, SK_CONTENT_TEXT))
            || (rv = skStreamBind(xargs, (opt_arg ? opt_arg : "stdin"))))
        {
            skStreamPrintLastErr(xargs, rv, &skAppPrintErr);
            skStreamDestroy(&xargs);
            skAppPrintErr("Invalid %s '%s'",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;
    } /* switch */

    return 0;                     /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
void appTeardown(void)
{
    static uint8_t teardownFlag = 0;
    int j;
    int rv;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    dynlibTeardown(dlISP);

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibTeardown(app_dynlib_list[j]);
        app_dynlib_list[j] = NULL;
    }
    app_dynlib_count = 0;

    tupleTeardown();
    filterTeardown();
    skOptionsNotesTeardown();
    fglobTeardown();

    closeAllDests();

    /* close stats */
    if (print_stat) {
        rv = skStreamClose(print_stat);
        switch (rv) {
          case SKSTREAM_OK:
          case SKSTREAM_ERR_NOT_OPEN:
          case SKSTREAM_ERR_CLOSED:
            break;
          default:
            skStreamPrintLastErr(print_stat, rv, &skAppPrintErr);
            skAppPrintErr("Error closing --print-stat stream '%s'",
                          skStreamGetPathname(print_stat));
            break;
        }
        skStreamDestroy(&print_stat);
    }

    /* close xargs stream */
    if (xargs) {
        skStreamDestroy(&xargs);
    }


    skAppUnregister();
}


/*
 *  appAddDynlib(dlpath, warnOnError);
 *
 *    Load the dynamic library at 'dlpath', store its information in
 *    the global 'app_dynlib_list', and increment the global
 *    'app_dynlib_count'.  Return 0 on success.  Return 1 if the
 *    library cannot be loaded or if too many dynamic libraries have
 *    been loaded.
 */
static int appAddDynlib(const char *dlpath, int warnOnError)
{
    dynlibInfoStruct_t *dl;

    if (app_dynlib_count == APP_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed",
                      APP_MAX_DYNLIBS);
        return 1;
    }

    dl = dynlibCreate(DYNLIB_SHAR_FILTER);
    if (dynlibLoad(dl, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (warnOnError) {
            skAppPrintErr("Error loading dynamic library '%s': %s",
                          dlpath, dynlibLastError(dl));
        }
        dynlibTeardown(dl);
        return 1;
    }

    app_dynlib_list[app_dynlib_count] = dl;
    ++app_dynlib_count;

    return 0;
}


/*
 *  stream_count = filterCheckInputs(argc)
 *
 *    Do basic checks for the input, and return the number of inputs
 *    selected from among: command-line files, fglob options, piped
 *    input, or xargs.  Returns -1 on error, such as trying to read
 *    binary records from a TTY.
 */
static int filterCheckInputs(int argc)
{
    int count = 0;
    int rv;

    /* Check for file names on command line */
    if (arg_index < argc) {
        ++count;
    }

    /* the input-pipe must be 'stdin' or it must be an existing FIFO.
     * If 'stdin', stdin must not be a TTY, since we expect binary. */
    if (input_pipe) {
        ++count;
        if (strcmp(input_pipe, "stdin") == 0) {
            if (FILEIsATty(stdin)) {
                skAppPrintErr("The input-pipe 'stdin' is connected"
                              " to a terminal. Abort.");
                return -1;
            }
        } else if ( !skFileExists(input_pipe)) {
            skAppPrintErr("The input-pipe '%s' does not exist",
                          input_pipe);
            return -1;
        } else if ( !isFIFO(input_pipe)) {
            skAppPrintErr("The input-pipe '%s' is not a pipe",
                          input_pipe);
            return -1;
        }
    }

    /* check if an --xargs value was given */
    if (xargs) {
        ++count;
    }

    if (count == 1) {
        /*
         * Let any traditional fglob options work as filters.
         */
        if (filterGetFGlobFilters()) {
            return -1;
        }
    }

    /* Check if fglob args were given */
    rv = fglobValid();
    if (rv == -1) {
        /* error with fglob options */
        return -1;
    }
    if (rv) {
        ++count;
    }

    return count;
}


/*
 *  stream_count = filterCheckOutputs()
 *
 *    Count the number of output streams, and do basic checks for the
 *    outputs, such as making certain only one stream uses stdout.
 *    Note that for this count, a WILLPROCESS dynamic library (the
 *    global dlISP) is considered an outputs.
 *
 *    Returns the number of output streams requested, or -1 on error.
 */
static int filterCheckOutputs(void)
{
    destination_t *dest;
    int count = 0;
    int stdout_used = 0;
    int dest_id;

    /* If the dynamic library is taking over all processing, warn if
     * --pass or --fail was also given. */
    if (dlISP && DYNLIB_WILLPROCESS == dynlibGetStatus(dlISP)) {
        ++count;
        if (dest_type[DEST_PASS].count) {
            skAppPrintErr("Warning: dynamic library overriding "
                          "--pass-destinations");
            closeOutputDests(DEST_PASS, 1);
        }
        if (dest_type[DEST_FAIL].count) {
            skAppPrintErr("Warning: dynamic library overriding "
                          "--fail-destinations");
            closeOutputDests(DEST_FAIL, 1);
        }
        if (print_stat) {
            skAppPrintErr("Warning: dynamic library overriding "
                          "--print-statistics");
            skStreamDestroy(&print_stat);
        }
    }

    /* Basic checks: Only allow one output stream to use 'stdout'.
     * Make certain no one uses 'stderr'. */
    for (dest_id = 0; dest_id < DESTINATION_TYPES; ++dest_id) {
        for (dest = dest_type[dest_id].dest_list;
             dest != NULL;
             dest = dest->next)
        {
            ++count;
            if (0 == strcmp("stderr", skStreamGetPathname(dest->ios))){
                skAppPrintErr("Error: stderr is not a valid output device");
                return -1;
            } else if (0 == strcmp("stdout", skStreamGetPathname(dest->ios))) {
                if (stdout_used) {
                    skAppPrintErr("Error: stdout is already allocated");
                    return -1;
                }
                stdout_used = 1;
            }
        }
    }

    /* Check the STATISTICS stream: increment the output count and
     * check whether stdout is already used if the stats are going to
     * stdout. */
    if (print_stat) {
        ++count;
        if (strcmp(skStreamGetPathname(print_stat), "stdout") == 0) {
            if (stdout_used) {
                skAppPrintErr("Error: stdout is already allocated");
                return -1;
            }
        }
    }

    return count;
}


/*
 *  status = filterOpenOutputs()
 *
 *    Open all output streams.  Return 0 on success, or -1 on failure.
 */
static int filterOpenOutputs(void)
{
    destination_t *dest = NULL;
    int dest_id;
    int rv = SKSTREAM_OK;

    /* open the STATISTICS stream */
    if (print_stat) {
        rv = skStreamOpen(print_stat);
        if (rv) {
            skStreamPrintLastErr(print_stat, rv, &skAppPrintErr);
            return -1;
        }
    }


    /* Open all the SiLK Flow output streams */
    for (dest_id = 0; dest_id < DESTINATION_TYPES; ++dest_id) {

        for (dest = dest_type[dest_id].dest_list;
             dest != NULL;
             dest = dest->next)
        {
            /* set compression level */
            rv = skHeaderSetCompressionMethod(skStreamGetSilkHeader(dest->ios),
                                              comp_method);
            if (rv) {
                goto DEST_ERROR;
            }
            rv = skStreamOpen(dest->ios);
            if (rv) {
                goto DEST_ERROR;
            }
        }
    }

    return 0;

  DEST_ERROR:
    skStreamPrintLastErr(dest->ios, rv, &skAppPrintErr);
    closeAllDests();
    return -1;
}


/*
 *  count = filterSetCheckers();
 *
 *    Set the array of function pointers to the pass/fail checking
 *    routines, and return the number of pointers that were set.  If a
 *    check-routine is a plug-in, call the plug-in's initialize()
 *    routine.  If the initialize() routine fails, return -1.  A
 *    return code of 0 means no filtering rules were specified.
 */
static int filterSetCheckers(void)
{
    int count = 0;
    int rv;
    int j;

    if (filterGetCheckCount() > 0) {
        checker[count] = (int (*)(rwRec*))filterCheck;
        ++count;
    }

    rv = tupleGetCheckCount();
    if (rv == -1) {
        return -1;
    }
    if (rv) {
        checker[count] = (int (*)(rwRec*))tupleCheck;
        ++count;
    }

    for (j = 0; j < app_dynlib_count; ++j) {
        if (dynlibCheckActive(app_dynlib_list[j])) {
            if (dynlibInitialize(app_dynlib_list[j])) {
                return -1;
            }
            checker[count] = dynlibGetRWProcessor(app_dynlib_list[j]);
            ++count;
#if SK_RWFILTER_THREADED
            /* do not use threading when the plug-in doesn't support
             * it */
            if ((thread_count > 1)
                && (dynlibSupportsThreads(app_dynlib_list[j]) == 0))
            {
                thread_count = 1;
            }
#endif  /* SK_RWFILTER_THREADED */
        }
    }

    if (dynlibCheckLoaded(dlISP)) {
        if (dynlibInitialize(dlISP)) {
            return -1;
        }
#if SK_RWFILTER_THREADED
        /* do not use threading when the plug-in doesn't support it */
        if ((thread_count > 1)
            && (dynlibSupportsThreads(dlISP) == 0))
        {
            thread_count = 1;
        }
#endif  /* SK_RWFILTER_THREADED */
        checker[count] = dynlibGetRWProcessor(dlISP);
        ++count;
    }

    return count;
}


/*
 *  status = filterOpenInputData(&stream, content_type, filename);
 *
 *    This is the function that plug-ins should use for opening any
 *    input files they use.  The function opens 'filename', a file
 *    having the specified 'content_type', and fills 'stream' with a
 *    handle to the file.  Returns 0 if the file was successfully
 *    opened, and -1 if the file could not be opened.
 */
int filterOpenInputData(
    skstream_t    **stream,
    skcontent_t     content_type,
    const char     *filename)
{
    skstream_t *s;
    int rv;


    if ((rv = skStreamCreate(&s, SK_IO_READ, content_type))
        || (rv = skStreamBind(s, filename))
        || (rv = skStreamOpen(s)))
    {
        skStreamPrintLastErr(s, rv, &skAppPrintErr);
        skStreamDestroy(&s);
        return -1;
    }
    *stream = s;
    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
