/*
** Copyright (C) 2004-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

#include "silk.h"

RCSIDENT("$SiLK: stream-cache.c 12525 2008-09-18 15:42:32Z mthomas $");

#include "utils.h"
#include "sklog.h"
#include "rwpack.h"
#include "stream-cache.h"
#include "sksite.h"
#include "redblack.h"


/* MACROS and DATA TYPES */

#ifndef DEBUG
#  define DEBUG 0
#endif

/* Message to print when out of memory */
#define FMT_MEM_FAILURE  "Out of memory at %s:%d", __FILE__, __LINE__
#define FMT_MUTEX_FAILURE \
    "Could not initialize a mutex at %s:%d", __FILE__, __LINE__

#define MUTEX_INIT(x) pthread_mutex_init((x), NULL)
#define MUTEX_DESTROY(x) pthread_mutex_destroy((x))

#if HAVE_PTHREAD_RWLOCK
#  define RWMUTEX pthread_rwlock_t
#  define READ_LOCK(x) pthread_rwlock_rdlock(x)
#  define WRITE_LOCK(x) pthread_rwlock_wrlock(x)
#  define RW_MUTEX_UNLOCK(x) pthread_rwlock_unlock(x)
#  define RW_MUTEX_INIT(x) pthread_rwlock_init((x), NULL)
#  define RW_MUTEX_DESTROY(x) pthread_rwlock_destroy(x)
#  define ASSERT_RW_MUTEX_LOCKED(x) assert(EBUSY==pthread_rwlock_trywrlock(x))
#else
#  define RWMUTEX pthread_mutex_t
#  define READ_LOCK MUTEX_LOCK
#  define WRITE_LOCK MUTEX_LOCK
#  define RW_MUTEX_UNLOCK MUTEX_UNLOCK
#  define RW_MUTEX_INIT MUTEX_INIT
#  define RW_MUTEX_DESTROY MUTEX_DESTROY
#  define ASSERT_RW_MUTEX_LOCKED(x) assert(EBUSY == pthread_mutex_trylock(x))
#endif


/*
 *  The stream_cache_t contains an array of cache_entry_t objects.
 *  The array is allocated with 'max_size' entries when the cache is
 *  created.  The number of valid entries in the array is specified by
 *  'size'.  The cache also contains a redblack tree that is used to
 *  index the entries.  Finally, there is a mutex on the cache; the
 *  mutex will be a pthread_rwlock_t mutex if read/write locks are
 *  supported on this system.
 */
struct stream_cache_st {
    /* Array of cache entries */
    cache_entry_t      *entries;
    /* the redblack tree used for searching */
    struct rbtree      *rbtree;
    /* function called by skCacheLookupOrOpenAdd() to open a file */
    cache_open_fn_t     open_callback;
    /* current number of valid entries */
    int                 size;
    /* maximum number of valid entries */
    int                 max_size;
    /* mutex for the cache */
    RWMUTEX             mutex;
};


/* FUNCTION DECLARATIONS */

static cache_entry_t *cacheEntryAdd(
    stream_cache_t     *cache,
    skstream_t         *rwios,
    const cache_key_t  *key);
static int cacheEntryCompare(
    const void *entry1_v,
    const void *entry2_v,
    const void UNUSED(*config));
static void cacheEntryDestroyFile(
    stream_cache_t *cache,
    cache_entry_t  *entry);
static void cacheEntryLogRecordCount(
    cache_entry_t  *entry);
static cache_entry_t *cacheEntryLookup(
    stream_cache_t     *cache,
    const cache_key_t  *key);


/* FUNCTION DEFINITIONS */

/*
 *  entry = cacheEntryAdd(cache, stream, key);
 *
 *    Add 'stream' to the stream cache 'cache' using 'key' as the key.
 *    The cache should be locked for writing.  The entry's
 *    last_accessed time is set to the current time and the entry is
 *    returned in a locked state.
 */
static cache_entry_t *cacheEntryAdd(
    stream_cache_t     *cache,
    skstream_t         *rwios,
    const cache_key_t  *key)
{
    cache_entry_t *entry;
    const void *rv;

    ASSERT_RW_MUTEX_LOCKED(&cache->mutex);
    assert(rwios);

#if DEBUG
    DEBUGMSG("Adding new entry to cache with %d/%d entries",
             cache->size, cache->max_size);
#endif

    if (cache->size < cache->max_size) {
        /* We're not to the max size yet, so use the next entry in the
         * array */
        entry = &cache->entries[cache->size];
        if (MUTEX_INIT(&entry->mutex)) {
            CRITMSG(FMT_MUTEX_FAILURE);
            return NULL;
        }
        ++cache->size;
    } else {
        /* The cache is full: flush, close and free the least recently
         * used stream */
        int i;
        time_t min;

        min = cache->entries[0].last_accessed;
        entry = &cache->entries[0];

        for (i = 1; i < cache->size; i++) {
            if (cache->entries[i].last_accessed < min) {
                min = cache->entries[i].last_accessed;
                entry = &cache->entries[i];
            }
        }
        cacheEntryDestroyFile(cache, entry);
    }

    /* fill the new entry */
    entry->key.time_stamp = key->time_stamp;
    entry->key.sensor_id = key->sensor_id;
    entry->key.flowtype_id = key->flowtype_id;
    entry->stream = rwios;
    entry->rec_count = skStreamGetRecordCount(rwios);
    entry->last_accessed = time(NULL);

    /* add the entry to the redblack tree */
    rv = rbsearch(entry, cache->rbtree);
    if (rv != entry) {
        if (rv == NULL) {
            CRITMSG("Out of memory at %s:%d", __FILE__, __LINE__);
            exit(EXIT_FAILURE);
        }
        CRITMSG(("Duplicate entries in stream cache "
                 "for time=%" PRId64 " sensor=%d flowtype=%d"),
                key->time_stamp, key->sensor_id, key->flowtype_id);
        abort();
    }


    /* lock the entry */
    MUTEX_LOCK(&entry->mutex);

    return entry;
}


/*
 *  direction = cacheEntryCompare(a, b, config);
 *
 *    The comparison function used by the redblack tree.
 */
static int cacheEntryCompare(
    const void *entry1_v,
    const void *entry2_v,
    const void UNUSED(*config))
{
    cache_key_t key1 = ((cache_entry_t*)entry1_v)->key;
    cache_key_t key2 = ((cache_entry_t*)entry2_v)->key;

    if (key1.sensor_id < key2.sensor_id) {
        return -1;
    }
    if (key1.sensor_id > key2.sensor_id) {
        return 1;
    }
    if (key1.flowtype_id < key2.flowtype_id) {
        return -1;
    }
    if (key1.flowtype_id > key2.flowtype_id) {
        return 1;
    }
    if (key1.time_stamp < key2.time_stamp) {
        return -1;
    }
    if (key1.time_stamp > key2.time_stamp) {
        return 1;
    }
    return 0;
}


/*
 *  cacheEntryDestroyFile(entry);
 *
 *    Close the stream that 'entry' wraps, destroy the stream, and
 *    remove the entry from the redblack tree.  In addition, log the
 *    number of records written.  Does not destroy the entry's mutex.
 *
 *    This function expects the caller to have a write lock on the
 *    cache.
 */
static void cacheEntryDestroyFile(
    stream_cache_t *cache,
    cache_entry_t  *entry)
{
    int rv;

    ASSERT_RW_MUTEX_LOCKED(&cache->mutex);

    MUTEX_LOCK(&entry->mutex);

#if DEBUG
    DEBUGMSG("Stream cache closing file %s",
             skStreamGetPathname(entry->stream));
#endif

    cacheEntryLogRecordCount(entry);
    rv = skStreamClose(entry->stream);
    if (rv) {
        skStreamPrintLastErr(entry->stream, rv, &NOTICEMSG);
    }
    skStreamDestroy(&entry->stream);
    rbdelete(entry, cache->rbtree);

    MUTEX_UNLOCK(&entry->mutex);
}


/*
 *  cacheEntryLogRecordCount(entry);
 *
 *    Write a message to the log giving the name of the file that
 *    'entry' wraps and the number of records written to that file
 *    since it was opened or last flushed.
 *
 *    The 'entry' will be updated with the new record count.
 */
static void cacheEntryLogRecordCount(
    cache_entry_t  *entry)
{
    uint64_t new_count;

    new_count = skStreamGetRecordCount(entry->stream);

    if (entry->rec_count == new_count) {
        return;
    }
    assert(entry->rec_count < new_count);

    INFOMSG(("%s: %" PRIu64 " recs"),
            skStreamGetPathname(entry->stream),(new_count - entry->rec_count));
    entry->rec_count = new_count;
}


/*
 *  entry = skCacheLookup(cache, key);
 *
 *    Return the stream entry for the specified key.  Return NULL if
 *    no stream entry for the specified values is found.  The cache
 *    should be locked for reading or writing.  The entry's
 *    last_accessed time is updated and the entry is returned in a
 *    locked state.
 */
static cache_entry_t *cacheEntryLookup(
    stream_cache_t     *cache,
    const cache_key_t  *key)
{
    cache_entry_t *entry;
    cache_entry_t search_key;

    ASSERT_RW_MUTEX_LOCKED(&cache->mutex);

    /* fill the search key */
    search_key.key.time_stamp = key->time_stamp;
    search_key.key.sensor_id = key->sensor_id;
    search_key.key.flowtype_id = key->flowtype_id;

    /* try to find the entry */
    entry = (cache_entry_t*)rbfind(&search_key, cache->rbtree);
#if DEBUG
    {
        char tstamp[SKTIMESTAMP_STRLEN];
        char sensor[SK_MAX_STRLEN_SENSOR+1];
        char flowtype[SK_MAX_STRLEN_FLOWTYPE+1];

        sktimestamp_r(tstamp, key->time_stamp, SKTIMESTAMP_NOMSEC),
        sksiteSensorGetName(sensor, sizeof(sensor), key->sensor_id);
        sksiteFlowtypeGetName(flowtype, sizeof(flowtype), key->flowtype_id);

        DEBUGMSG("Cache %s for stream %s %s %s",
                 ((entry == NULL) ? "miss" : "hit"),
                 tstamp, sensor, flowtype);
    }
#endif /* DEBUG */

    /* found it, lock it and update its last_accessed timestamp */
    if (entry) {
        MUTEX_LOCK(&entry->mutex);
        entry->last_accessed = time(NULL);
    }

    return entry;
}


/* add an entry to the cache.  return entry in locked state. */
cache_entry_t *skCacheAdd(
    stream_cache_t     *cache,
    skstream_t         *rwios,
    const cache_key_t  *key)
{
    cache_entry_t *entry;

    WRITE_LOCK(&cache->mutex);

    entry = cacheEntryAdd(cache, rwios, key);

    RW_MUTEX_UNLOCK(&cache->mutex);

    return entry;
}


/* lock cache, then close and destroy all streams.  unlock cache. */
void skCacheCloseAll(stream_cache_t *cache)
{
    if (NULL == cache) {
        return;
    }

    skCacheLockAndCloseAll(cache);

    RW_MUTEX_UNLOCK(&cache->mutex);
}


/* create a cache with the specified size and open callback function */
stream_cache_t *skCacheCreate(
    int                 max_size,
    cache_open_fn_t     open_fn)
{
    stream_cache_t *cache = NULL;

    /* verify input */
    if (max_size < STREAM_CACHE_MINIMUM_SIZE) {
        CRITMSG("Illegal maximum size (%d) for stream cache; use value > 1",
                max_size);
        return NULL;
    }

    cache = calloc(1, sizeof(stream_cache_t));
    if (cache == NULL) {
        CRITMSG(FMT_MEM_FAILURE);
        return NULL;
    }

    if (RW_MUTEX_INIT(&cache->mutex)) {
        CRITMSG(FMT_MUTEX_FAILURE);
        free(cache);
        return NULL;
    }

    cache->entries = (cache_entry_t *)calloc(max_size, sizeof(cache_entry_t));
    if (cache->entries == NULL) {
        CRITMSG(FMT_MEM_FAILURE);
        RW_MUTEX_DESTROY(&cache->mutex);
        free(cache);
        return NULL;
    }

    cache->rbtree = rbinit(&cacheEntryCompare, NULL);
    if (cache->rbtree == NULL) {
        CRITMSG(FMT_MEM_FAILURE);
        RW_MUTEX_DESTROY(&cache->mutex);
        free(cache->entries);
        free(cache);
        return NULL;
    }

    cache->max_size = max_size;
    cache->open_callback = open_fn;

    return cache;
}


/* close all streams, destroy them, and destroy the cache */
void skCacheDestroy(stream_cache_t *cache)
{
    if (NULL == cache) {
        INFOMSG("Tried to destroy unitialized stream cache.");
        return;
    }

#if DEBUG
    DEBUGMSG("Destroying cache with %d entries", cache->size);
#endif

    /* close any open files */
    skCacheLockAndCloseAll(cache);

    /* destroy the redblack tree */
    rbdestroy(cache->rbtree);

    /* Destroy the entries array */
    free(cache->entries);

    RW_MUTEX_UNLOCK(&cache->mutex);
    RW_MUTEX_DESTROY(&cache->mutex);

    /* Free the structure itself */
    free(cache);
}


/* flush all streams in the cache */
void skCacheFlush(stream_cache_t *cache)
{
    int i;
    int rv;

    if (NULL == cache) {
        return;
    }

    READ_LOCK(&cache->mutex);

#if DEBUG
    DEBUGMSG("Flushing cache with %d entries", cache->size);
#endif

    for (i = 0; i < cache->size; i++) {
        cache_entry_t *entry = &cache->entries[i];

        MUTEX_LOCK(&entry->mutex);

        rv = skStreamFlush(entry->stream);
        if (rv) {
            skStreamPrintLastErr(entry->stream, rv, &NOTICEMSG);
        }
        cacheEntryLogRecordCount(entry);

        MUTEX_UNLOCK(&entry->mutex);
    }

    RW_MUTEX_UNLOCK(&cache->mutex);
}


/* lock cache, then close and destroy all streams. do not unlock cache */
void skCacheLockAndCloseAll(
    stream_cache_t *cache)
{
    int i;

    if (NULL == cache) {
        return;
    }

    WRITE_LOCK(&cache->mutex);

#if DEBUG
    DEBUGMSG("Closing files in cache with %d entries", cache->size);
#endif

    for (i = 0; i < cache->size; i++) {
        cache_entry_t *entry = &cache->entries[i];
        cacheEntryDestroyFile(cache, entry);
        MUTEX_DESTROY(&entry->mutex);
    }

    cache->size = 0;
}


/* find an entry in the cache.  return entry in locked state. */
cache_entry_t *skCacheLookup(
    stream_cache_t     *cache,
    const cache_key_t  *key)
{
    cache_entry_t *entry;

    READ_LOCK(&cache->mutex);

    entry = cacheEntryLookup(cache, key);

    RW_MUTEX_UNLOCK(&cache->mutex);

    return entry;
}


/* find an entry in the cache.  if not present, use the open-callback
 * function to open/create the stream and then add it. */
cache_entry_t *skCacheLookupOrOpenAdd(
    stream_cache_t     *cache,
    const cache_key_t  *key,
    void               *caller_data)
{
    cache_entry_t *entry;
    skstream_t *rwio;

    /* do a standard lookup */
    READ_LOCK(&cache->mutex);
    entry = cacheEntryLookup(cache, key);

    /* found it; we can return */
    if (entry) {
        goto END;
    }

#if HAVE_PTHREAD_RWLOCK
    /* need to add the stream.  Get a write lock, but first we must
     * release the read lock.  No need to change the type of lock and
     * re-do the search if we aren't using read/write locks. */
    RW_MUTEX_UNLOCK(&cache->mutex);
    WRITE_LOCK(&cache->mutex);

    /* search for the entry again, in case it was added between
     * releasing the read lock and getting the write lock */
    entry = cacheEntryLookup(cache, key);
    if (entry) {
        /* found it.  we can return */
        goto END;
    }
#endif  /* HAVE_PTHREAD_RWLOCK */

    /* use the callback to open the file */
    rwio = cache->open_callback(key, caller_data);
    if (NULL == rwio) {
        goto END;
    }

    /* add the newly opened file to the cache */
    entry = cacheEntryAdd(cache, rwio, key);
    if (NULL == entry) {
        skStreamDestroy(&rwio);
    }

  END:
    RW_MUTEX_UNLOCK(&cache->mutex);
    return entry;
}


/* unlocks a cache locked by skCacheLockAndCloseAll(). */
void skCacheUnlock(
    stream_cache_t *cache)
{
    assert(cache);

    RW_MUTEX_UNLOCK(&cache->mutex);
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
