/*
** Copyright (C) 2001-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwuniq.c
**
**  Implementation of the rwuniq application.
**
**  rwuniq reads SiLK flow records---from files listed on the command
**  line or from the standard input when no filenames are given---and
**  bins those flows by a key composed of user-selected fields of an
**  rwRec, or by fields generated from a plug-in.  For each bin, a
**  user-selected combination of bytes, packets, flows, earliest
**  start-time, latest end-time, distinct sIPs, and/or distinct dIPs
**  may be computed.
**
**  Once the input is read, the keys fields and computed values are
**  printed for each bin that meets the user-specified minimum and
**  maximum.
**
**  Normally, rwuniq uses the hashlib hash table to store the
**  key-volume pairs for each bin.  If this hash table runs out of
**  memory, the contents of the table are sorted and then saved to
**  disk in a temporary file.  More records are then read into a fresh
**  hash table.  The process repeats until all records are read or the
**  maximum number of temp files is reached.  The on-disk files are
**  then merged to produce the final output.
**
**  When the --presorted-input switch is given, rwuniq assumes rwsort
**  has been used to sort the data with the same --fields value that
**  rwuniq is using.  In this case, the hash table is not used.
**  Instead, rwuniq just watches for the key to change, and prints the
**  key-volume when it does.
**
**  For the --presorted-input case or when more than one distinct IP
**  count is requested for the unsorted case, an IPSet is used to keep
**  track of the IPs we have seen.  Since IPSets do not yet support
**  IPv6, this limits rwuniq's ability when IPv6 is active.  Also,
**  these IPSets can exhaust the ram, which would lead to an incorrect
**  count of IPs.  Could consider using a hashlib instead of an IPSet
**  for the values to get around the IPv6 issue.
**
*/

#include "silk.h"

RCSIDENT("$SiLK: rwuniq.c 13074 2008-11-20 20:28:45Z mthomas $");

#include "rwuniq.h"
#include "rwuniqtemp.h"
#include "hashlib.h"


/* TYPEDEFS AND DEFINES */

/* Initializer for hash table creation */
#define HASH_INITIAL_SIZE 500000

/* The summation values are stored in an array of bytes; the offsets
 * into this array are stored in the variable value_fields[].  The
 * following macros access the memory at these offsets or copy the
 * values from/to the byte-array and a variable. */
#define VAL_FIELD_GET_ADDR32(val_ptr, i)                        \
    ((uint32_t*)((val_ptr) + value_fields[(i)].vf_offset))

#define VAL_FIELD_MEMGET(val_ptr, i, out_val)                     \
    memcpy((out_val), ((val_ptr) + value_fields[(i)].vf_offset),  \
           value_fields[(i)].vf_byte_width)

#define VAL_FIELD_MEMSET(val_ptr, i, in_val)                      \
    memcpy(((val_ptr) + value_fields[(i)].vf_offset), (in_val),   \
           value_fields[(i)].vf_byte_width)

/*
 * Macro to take the key from the hash table 'hashkey' and copy the
 * distinct field at position 'dnum' from it and into 'ip_ptr'
 */
#if  RWUNIQ_USE_MEMCPY
#define GET_DISTINCT_IP(hashkey, dnum, ip_ptr)                  \
    memcpy((ip_ptr),                                            \
           (hashkey) + distinct_fields[(dnum)].df_offset,       \
           distinct_fields[(dnum)].df_byte_width)
#else
#define GET_DISTINCT_IP(hashkey, dnum, ip_ptr)                  \
    (*(ip_ptr)) = *(uint32_t*)((hashkey) + distinct_fields[(dnum)].df_offset)
#endif

/*
 *  Print a message (one time) if we fail to add an IP to a IPset,
 *  meaning that the distinct counts are incorrect.  Uses the global
 *  variable 'distinct_err'.
 */
#define MEMCHECK_IPSET_ADD(ipset, ipaddr, distinct_count)       \
    if (skIPTreeCheckAddress((ipset), (ipaddr))) {              \
        /* no need to do anything */                            \
    } else if (skIPTreeAddAddress((ipset), (ipaddr)) == 0) {    \
        ++(distinct_count);                                     \
    } else if (!distinct_err) {                                 \
        skAppPrintErr("Out of memory counting distinct IPs."    \
                      " Counts may be incorrect");              \
        distinct_err = 1;                                       \
    }



#define KEY_FIELD_PTR(keyarray, idx)            \
    ((keyarray) + key_fields[(idx)].kf_offset)

/*
 *  These macros either get a value from a record, or set a value on
 *  the record.  The value will either be uint16 or a uint32, and it
 *  may be in network or native byte order.
 */
#define REC_TO_KEY_NATIVE_08(func, rec, keyarray, idx)  \
    { *(KEY_FIELD_PTR(keyarray, idx)) = func(rec); }

#define KEY_TO_REC_NATIVE_08(func, rec, keyarray, idx)  \
    func((rec), *(KEY_FIELD_PTR(keyarray, idx)))

#define REC_TO_KEY_NETWORK_08  REC_TO_KEY_NATIVE_08

#define KEY_TO_REC_NETWORK_08  KEY_TO_REC_NATIVE_08

#if  !RWUNIQ_USE_MEMCPY

#define REC_TO_KEY_NATIVE_16(func, rec, keyarray, idx)          \
    { *((uint16_t*)KEY_FIELD_PTR(keyarray, idx)) = func(rec); }

#define REC_TO_KEY_NETWORK_16(func, rec, keyarray, idx)                 \
    { *((uint16_t*)KEY_FIELD_PTR(keyarray, idx)) = htons(func(rec)); }

#define REC_TO_KEY_NATIVE_32(func, rec, keyarray, idx)          \
    { *((uint32_t*)KEY_FIELD_PTR(keyarray, idx)) = func(rec); }

#define REC_TO_KEY_NETWORK_32(func, rec, keyarray, idx)                 \
    { *((uint32_t*)KEY_FIELD_PTR(keyarray, idx)) = htonl(func(rec)); }

#define KEY_TO_REC_NATIVE_16(func, rec, keyarray, idx)          \
    func((rec), *((uint16_t*)KEY_FIELD_PTR(keyarray, idx)))

#define KEY_TO_REC_NETWORK_16(func, rec, keyarray, idx)                 \
    func((rec), htons(*((uint16_t*)KEY_FIELD_PTR(keyarray, idx))))

#define KEY_TO_REC_NATIVE_32(func, rec, keyarray, idx)          \
    func((rec), *((uint32_t*)KEY_FIELD_PTR(keyarray, idx)))

#define KEY_TO_REC_NETWORK_32(func, rec, keyarray, idx)                 \
    func((rec), htonl(*((uint32_t*)KEY_FIELD_PTR(keyarray, idx))))

#else  /* following is when used when RWUNIQ_USE_MEMCPY is true */

#define REC_TO_KEY_NATIVE_16(func, rec, keyarray, idx)                  \
    {                                                                   \
        uint16_t t16 = func(rec);                                       \
        memcpy(KEY_FIELD_PTR(keyarray, idx), &t16, sizeof(t16));        \
    }

#define REC_TO_KEY_NETWORK_16(func, rec, keyarray, idx)                 \
    {                                                                   \
        uint16_t t16 = htons(func(rec));                                \
        memcpy(KEY_FIELD_PTR(keyarray, idx), &t16, sizeof(t16));        \
    }

#define REC_TO_KEY_NATIVE_32(func, rec, keyarray, idx)                  \
    {                                                                   \
        uint32_t t32 = func(rec);                                       \
        memcpy(KEY_FIELD_PTR(keyarray, idx), &t32, sizeof(t32));        \
    }

#define REC_TO_KEY_NETWORK_32(func, rec, keyarray, idx)                 \
    {                                                                   \
        uint32_t t32 = htonl(func(rec));                                \
        memcpy(KEY_FIELD_PTR(keyarray, idx), &t32, sizeof(t32));        \
    }

#define KEY_TO_REC_NATIVE_16(func, rec, keyarray, idx)                  \
    {                                                                   \
        uint16_t t16;                                                   \
        memcpy(&t16, KEY_FIELD_PTR(keyarray, idx), sizeof(t16));        \
        func((rec), t16);                                               \
    }

#define KEY_TO_REC_NETWORK_16(func, rec, keyarray, idx)                 \
    {                                                                   \
        uint16_t t16;                                                   \
        memcpy(&t16, KEY_FIELD_PTR(keyarray, idx), sizeof(t16));        \
        func((rec), htons(t16));                                        \
    }

#define KEY_TO_REC_NATIVE_32(func, rec, keyarray, idx)                  \
    {                                                                   \
        uint32_t t32;                                                   \
        memcpy(&t32, KEY_FIELD_PTR(keyarray, idx), sizeof(t32));        \
        func((rec), t32);                                               \
    }

#define KEY_TO_REC_NETWORK_32(func, rec, keyarray, idx)                 \
    {                                                                   \
        uint32_t t32;                                                   \
        memcpy(&t32, KEY_FIELD_PTR(keyarray, idx), sizeof(t32));        \
        func((rec), htonl(t32));                                        \
    }

#endif  /* RWUNIQ_USE_MEMCPY */


/* EXPORTED VARIABLES */

/* delimiter between output columns */
char g_delimiter = '|';

/* the columns/fields that make up our key for this run */
key_field_t *key_fields = NULL;

/* the number of actual columns/fields that make up the key */
size_t key_num_fields = 0;

/* the total byte count of the key */
size_t key_byte_width = 0;

/* Information about each potential "distinct" field the user can
 * choose to compute and display. */
distinct_field_t distinct_fields[NUM_DISTINCTS] = {
    {"Unique_SIP", NULL, 1, UINT64_MAX, &rwrec_MemGetSIPv4,
     sizeof(uint32_t), 0, 10, DISTINCT_SIP, 0},
    {"Unique_DIP", NULL, 1, UINT64_MAX, &rwrec_MemGetDIPv4,
     sizeof(uint32_t), 0, 10, DISTINCT_DIP, 0}
};

/* Number of distinct fields the user has requested */
size_t distinct_num_fields = 0;

/* Total byte count for all distinct_fields that are being computed */
size_t distinct_byte_width = 0;

/* Information about each potential "value" field the user can choose
 * to compute and display. */
value_field_t value_fields[NUM_VALUES] = {
    {"Bytes",      NULL, 1, UINT64_MAX, sizeof(uint64_t), 0, 20, 0,
     VALUE_BYTES,   1, 0},
    {"Packets",    NULL, 1, UINT64_MAX, sizeof(uint32_t), 0, 10, 1,
     VALUE_PACKETS, 1, 0},
    {"Records",    NULL, 1, UINT64_MAX, sizeof(uint32_t), 0, 10, 2,
     VALUE_FLOWS,   1, 0},
    {"min_sTime",  NULL, 1, UINT64_MAX, sizeof(uint32_t), 0, 19, 3,
     VALUE_STIME,   1, 0},
    {"max_eTime",  NULL, 1, UINT64_MAX, sizeof(uint32_t), 0, 19, 4,
     VALUE_ETIME,   1, 0}
};

/* Number of value fields the user has requested */
size_t value_num_fields = 0;

/* Total byte count for all value_fields that are being computed */
size_t value_byte_width = 0;

/* the byte count of the key passed to the hash table, which is the
 * length of the "key" and any "distinct" fields  */
size_t hash_key_bytes = 0;

/* the byte length of the value used by the hash table, which has a
 * minimum length of 1  */
size_t hash_value_bytes = 0;

/* the byte length of the key-value pair in the hash table, which is
 * comprised of the key and the value */
size_t hash_node_bytes = 0;

/* Array of out_stream_t* */
out_stream_t **outputs;

/* Count of 'outputs' */
size_t output_count;

/* the real output */
uniq_io_t uniq_io;

/* flags set by the user options */
uniq_flags_t uniq_flags;

/* how to print timestamps */
uint32_t time_flags = SKTIMESTAMP_NOMSEC;

/* how to handle IPv6 flows */
sk_ipv6policy_t ipv6_policy = SK_IPV6POLICY_MIX;


/* LOCAL VARIABLES */

/* whether we encounted an error counting distinct IP addresses */
static int distinct_err = 0;

/* number of records read */
static uint64_t record_count = 0;

/* scratch space for a node */
static uint8_t scratch[MAX_HASH_NODE_BYTES];


/* LOCAL FUNCTION PROTOTYPES */

/* I/O Functions */
static void writeColTitles(void);



/* FUNCTION DEFINITIONS */


/*
 *  writeColTitles();
 *
 *    Print the column titles to the global 'uniq_io.outF'.
 */
static void writeColTitles(void)
{
    size_t i;
    char buf[VALUE_OUT_BUF_SIZE];
    char delim[2] = {'\0', '\0'};
    out_stream_t *ostream;

    if (uniq_flags.no_titles) {
        return;
    }

    /* print key titles */
    for (i = 0; i < output_count; ++i) {
        ostream = outputs[i];
        if (ostream->a_stream) {
            rwAsciiPrintTitles(ostream->a_stream);
        } else {
            /* call the uniq() function on the plug-in to get the title */
            assert(ostream->k_field->kf_text_len <= (int)sizeof(buf));
            ostream->k_field->kf_fxn(ostream->k_field->kf_fxn_id, 0, buf,
                                     ostream->k_field->kf_text_len, NULL);
            if (uniq_flags.no_columns) {
                fprintf(uniq_io.outF, "%s%c", buf, g_delimiter);
            } else {
                fprintf(uniq_io.outF, "%*s%c",
                        ostream->k_field->kf_text_len-1,
                        buf, g_delimiter);
            }
        }
    }

    /* print value titles */
    for (i = 0; i < value_num_fields; ++i) {
        if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%s",
                    delim, value_fields[i].vf_title);
        } else {
            fprintf(uniq_io.outF, "%s%*.*s",
                    delim,
                    value_fields[i].vf_text_len,
                    value_fields[i].vf_text_len,
                    value_fields[i].vf_title);
        }
        delim[0] = g_delimiter;
    }

    /* print distinct titles */
    if (distinct_num_fields) {
        for (i = 0; i < distinct_num_fields; ++i) {
            if (uniq_flags.no_columns) {
                fprintf(uniq_io.outF, "%s%s",
                        delim, distinct_fields[i].df_title);
            } else {
                fprintf(uniq_io.outF, "%s%*.*s",
                        delim,
                        distinct_fields[i].df_text_len,
                        distinct_fields[i].df_text_len,
                        distinct_fields[i].df_title);
            }
            delim[0] = g_delimiter;
        }
    }

    if (uniq_flags.no_final_delimiter) {
        delim[0] = '\0';
    }

    fprintf(uniq_io.outF, "%s\n", delim);
}


/*
 *  writeAsciiRecord(key, value, distincts);
 *
 *    Verifies that the 'value' and 'distincts' values are within the
 *    limits specified by the user.  If they are not, the function
 *    returns without printing anything.
 *
 *    Unpacks the fields from 'key' and prints the key fields, the
 *    value fields, and the distinct fields to the global output
 *    stream 'uniq_io.outF'.
 */
static void writeAsciiRecord(
    uint8_t        *key,
    const uint64_t  values[NUM_VALUES],
    const uint64_t  distincts[NUM_DISTINCTS])
{
    size_t i;
    out_stream_t *ostream;
    char buf[VALUE_OUT_BUF_SIZE];
    rwRec rwrec;
    char delim[2] = {'\0', '\0'};

    /* see if values are within limits */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
          case VALUE_PACKETS:
          case VALUE_FLOWS:
            if ((values[i] < value_fields[i].vf_min)
                || (values[i] > value_fields[i].vf_max))
            {
                return;
            }
            break;

          case VALUE_STIME:
          case VALUE_ETIME:
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }

    /* see if distinct counts are within limits */
    for (i = 0; i < distinct_num_fields; ++i) {
        if ((distincts[i] < distinct_fields[i].df_min)
            || (distincts[i] > distinct_fields[i].df_max))
        {
            return;
        }
    }


    /* Zero out rwrec to avoid display errors---specifically with msec
     * fields and eTime. */
    RWREC_CLEAR(&rwrec);

    /* Initialize the protocol to 1, so that if the user has requested
     * ICMP without the protocol, we get ICMP values. */
    rwRecSetProto(&rwrec, 1);

#if SK_ENABLE_IPV6
    if (ipv6_policy >= SK_IPV6POLICY_MIX) {
        /* Force records to be in IPv6 format */
        rwRecSetIPv6(&rwrec);
    }
#endif /* SK_ENABLE_IPV6 */

    /* unpack the key into 'rwrec' */
    if (uniq_flags.presorted_input || uniq_flags.sort_output) {
        /* create key in network byte-order */
        for (i = 0; i < key_num_fields; ++i) {
            switch (key_fields[i].kf_id) {
              case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetSIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NETWORK_32(rwRecSetSIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetDIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NETWORK_32(rwRecSetDIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetNhIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NETWORK_32(rwRecSetNhIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_SPORT:
                KEY_TO_REC_NETWORK_16(rwRecSetSPort, &rwrec, key, i);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                KEY_TO_REC_NETWORK_16(rwRecSetDPort, &rwrec, key, i);
                break;

              case RWREC_FIELD_PROTO:
                KEY_TO_REC_NETWORK_08(rwRecSetProto, &rwrec, key, i);
                break;

              case RWREC_FIELD_PKTS:
                KEY_TO_REC_NETWORK_32(rwRecSetPkts, &rwrec, key, i);
                break;

              case RWREC_FIELD_BYTES:
                KEY_TO_REC_NETWORK_32(rwRecSetBytes, &rwrec, key, i);
                break;

              case RWREC_FIELD_FLAGS:
                KEY_TO_REC_NETWORK_08(rwRecSetFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_SID:
                KEY_TO_REC_NETWORK_16(rwRecSetSensor, &rwrec, key, i);
                break;

              case RWREC_FIELD_INPUT:
                KEY_TO_REC_NETWORK_16(rwRecSetInput, &rwrec, key, i);
                break;

              case RWREC_FIELD_OUTPUT:
                KEY_TO_REC_NETWORK_16(rwRecSetOutput, &rwrec, key, i);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                KEY_TO_REC_NETWORK_08(rwRecSetInitFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_REST_FLAGS:
                KEY_TO_REC_NETWORK_08(rwRecSetRestFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_TCP_STATE:
                KEY_TO_REC_NETWORK_08(rwRecSetTcpState, &rwrec, key, i);
                break;

              case RWREC_FIELD_APPLICATION:
                KEY_TO_REC_NETWORK_16(rwRecSetApplication, &rwrec, key, i);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                KEY_TO_REC_NETWORK_08(rwRecSetFlowType, &rwrec, key, i);
                break;

              default:
                key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                     KEY_FIELD_PTR(key, i),
                                     key_fields[i].kf_textbuf,
                                     key_fields[i].kf_text_len,
                                     &rwrec);
            }
        }
    } else {
        /* key can be in native byte-order */
        for (i = 0; i < key_num_fields; ++i) {
            switch (key_fields[i].kf_id) {
              case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetSIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NATIVE_32(rwRecSetSIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetDIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NATIVE_32(rwRecSetDIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemSetNhIPv6(&rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                KEY_TO_REC_NATIVE_32(rwRecSetNhIPv4, &rwrec, key, i);
                break;

              case RWREC_FIELD_SPORT:
                KEY_TO_REC_NATIVE_16(rwRecSetSPort, &rwrec, key, i);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                KEY_TO_REC_NATIVE_16(rwRecSetDPort, &rwrec, key, i);
                break;

              case RWREC_FIELD_PROTO:
                KEY_TO_REC_NATIVE_08(rwRecSetProto, &rwrec, key, i);
                break;

              case RWREC_FIELD_PKTS:
                KEY_TO_REC_NATIVE_32(rwRecSetPkts, &rwrec, key, i);
                break;

              case RWREC_FIELD_BYTES:
                KEY_TO_REC_NATIVE_32(rwRecSetBytes, &rwrec, key, i);
                break;

              case RWREC_FIELD_FLAGS:
                KEY_TO_REC_NATIVE_08(rwRecSetFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_SID:
                KEY_TO_REC_NATIVE_16(rwRecSetSensor, &rwrec, key, i);
                break;

              case RWREC_FIELD_INPUT:
                KEY_TO_REC_NATIVE_16(rwRecSetInput, &rwrec, key, i);
                break;

              case RWREC_FIELD_OUTPUT:
                KEY_TO_REC_NATIVE_16(rwRecSetOutput, &rwrec, key, i);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                KEY_TO_REC_NATIVE_08(rwRecSetInitFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_REST_FLAGS:
                KEY_TO_REC_NATIVE_08(rwRecSetRestFlags, &rwrec, key, i);
                break;

              case RWREC_FIELD_TCP_STATE:
                KEY_TO_REC_NATIVE_08(rwRecSetTcpState, &rwrec, key, i);
                break;

              case RWREC_FIELD_APPLICATION:
                KEY_TO_REC_NATIVE_16(rwRecSetApplication, &rwrec, key, i);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                KEY_TO_REC_NATIVE_08(rwRecSetFlowType, &rwrec, key, i);
                break;

              default:
                key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                     KEY_FIELD_PTR(key, i),
                                     key_fields[i].kf_textbuf,
                                     key_fields[i].kf_text_len,
                                     &rwrec);
            }
        }
    }

    /* print key fields */
    for (i = 0; i < output_count; ++i) {
        ostream = outputs[i];
        if (ostream->a_stream) {
            rwAsciiPrintRec(ostream->a_stream, &rwrec);
        } else if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%c",
                    ostream->k_field->kf_textbuf, g_delimiter);
        } else {
            fprintf(uniq_io.outF, "%*s%c",
                    ostream->k_field->kf_text_len-1,
                    ostream->k_field->kf_textbuf, g_delimiter);
        }
    }

    /* print value fields */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
          case VALUE_PACKETS:
          case VALUE_FLOWS:
            sprintf(buf, ("%" PRIu64), values[i]);
            break;

          case VALUE_STIME:
          case VALUE_ETIME:
            sktimestamp_r(buf, ((sktime_t)1000 * values[i]), time_flags);
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }

        if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, "%s%s", delim, buf);
        } else {
            fprintf(uniq_io.outF, "%s%*s",
                    delim, value_fields[i].vf_text_len, buf);
        }
        delim[0] = g_delimiter;
    }

    /* print distinct fields */
    for (i = 0; i < distinct_num_fields; ++i) {
        if (uniq_flags.no_columns) {
            fprintf(uniq_io.outF, ("%s%" PRIu64), delim, distincts[i]);
        } else {
            fprintf(uniq_io.outF, ("%s%*" PRIu64),
                    delim, distinct_fields[i].df_text_len, distincts[i]);
        }
        delim[0] = g_delimiter;
    }

    if (uniq_flags.no_final_delimiter) {
        delim[0] = '\0';
    }

    fprintf(uniq_io.outF, "%s\n", delim);
}


/*
 *  packValues(rwrec, out_val_ptr);
 *
 *    For each value field that is being computed---determined by the
 *    global value_fields[]---add the values from the current 'rwrec'
 *    to the values that are packed into the byte-array 'out_val_ptr'.
 */
static void packValues(
    const rwRec    *rwrec,
    uint8_t        *out_val_ptr)
{
    uint64_t val64;
#if  RWUNIQ_USE_MEMCPY
    uint32_t val32;
#else
    uint32_t *pval32;
#endif
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
            VAL_FIELD_MEMGET(out_val_ptr, i, &val64);
            val64 += rwRecGetBytes(rwrec);
            VAL_FIELD_MEMSET(out_val_ptr, i, &val64);
            break;

          case VALUE_PACKETS:
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(out_val_ptr, i, &val32);
            val32 += rwRecGetPkts(rwrec);
            VAL_FIELD_MEMSET(out_val_ptr, i, &val32);
#else
            *VAL_FIELD_GET_ADDR32(out_val_ptr, i) += rwRecGetPkts(rwrec);
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case VALUE_FLOWS:
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(out_val_ptr, i, &val32);
            ++val32;
            VAL_FIELD_MEMSET(out_val_ptr, i, &val32);
#else
            ++*VAL_FIELD_GET_ADDR32(out_val_ptr, i);
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case VALUE_STIME:
            /* earliest sTime */
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(out_val_ptr, i, &val32);
            if ((val32 == 0) || (rwRecGetStartSeconds(rwrec) < val32)) {
                val32 = rwRecGetStartSeconds(rwrec);
                VAL_FIELD_MEMSET(out_val_ptr, i, &val32);
            }
#else
            pval32 = VAL_FIELD_GET_ADDR32(out_val_ptr, i);
            if ((*pval32 == 0) || (rwRecGetStartSeconds(rwrec) < *pval32)) {
                *pval32 = rwRecGetStartSeconds(rwrec);
            }
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case VALUE_ETIME:
            /* latest eTime */
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(out_val_ptr, i, &val32);
            if (val32 < rwRecGetEndSeconds(rwrec)) {
                val32 = rwRecGetEndSeconds(rwrec);
                VAL_FIELD_MEMSET(out_val_ptr, i, &val32);
            }
#else
            pval32 = VAL_FIELD_GET_ADDR32(out_val_ptr, i);
            if (*pval32 < rwRecGetEndSeconds(rwrec)) {
                *pval32 = rwRecGetEndSeconds(rwrec);
            }
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }
}


/*
 *  unpackValues(out_val, packed_val);
 *
 *    Does reverse of the packValues() subroutine: Unpacks the
 *    'packed_val' values into 'out_val', an array of NUM_VALUES
 *    elements.  Values unpacked depend on the elements of the
 *    value_fields[] array and occur in that order.
 */
static void unpackValues(
    uint64_t        out_val[NUM_VALUES],
    uint8_t        *packed_val)
{
#if  RWUNIQ_USE_MEMCPY
    uint32_t val32;
#endif
    size_t i;

    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
            VAL_FIELD_MEMGET(packed_val, i, &out_val[i]);
            break;

          case VALUE_PACKETS:
          case VALUE_FLOWS:
          case VALUE_STIME:
          case VALUE_ETIME:
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(packed_val, i, &val32);
            out_val[i] = val32;
#else
            out_val[i] = *VAL_FIELD_GET_ADDR32(packed_val, i);
#endif
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }
}


/*
 *  unpackAndAddValues(val, packed_val);
 *
 *    Similar to unpackValues(), except instead of overwriting the
 *    values in the 'val' array, the values in the 'packed_val' byte
 *    array are added to the current value.  Values unpacked depend on
 *    the elements of the value_fields[] array and occur in that
 *    order.
 */
static void unpackAndAddValues(
    uint64_t        val[NUM_VALUES],
    uint8_t        *packed_val)
{
    size_t i;
    uint64_t tmp64;
#if  RWUNIQ_USE_MEMCPY
    uint32_t tmp32;
#endif

    /* add this entry's values */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
            VAL_FIELD_MEMGET(packed_val, i, &tmp64);
            val[i] += tmp64;
            break;

          case VALUE_PACKETS:
          case VALUE_FLOWS:
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(packed_val, i, &tmp32);
            val[i] += tmp32;
#else
            val[i] += *VAL_FIELD_GET_ADDR32(packed_val, i);
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case VALUE_STIME:
            /* earliest sTime */
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(packed_val, i, &tmp32);
            if (val[i] > tmp32) {
                val[i] = tmp32;
            }
#else
            if (val[i] > *VAL_FIELD_GET_ADDR32(packed_val, i)) {
                val[i] = *VAL_FIELD_GET_ADDR32(packed_val, i);
            }
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case VALUE_ETIME:
            /* latest eTime */
#if  RWUNIQ_USE_MEMCPY
            VAL_FIELD_MEMGET(packed_val, i, &tmp32);
            if (val[i] < tmp32) {
                val[i] = tmp32;
            }
#else
            if (val[i] < *VAL_FIELD_GET_ADDR32(packed_val, i)) {
                val[i] = *VAL_FIELD_GET_ADDR32(packed_val, i);
            }
#endif  /* RWUNIQ_USE_MEMCPY */
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }
}


/*
 *  addValuesDistincts(rwrec, values, sip_set, dip_set);
 *
 *    For each value field that is being computed---determined by the
 *    global value_fields[]---add the values from the current 'rwrec'
 *    to the values that are in the array 'values'.  Similar to
 *    packValues(), except the values are stored in an array.
 *
 *    In addition, the sIP and/or dIP will be stored in the 'sip_set'
 *    and/or 'dip_set' IPSets, if those values are non-NULL.
 */
static void addValuesDistincts(
    const rwRec    *rwrec,
    uint64_t        values[NUM_VALUES],
    uint64_t        distincts[NUM_DISTINCTS],
    skIPTree_t     *sip_set,
    skIPTree_t     *dip_set)
{
    size_t i;

    /* we want to ensure that the IPs come out of the IPset in
     * memcmp() order, so we need to byte-swap the values on
     * little-endian machines. for counting, the byte-order is
     * irrelevant. */
    for (i = 0; i < distinct_num_fields; ++i) {
        switch ((distinct_types_t)distinct_fields[i].df_id) {
          case DISTINCT_SIP:
            /* unique source ips */
            MEMCHECK_IPSET_ADD(sip_set, htonl(rwRecGetSIPv4(rwrec)),
                               distincts[i]);
            break;

          case DISTINCT_DIP:
            /* unique destination ips */
            MEMCHECK_IPSET_ADD(dip_set, htonl(rwRecGetDIPv4(rwrec)),
                               distincts[i]);
            break;

          case _NUM_DISTINCTS_:
            skAppPrintErr("Unhandled distinct_fields index %d", (int)i);
            abort();
        }
    }

    /* add this entry's values */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
            values[i] += rwRecGetBytes(rwrec);
            break;

          case VALUE_PACKETS:
            values[i] += rwRecGetPkts(rwrec);
            break;

          case VALUE_FLOWS:
            ++values[i];
            break;

          case VALUE_STIME:
            /* earliest sTime */
            if ((values[i] == 0)
                || (rwRecGetStartSeconds(rwrec) < values[i]))
            {
                values[i] = rwRecGetStartSeconds(rwrec);
            }
            break;

          case VALUE_ETIME:
            /* latest eTime */
            if (values[i] < rwRecGetEndSeconds(rwrec)) {
                values[i] = rwRecGetEndSeconds(rwrec);
            }
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }
}


/*
 *  status = fillRecordAndKey(rwios, rwrec, key);
 *
 *    Reads a flow record from 'rwios' into the memory pointed at by
 *    'rwrec', computes the key based on the global key_fields[]
 *    settings, and fills in the parameter 'key' with that value.
 *    Return 1 if a record was read, or 0 if it was not.
 */
static int fillRecordAndKey(
    skstream_t *rwios,
    rwRec      *rwrec,
    uint8_t    *key)
{
    size_t i;

    /* get next record */
    if (skStreamReadRecord(rwios, rwrec)) {
        return 0;
    }
    ++record_count;

    if (uniq_flags.presorted_input || uniq_flags.sort_output) {
        /* create key in network byte-order */
        for (i = 0; i < key_num_fields; ++i) {
            switch (key_fields[i].kf_id) {
              case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetSIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NETWORK_32(rwRecGetSIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetDIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NETWORK_32(rwRecGetDIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetNhIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NETWORK_32(rwRecGetNhIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_SPORT:
                REC_TO_KEY_NETWORK_16(rwRecGetSPort, rwrec, key, i);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                REC_TO_KEY_NETWORK_16(rwRecGetDPort, rwrec, key, i);
                break;

              case RWREC_FIELD_PROTO:
                REC_TO_KEY_NETWORK_08(rwRecGetProto, rwrec, key, i);
                break;

              case RWREC_FIELD_PKTS:
                REC_TO_KEY_NETWORK_32(rwRecGetPkts, rwrec, key, i);
                break;

              case RWREC_FIELD_BYTES:
                REC_TO_KEY_NETWORK_32(rwRecGetBytes, rwrec, key, i);
                break;

              case RWREC_FIELD_FLAGS:
                REC_TO_KEY_NETWORK_08(rwRecGetFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_SID:
                REC_TO_KEY_NETWORK_16(rwRecGetSensor, rwrec, key, i);
                break;

              case RWREC_FIELD_INPUT:
                REC_TO_KEY_NETWORK_16(rwRecGetInput, rwrec, key, i);
                break;

              case RWREC_FIELD_OUTPUT:
                REC_TO_KEY_NETWORK_16(rwRecGetOutput, rwrec, key, i);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                REC_TO_KEY_NETWORK_08(rwRecGetInitFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_REST_FLAGS:
                REC_TO_KEY_NETWORK_08(rwRecGetRestFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_TCP_STATE:
                REC_TO_KEY_NETWORK_08(rwRecGetTcpState, rwrec, key, i);
                break;

              case RWREC_FIELD_APPLICATION:
                REC_TO_KEY_NETWORK_16(rwRecGetApplication, rwrec, key, i);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                REC_TO_KEY_NETWORK_08(rwRecGetFlowType, rwrec, key, i);
                break;

              default:
                key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                     KEY_FIELD_PTR(key, i),
                                     NULL, 0, rwrec);
            }
        }
    } else {
        /* key can be in native byte-order */
        for (i = 0; i < key_num_fields; ++i) {
            switch (key_fields[i].kf_id) {
              case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetSIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NATIVE_32(rwRecGetSIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetDIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NATIVE_32(rwRecGetDIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    rwRecMemGetNhIPv6(rwrec, KEY_FIELD_PTR(key, i));
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                REC_TO_KEY_NATIVE_32(rwRecGetNhIPv4, rwrec, key, i);
                break;

              case RWREC_FIELD_SPORT:
                REC_TO_KEY_NATIVE_16(rwRecGetSPort, rwrec, key, i);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                REC_TO_KEY_NATIVE_16(rwRecGetDPort, rwrec, key, i);
                break;

              case RWREC_FIELD_PROTO:
                REC_TO_KEY_NATIVE_08(rwRecGetProto, rwrec, key, i);
                break;

              case RWREC_FIELD_PKTS:
                REC_TO_KEY_NATIVE_32(rwRecGetPkts, rwrec, key, i);
                break;

              case RWREC_FIELD_BYTES:
                REC_TO_KEY_NATIVE_32(rwRecGetBytes, rwrec, key, i);
                break;

              case RWREC_FIELD_FLAGS:
                REC_TO_KEY_NATIVE_08(rwRecGetFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_SID:
                REC_TO_KEY_NATIVE_16(rwRecGetSensor, rwrec, key, i);
                break;

              case RWREC_FIELD_INPUT:
                REC_TO_KEY_NATIVE_16(rwRecGetInput, rwrec, key, i);
                break;

              case RWREC_FIELD_OUTPUT:
                REC_TO_KEY_NATIVE_16(rwRecGetOutput, rwrec, key, i);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                REC_TO_KEY_NATIVE_08(rwRecGetInitFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_REST_FLAGS:
                REC_TO_KEY_NATIVE_08(rwRecGetRestFlags, rwrec, key, i);
                break;

              case RWREC_FIELD_TCP_STATE:
                REC_TO_KEY_NATIVE_08(rwRecGetTcpState, rwrec, key, i);
                break;

              case RWREC_FIELD_APPLICATION:
                REC_TO_KEY_NATIVE_16(rwRecGetApplication, rwrec, key, i);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                REC_TO_KEY_NATIVE_08(rwRecGetFlowType, rwrec, key, i);
                break;

              default:
                key_fields[i].kf_fxn(key_fields[i].kf_fxn_id,
                                     KEY_FIELD_PTR(key, i),
                                     NULL, 0, rwrec);
            }
        }
    }

    return 1;
}


/*
 *  status = fillRecordAndKeyForHash(rwrec, key);
 *
 *    Fills in 'rwrec' with a flow read from the input.  Processes
 *    stdin or the files listed on the command line.  Computes the key
 *    based on the global key_fields[] and distinct_fields[] settings
 *    and fills in the parameter 'key' with that value.
 */
static int fillRecordAndKeyForHash(
    rwRec      *rwrec,
    uint8_t    *key)
{
    static skstream_t *rwios = NULL;
    size_t i;
    int rv;

    do {
        if (rwios && fillRecordAndKey(rwios, rwrec, key)) {
            /* Add the distinct fields to the key. */
            for (i = 0; i < distinct_num_fields; ++i) {
                distinct_fields[i].df_r2k(rwrec,
                                          key +distinct_fields[i].df_offset);
            }
            return 1;
        }

        /* no more records; close input file */
        skStreamDestroy(&rwios);

        /* goto next file */
        rv = appNextInput(&rwios);
    } while (rv == 0);

    if (rv != 1) {
        /* unexpected error */
        appExit(EXIT_FAILURE);
    }

    /* out of files */
    return 0;
}


/*
 *  printTableSimple(hash_table);
 *
 *    Just print the key/value pairs from the hash table.  There are
 *    no distinct values to keep track of.
 */
static void printTableSimple(HashTable *ht)
{
    HASH_ITER it;
    uint8_t *key_ptr;
    uint8_t *val_ptr;
    uint64_t values[NUM_VALUES];
    uint64_t distincts[NUM_DISTINCTS];

    /* We can do simple iteration through the hash table, though we
     * need to sort the hash table if sorted output requested */
    if (uniq_flags.sort_output) {
        hashlib_sort_entries(ht);
    }

    /* create the iterator */
    it = hashlib_create_iterator(ht);

    while (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) != ERR_NOMOREENTRIES) {
        unpackValues(values, val_ptr);
        writeAsciiRecord(key_ptr, values, distincts);
    }
}


/*
 *  printTableSingleDistinct(hash_table);
 *
 *    Print the entries from the hash table when one distinct field
 *    was requested.  Part of each entry's key is the column we are
 *    computing the 'distinct' count of.
 */
static void printTableSingleDistinct(
    HashTable  *ht)
{
    HASH_ITER it;
    uint8_t *cached_key = scratch;
    uint8_t *key_ptr;
    uint8_t *val_ptr;
    uint64_t values[NUM_VALUES];
    uint64_t distincts[NUM_DISTINCTS];

    /* Sort the entries then iterate through the hash table,
     * watching for the key (minus the distinct column) to
     * change. */
    hashlib_sort_entries(ht);

    /* create the iterator */
    it = hashlib_create_iterator(ht);

    /* get first key/value from hash table */
    if (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) == ERR_NOMOREENTRIES) {
        /* no data.  done. */
        return;
    }

    /* cache the key */
    memcpy(cached_key, key_ptr, key_byte_width);

    /* unpack the values */
    unpackValues(values, val_ptr);

    /* set the distincts */
    distincts[0] = 1;

    /* process the remaining hash entries */
    while (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) != ERR_NOMOREENTRIES) {
        /* compare keys */
        if (0 != memcmp(cached_key, key_ptr, key_byte_width)) {
            /* keys differ, print the old key and the values */
            writeAsciiRecord(cached_key, values, distincts);

            /* remember the key and reset values and distincts */
            memcpy(cached_key, key_ptr, key_byte_width);
            unpackValues(values, val_ptr);
            distincts[0] = 1;
        } else {
            /* keys match; add current 'val_ptr' to the 'values' array */
            unpackAndAddValues(values, val_ptr);
            ++distincts[0];
        }
    }

    /* print final entry */
    writeAsciiRecord(cached_key, values, distincts);
}


/*
 *  printTableMultipleDistinct(hash_table)
 *
 *    Print the entries from the hash table when multiple distinct
 *    fields were requested.  Part of each entry's key are the columns
 *    we are computing the 'distinct' count of.
 */
static void printTableMultipleDistinct(
    HashTable  *ht)
{
    HASH_ITER it;
    uint8_t *cached_key = scratch;
    uint8_t *key_ptr;
    uint8_t *val_ptr;
    uint64_t values[NUM_VALUES];
    uint64_t distincts[NUM_DISTINCTS];
    skIPTree_t *ipset = NULL;
    uint32_t ipaddr;
    uint32_t tmp32;

    /* multiple distinct columns.  sort the hash entries.  the first
     * distinct column we get for "free" by watching the values go by.
     * For the other columns, we must maintain a list of values we
     * have seen. */
    hashlib_sort_entries(ht);

    /* create the iterator */
    it = hashlib_create_iterator(ht);

    /* create an IPset for keeping track of values we see */
    if (skIPTreeCreate(&ipset)) {
        skAppPrintErr("Unable to create IPset");
        appExit(EXIT_FAILURE);
    }

    /* get first key/value from hash table */
    if (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) == ERR_NOMOREENTRIES) {
        /* no data.  done */
        skIPTreeDelete(&ipset);
        return;
    }

    /* cache the key */
    memcpy(cached_key, key_ptr, key_byte_width);

    /* unpack the values */
    unpackValues(values, val_ptr);

    /* cache the first distinct value---which we know is an IP
     * address */
    distincts[0] = 1;
    GET_DISTINCT_IP(key_ptr, 0, &ipaddr);

    /* begin list of the other distinct values (also an IP) */
    distincts[1] = 0;
    GET_DISTINCT_IP(key_ptr, 1, &tmp32);
    MEMCHECK_IPSET_ADD(ipset, tmp32, distincts[1]);

    /* process the remaining hash entries */
    while (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) != ERR_NOMOREENTRIES) {
        /* compare keys */
        if (0 != memcmp(cached_key, key_ptr, key_byte_width)) {
            /* keys differ, print the old key and the values */
            writeAsciiRecord(cached_key, values, distincts);
            skIPTreeRemoveAll(ipset);
            distincts[1] = 0;

            /* remember the key, set values, store distincts */
            memcpy(cached_key, key_ptr, key_byte_width);
            unpackValues(values, val_ptr);
            distincts[0] = 1;
            GET_DISTINCT_IP(key_ptr, 0, &ipaddr);
            GET_DISTINCT_IP(key_ptr, 1, &tmp32);
            MEMCHECK_IPSET_ADD(ipset, tmp32, distincts[1]);
        } else {
            /* keys match; add current 'val_ptr' to the 'values' array */
            unpackAndAddValues(values, val_ptr);

            /* handle the distinct fields */
            GET_DISTINCT_IP(key_ptr, 0, &tmp32);
            if (ipaddr != tmp32) {
                ++distincts[0];
                ipaddr = tmp32;
            }
            GET_DISTINCT_IP(key_ptr, 1, &tmp32);
            MEMCHECK_IPSET_ADD(ipset, tmp32, distincts[1]);
        }
    }

    /* print final entry */
    skIPTreeDelete(&ipset);

    writeAsciiRecord(cached_key, values, distincts);
}


/*
 *  count = openAllTempFiles(fp[], idx_1, idx_2);
 *
 *    Open all temporary files numbered between 'idx_1' and 'idx_2',
 *    put the file handles in the array 'fp', and return the number of
 *    files opened.  Normally, the number of returned file handles is
 *    given by (1 + idx_2 - idx_1).
 *
 *    If it is impossible to open all files due to a lack of file
 *    handles, the existing temporary files will be merged into new
 *    temporary files, and then another attempt will be made to open
 *    all files.
 *
 *    This function will only return when it is possible to return a
 *    file handle to every existing temporary file.  If it is unable
 *    to create a new temporary file, it causes the application to
 *    exit.
 */
static int openAllTempFiles(
    FILE   *fps[],
    int     tmp_idx_1,
    int     tmp_idx_2)
{
    uint8_t node[MAX_TMP_FILES][MAX_HASH_NODE_BYTES];
    size_t has_data[MAX_TMP_FILES];
    int i, j;
    int count = 0;
    int lowest;
    int tmp_idx_a = tmp_idx_1;
    int tmp_idx_b = tmp_idx_2;
    FILE *fp_intermediate = NULL;
    int tmp_idx_intermediate = tmp_idx_2;

    memset(has_data, 0, sizeof(has_data));

    /* This loop repeats as long as we haven't read all of the temp
     * files generated while reading the flows. */
    while (1) {

        PRINTDEBUG((stderr, "opening tmp files %d through %d\n",
                    tmp_idx_a, tmp_idx_b));

        /* open an intermediate temp file.  The merge-sort will have
         * to write records here if there are not enough file handles
         * available to open all the tempoary files we wrote while
         * reading the data. */
        ++tmp_idx_intermediate;
        fp_intermediate = tempFileCreate(tmp_idx_intermediate);
        if (fp_intermediate == NULL) {
            appExit(EXIT_FAILURE);
        }

        /* Repeat for as many temp files as we can open before we hit
         * EMFILE or ENOMEM, or all of them if we can */
        for (count = 0, j = tmp_idx_a; j <= tmp_idx_b; count++, j++) {
            if (count == MAX_TMP_FILES) {
                PRINTDEBUG((stderr, "too many temporary files opened\n"));
                appExit(EXIT_FAILURE);
            }
            fps[count] = tempFileOpen(j);
            if (fps[count] == NULL) {
                if ((count > 0) && ((errno == EMFILE) || (errno == ENOMEM))) {
                    /* Blast!  We can't open any more temp files.  So,
                     * we rewind by one to catch this one the next
                     * time around. */
                    tmp_idx_b = j - 1;
                    PRINTDEBUG((stderr, ("EMFILE limit hit--"
                                         "merging %d through %d to %d\n"),
                                tmp_idx_a, tmp_idx_b, tmp_idx_intermediate));
                    break;
                } else {
                    PRINTDEBUG((stderr, "couldn't open tmp file %s\n",
                                tempFileGetName(j)));
                    appExit(EXIT_FAILURE);
                }
            }
        }
        PRINTDEBUG((stderr,"opened %d temp files\n", count));

        /* Check to see if we've opened all temp files.  If so, close
         * the intermediate file and return */
        if (tmp_idx_b == tmp_idx_2) {
            PRINTDEBUG((stderr, "successfully opened all temp files\n"));
            if (EOF == fclose(fp_intermediate)) {
                skAppPrintErr("Error closing file %s: %s",
                              tempFileGetName(tmp_idx_intermediate),
                              strerror(errno));
                appExit(EXIT_FAILURE);
            }
            return count;
        }
        /* Else, we could not open all temp files, so merge all opened
         * temp files into the intermediate file */

        /* Read the first key/value pair from each temp file into the
         * work buffers. */
        for (i = 0; i < count; i++) {
            has_data[i] = fread(node[i], hash_node_bytes, 1, fps[i]);
        }

        /* assign an arbitary key/value pair as the lowest */
        lowest = 0;

        /* inner-while: loop until we process all key/value pairs in
         * all opened files */
        while (1) {

            /* make certain 'lowest' key/value pair has data; if not
             * assign a new key/value pair arbitrarily as lowest. */
            if ( !has_data[lowest]) {
                for (lowest = 0; lowest < count; ++lowest) {
                    if (has_data[lowest]) {
                        break;
                    }
                }
                /* exit the inner-while() if all key/value pairs for
                 * all opened files have been processed */
                if (lowest == count) {
                    break;
                }
            }

            /* Loop over each of the key/value pairs */
            for (i = 0; i < count; i++) {
                /* If the temp file has not reached EOF yet and if the
                 * temp file's key is lower than the lowest... */
                if ((i != lowest)
                    && has_data[i]
                    && (memcmp(node[i], node[lowest], hash_key_bytes) < 0))
                {
                    /* record the index of this temp file as having
                     * the lowest key */
                    lowest = i;
                }
            }

            /* write the lowest key/value pair to the intermediate
             * temp file */
            if (fwrite(node[lowest], hash_node_bytes, 1, fp_intermediate) != 1)
            {
                skAppPrintErr("Could not write to tmpfile %s",
                              tempFileGetName(tmp_idx_intermediate));
                appExit(EXIT_FAILURE);
            }

            /* replace the node we just wrote */
            has_data[lowest] = fread(node[lowest], hash_node_bytes,
                                     1, fps[lowest]);
        }

        /* Close and delete all of the temp files that we processed
         * this time. */
        for (i = 0; i < count; i++) {
            if (fps[i]) {
                fclose(fps[i]);
            }
            tempFileRemove(tmp_idx_a + i);
        }

        /* Close the intermediate temp file. */
        if (fp_intermediate) {
            if (EOF == fclose(fp_intermediate)) {
                skAppPrintErr("Error closing file %s: %s",
                              tempFileGetName(tmp_idx_intermediate),
                              strerror(errno));
                appExit(EXIT_FAILURE);
            }
            fp_intermediate = NULL;
        }

        /* Start the next merge with the next input temp file */
        tmp_idx_a = tmp_idx_b + 1;

        /* End the next merge with the temp file we just created */
        tmp_idx_2 = tmp_idx_b = tmp_idx_intermediate;
    }

    return -1;    /* NOT REACHED */
}


/*
 *  mergeFiles(temp_id_1, temp_id_2)
 *
 *    Merge the temporary files numbered from 'temp_id_1' to
 *    'temp_id_2' into the output file 'out_ios', maintaining sorted
 *    order.  Exits the application if an error occurs.
 */
static void mergeFiles(int tmp_idx_1, int tmp_idx_2)
{
    uint8_t *cached_key = scratch;
    FILE *fps[MAX_TMP_FILES];
    uint8_t node[MAX_TMP_FILES][MAX_HASH_NODE_BYTES];
    size_t has_data[MAX_TMP_FILES];
    uint8_t *key_ptr;
    uint8_t *val_ptr;
    uint64_t values[NUM_VALUES];
    uint64_t distincts[NUM_DISTINCTS];
    skIPTree_t *ipset = NULL;
    uint32_t cached_ip = 0;
    uint32_t ipaddr;
    int lowest = -1;
    int count;
    int i;

    memset(has_data, 0, sizeof(has_data));
    memset(distincts, 0, sizeof(distincts));

    /* open all temp files. */
    count = openAllTempFiles(fps, tmp_idx_1, tmp_idx_2);

    /* Read the first key/value pair from each temp file into the
     * work buffers and find the one with the lowest key. */
    for (i = 0; i < count; i++) {
        has_data[i] = fread(node[i], hash_node_bytes, 1, fps[i]);
        if (has_data[i]) {
            if (lowest == -1) {
                /* first node we've seen; so it is the lowest */
                lowest = i;
            } else if (memcmp(node[i], node[lowest], hash_key_bytes) < 0) {
                /* record the index of this temp file as having
                 * the lowest key */
                lowest = i;
            }
        }
    }

    if (lowest == -1) {
        skAppPrintErr("Could not read records from any temp files");
        appExit(EXIT_FAILURE);
    }

    /* found the lowest */
    key_ptr = node[lowest];
    val_ptr = node[lowest] + hash_key_bytes;

    /* cache this first key, unpack its values, and initialize
     * everything required to keep track of distinct counts. */
    memcpy(cached_key, key_ptr, key_byte_width);
    unpackValues(values, val_ptr);
    switch (distinct_num_fields) {
      case 2:
        if (skIPTreeCreate(&ipset)) {
            skAppPrintErr("Unable to create IPset");
            appExit(EXIT_FAILURE);
        }
        GET_DISTINCT_IP(key_ptr, 1, &ipaddr);
        MEMCHECK_IPSET_ADD(ipset, ipaddr, distincts[1]);
        /* FALLTHROUGH */

      case 1:
        distincts[0] = 1;
        GET_DISTINCT_IP(key_ptr, 0, &cached_ip);
        break;

      default:
        assert(distinct_num_fields == 0);
        break;
    }

    /* process all remaining key/value pairs in all opened files */
    while (1) {
        /* replace the key/value pair we just processed by reading
         * from the file.  If we're at the end of that file, assign a
         * new key/value pair arbitrarily as lowest. */
        has_data[lowest] = fread(node[lowest], hash_node_bytes, 1,fps[lowest]);
        if (has_data[lowest] == 0) {
            for (lowest = 0; lowest < count; ++lowest) {
                if (has_data[lowest]) {
                    break;
                }
            }

            /* exit the while once all key/value pairs for all opened
             * files have been processed */
            if (lowest == count) {
                break;
            }
        }

        /* find the lowest key, using entire hash_key */
        for (i = 0; i < count; i++) {
            /* If the temp file has not reached EOF yet and if the
             * temp file's key is lower than the lowest... */
            if ((i != lowest)
                && has_data[i]
                && (memcmp(node[i], node[lowest], hash_key_bytes) < 0))
            {
                /* record the index of this temp file as having
                 * the lowest key */
                lowest = i;
            }
        }

        /* found the lowest key */
        key_ptr = node[lowest];
        val_ptr = node[lowest] + hash_key_bytes;

        /* compare this key with the cached_key */
        if (0 != memcmp(cached_key, key_ptr, key_byte_width)) {
            /* keys differ, print the old key and the values */
            writeAsciiRecord(cached_key, values, distincts);

            /* remember the key and reset values and distincts */
            memcpy(cached_key, key_ptr, key_byte_width);
            unpackValues(values, val_ptr);
            switch (distinct_num_fields) {
              case 2:
                distincts[1] = 0;
                skIPTreeRemoveAll(ipset);
                GET_DISTINCT_IP(key_ptr, 1, &ipaddr);
                MEMCHECK_IPSET_ADD(ipset, ipaddr, distincts[1]);
                /* FALLTHROUGH */
              case 1:
                distincts[0] = 1;
                GET_DISTINCT_IP(key_ptr, 0, &cached_ip);
                break;
              default:
                assert(distinct_num_fields == 0);
                break;
            }
        } else {
            /* keys are same; add the current values from 'val_ptr' to
             * the 'values' array */
            unpackAndAddValues(values, val_ptr);

            /* handle the distinct fields */
            switch (distinct_num_fields) {
              case 2:
                GET_DISTINCT_IP(key_ptr, 1, &ipaddr);
                MEMCHECK_IPSET_ADD(ipset, ipaddr, distincts[1]);
                /* FALLTHROUGH */
              case 1:
                GET_DISTINCT_IP(key_ptr, 0, &ipaddr);
                if (cached_ip != ipaddr) {
                    ++distincts[0];
                    cached_ip = ipaddr;
                }
                break;
              default:
                assert(distinct_num_fields == 0);
                break;
            }
        }
    }

    /* output the final entry */
    if (ipset) {
        skIPTreeDelete(&ipset);
    }
    writeAsciiRecord(cached_key, values, distincts);
}


/*
 *  status = dumpHashToTemp(tmp_idx, ht);
 *
 *    Write the entries in the hash table 'ht' to the temporary file
 *    whose index is 'tmp_idx'.  The entries are written in sorted
 *    order.  Return 0 on success, or -1 on failure.
 */
static int dumpHashToTemp(
    int         tmp_idx,
    HashTable  *ht)
{
    HASH_ITER it;
    uint8_t *key_ptr;
    uint8_t *val_ptr;
    FILE* temp_filep = NULL;
    int rv = -1; /* return value */

    /* sort the hash entries. */
    hashlib_sort_entries(ht);

    /* create the iterator */
    it = hashlib_create_iterator(ht);

    /* create temp file */
    temp_filep = tempFileCreate(tmp_idx);
    if (temp_filep == NULL) {
        return rv;
    }

    PRINTDEBUG((stderr, "Writing %u key/value pairs to '%s'...",
                hashlib_count_entries(ht), tempFileGetName(tmp_idx)));

    /* write the sorted entries to disk */
    while (hashlib_iterate(ht, &it, &key_ptr, &val_ptr) != ERR_NOMOREENTRIES) {
        if (!fwrite(key_ptr, hash_key_bytes, 1, temp_filep)
            || !fwrite(val_ptr, hash_value_bytes, 1, temp_filep))
        {
            /* error writing, errno may or may not be set */
            skAppPrintErr("Error writing records to temp file %s: %s",
                          tempFileGetName(tmp_idx), strerror(errno));
            PRINTDEBUG((stderr, "ERROR!\n"));
            goto END;
        }
    }

    /* Success so far */
    PRINTDEBUG((stderr, "done\n"));
    rv = 0;

  END:
    /* close the file */
    if (fclose(temp_filep) == EOF) {
        /* error closing file; only report error if status so far
         * is good. */
        if (rv == 0) {
            skAppPrintErr("Error closing temp file %s: %s",
                          tempFileGetName(tmp_idx), strerror(errno));
            PRINTDEBUG((stderr, "error on close\n"));
            rv = -1;
        }
    }

    return rv;
}


/*
 *  uniqHash();
 *
 *    Main control function that creates a hash table, processes the
 *    input (files or stdin), and prints the results.
 */
static void uniqHash(void)
{
    HashTable *ht;
    uint8_t *no_val_ptr;
    uint8_t *key_ptr = scratch; /* use global scratch space */
    uint8_t *val_ptr;
    int temp_file_count = 0;
    int rv = 0;
    rwRec rwrec;

    /* Create the value that the hash table uses to indicate a key
     * without a value */
    no_val_ptr = (uint8_t *)malloc(hash_value_bytes);
    if (!no_val_ptr) {
        skAppPrintErr("Unable to allocate no-value default");
        appExit(EXIT_FAILURE);
    }
    memset(no_val_ptr, 0xFF, hash_value_bytes);

    /* Create the hash table */
    ht = hashlib_create_table(hash_key_bytes, hash_value_bytes,
                              HTT_INPLACE,
                              no_val_ptr,
                              NULL,
                              0,
                              HASH_INITIAL_SIZE,
                              DEFAULT_LOAD_FACTOR);
    if (NULL == ht) {
        skAppPrintErr("Unable to create hash table");
        appExit(EXIT_FAILURE);
    }

    /* Fill in the record and specified key from the input stream */
    while (fillRecordAndKeyForHash(&rwrec, key_ptr)) {

        /* the 'insert' will set 'val_ptr' to the memory to use to
         * store the values. either fresh memory or the existing
         * value(s). */
        rv = hashlib_insert(ht, key_ptr, &val_ptr);
        switch (rv) {
          case OK:
            /* new value; clear the memory */
            memset(val_ptr, 0, hash_value_bytes);
            /* FALLTHROUGH */

          case OK_DUPLICATE:
            /* existing value; add new value */
            packValues(&rwrec, val_ptr);
            break;

          case ERR_OUTOFMEMORY:
          case ERR_NOMOREBLOCKS:
            /* out of memory */
            if (dumpHashToTemp(temp_file_count, ht)) {
                skAppPrintErr("Unable to create temporary file");
                appExit(EXIT_FAILURE);
            }
            ++temp_file_count;
            /* destroy and re-create the hash table */
            hashlib_free_table(ht);
            ht = hashlib_create_table(hash_key_bytes, hash_value_bytes,
                                      HTT_INPLACE,
                                      no_val_ptr,
                                      NULL,
                                      0,
                                      HASH_INITIAL_SIZE,
                                      DEFAULT_LOAD_FACTOR);
            if (NULL == ht) {
                skAppPrintErr("Unable to create hash table");
                rv = 1;
                goto END;
            }
            /* add this entry to the newly created hash table */
            rv = hashlib_insert(ht, key_ptr, &val_ptr);
            if (rv == OK) {
                memset(val_ptr, 0, hash_value_bytes);
                packValues(&rwrec, val_ptr);
            } else {
                skAppPrintErr(("Unexpected return code '%d'"
                               " from hash table insert on new hash table"),
                              rv);
                appExit(EXIT_FAILURE);
            }
            break;

          default:
            skAppPrintErr(("Unexpected return code '%d'"
                           " from hash table insert"),
                          rv);
            appExit(EXIT_FAILURE);
        }
    }

    /* Write out the headings */
    writeColTitles();

    if (temp_file_count) {
        /* dump the current/final hash entries to a file */
        if (dumpHashToTemp(temp_file_count, ht)) {
            skAppPrintErr("Unable to create temporary file");
            appExit(EXIT_FAILURE);
        }
        /* destroy the hash table; we're done with it */
        hashlib_free_table(ht);
        ht = NULL;

        mergeFiles(0, temp_file_count);
    } else if (distinct_num_fields == 0) {
        printTableSimple(ht);
    } else if (distinct_num_fields == 1) {
        printTableSingleDistinct(ht);
    } else {
        printTableMultipleDistinct(ht);
    }

  END:
    if (ht) {
        hashlib_free_table(ht);
    }
    free(no_val_ptr);
}


/*
 *  addNodeToTemp(fp, key, values, sip_set, dip_set);
 *
 *    Write new key/value pair(s) to the temp file 'fp'.  The
 *    parameter 'key' holds the current key encoded as a byte array;
 *    'values' is an array with the current byte/packet/flow etc
 *    values.  The 'sip_set' and 'dip_set' are either NULL or IPSets
 *    with the sIP and dIP seem for this 'key'.
 *
 *    'key' must be a byte array large enough to hold the current key,
 *    any distinct fields, and all value fields.  The function writes
 *    entires to the file until all sIP and dIP values have been
 *    written.
 */
static void addNodeToTemp(
    FILE       *fp,
    uint8_t    *node,
    uint64_t    values[NUM_VALUES],
    skIPTree_t *sip_set,
    skIPTree_t *dip_set)
{
    uint8_t *val_ptr = node + hash_key_bytes;
    skIPTreeIterator_t sip_it;
    skIPTreeIterator_t dip_it;
    skIteratorStatus_t sip_rv = SK_ITERATOR_NO_MORE_ENTRIES;
    skIteratorStatus_t dip_rv = SK_ITERATOR_NO_MORE_ENTRIES;
    uint8_t *sip_pos = NULL;
    uint8_t *dip_pos = NULL;
    uint32_t ip;
    uint32_t val32;
    size_t i;

    /* encode the values into the node */
    for (i = 0; i < value_num_fields; ++i) {
        switch ((value_types_t)value_fields[i].vf_id) {
          case VALUE_BYTES:
            VAL_FIELD_MEMSET(val_ptr, i, &(values[i]));
            break;

          case VALUE_PACKETS:
          case VALUE_FLOWS:
          case VALUE_STIME:
          case VALUE_ETIME:
            val32 = (uint32_t)values[i];
            VAL_FIELD_MEMSET(val_ptr, i, &val32);
            break;

          case _NUM_VALUES_:
            skAppPrintErr("Illegal value_fields index %d", (int)i);
            abort();
        }
    }

    /* encode the distinct values into the node.  on little endian
     * machines, the values were swapped before being added to the
     * IPset, so swap them again as we pull them out of the IPset.
     * The swapping ensures the IPSet Iterator returns them in
     * memcmp() order; i.e., 7.6.5.4 is returned before 2.3.4.5 */
    if (sip_set) {
        sip_pos = node + distinct_fields[DISTINCT_SIP].df_offset;
        assert(sip_pos < val_ptr);

        skIPTreeIteratorBind(&sip_it, sip_set);
        sip_rv = skIPTreeIteratorNext(&ip, &sip_it);
        ip = ntohl(ip);
        memcpy(sip_pos, &ip, sizeof(uint32_t));
    }
    if (dip_set) {
        dip_pos = node + distinct_fields[DISTINCT_DIP].df_offset;
        assert(dip_pos < val_ptr);

        skIPTreeIteratorBind(&dip_it, dip_set);
        dip_rv = skIPTreeIteratorNext(&ip, &dip_it);
        ip = ntohl(ip);
        memcpy(dip_pos, &ip, sizeof(uint32_t));
    }

    do {
        /* write key/value pair to temp file */
        if (fwrite(node, hash_node_bytes, 1, fp) != 1) {
            skAppPrintSyserror("Unable to write to temp file");
            appExit(EXIT_FAILURE);
        }

        /* get next sip and/or dip. the iterator does not modify the
         * value if there are no more entries. when one IPSet has no
         * more entries, just print the final value until the other
         * IPSet is done */
        if (sip_set) {
            sip_rv = skIPTreeIteratorNext(&ip, &sip_it);
            if (sip_rv == SK_ITERATOR_OK) {
                ip = ntohl(ip);
                memcpy(sip_pos, &ip, sizeof(uint32_t));
            }
        }
        if (dip_set) {
            dip_rv = skIPTreeIteratorNext(&ip, &dip_it);
            if (dip_rv == SK_ITERATOR_OK) {
                ip = ntohl(ip);
                memcpy(dip_pos, &ip, sizeof(uint32_t));
            }
        }
    } while (sip_rv == SK_ITERATOR_OK || dip_rv == SK_ITERATOR_OK);
}


/*
 *  uniqPresorted();
 *
 *    Main control function that reads presorted flow records from
 *    files or stdin and prints the results.
 */
static void uniqPresorted(void)
{
    skstream_t *rwios[MAX_TMP_FILES];
    rwRec rwrec[MAX_TMP_FILES];
    int has_data[MAX_TMP_FILES];
    uint8_t node[MAX_TMP_FILES][MAX_KEY_BYTE_WIDTH];
    uint8_t *cached_key = scratch; /* use global scratch space */
    uint64_t values[NUM_VALUES];
    size_t i;
    int rv = 0;
    uint64_t distincts[NUM_DISTINCTS] = {0, 0};
    size_t sip_idx = NUM_DISTINCTS;
    size_t dip_idx = NUM_DISTINCTS;
    skIPTree_t *sip_set = NULL;
    skIPTree_t *dip_set = NULL;
    FILE *fp_intermediate = NULL;
    int more_input_files = 1;
    int temp_file_count = 0;
    int temp_idx;
    size_t lowest;
    size_t count;

    memset(node, 0, sizeof(node));

    /* If DISTINCT_SIP and/or DISTINCT_DIP were specified, determine
     * their index in the distinct_fields array and create IPsets for
     * them. */
    for (i = 0; i < distinct_num_fields; ++i) {
        switch ((distinct_types_t)distinct_fields[i].df_id) {
          case DISTINCT_SIP:
            sip_idx = i;
            if (skIPTreeCreate(&sip_set)) {
                skAppPrintErr("Unable to create IPset");
                appExit(EXIT_FAILURE);
            }
            break;

          case DISTINCT_DIP:
            dip_idx = i;
            if (skIPTreeCreate(&dip_set)) {
                skAppPrintErr("Unable to create IPset");
                appExit(EXIT_FAILURE);
            }
            break;

          case _NUM_DISTINCTS_:
            skAppPrintErr("Unhandled distinct_fields index %d", (int)i);
            abort();
        }
    }

    /* Write out the headings */
    writeColTitles();

    /* This loop repeats as long as we haven't read all the input
     * files */
    do {
        /* open an intermediate temp file that we will use if there
         * are not enough file handles available to open all the input
         * files. */
        temp_idx = temp_file_count;
        fp_intermediate = tempFileCreate(temp_idx);
        if (fp_intermediate == NULL) {
            appExit(EXIT_FAILURE);
        }

        /* Open as many input files as we can before we run out of
         * files, hit EMFILE, or run out of memory */
        for (count = 0; count < MAX_TMP_FILES; ++count) {
            rv = appNextInput(&(rwios[count]));
            if (rv != 0) {
                break;
            }
        }
        switch (rv) {
          case 1:
            /* successfully opened all input files */
            PRINTDEBUG((stderr, "opened all inputs\n"));
            more_input_files = 0;
            break;
          case -1:
            /* unexpected error opening a file */
            appExit(EXIT_FAILURE);
          case -2:
            /* ran out of memory or file descriptors */
            PRINTDEBUG((stderr, "unable to open all inputs\n"));
            ++temp_file_count;
            break;
          case 0:
            if (count == MAX_TMP_FILES) {
                /* ran out of pointers for this run */
                PRINTDEBUG((stderr, "unable to open all inputs\n"));
                ++temp_file_count;
                break;
            }
            /* no other way that rv == 0 */
            abort();
          default:
            /* unexpected error */
            PRINTDEBUG((stderr, "got unexpected rv value = %d\n", rv));
            abort();
        }

        PRINTDEBUG((stderr, "count: %d\n", (int)count));

        /* Read the first record from each file and generate the key from
         * each record. */
        for (i = 0; i < count; i++) {
            has_data[i] = fillRecordAndKey(rwios[i], &rwrec[i], node[i]);
        }

        /* Assign arbitary key as lowest */
        lowest = 0;

        /* the inner-while loops until we process all records in all
         * open files */
        while (1) {

            /* make certain 'lowest' key has data; if not assign a
             * new key arbitrarily as lowest. */
            if ( !has_data[lowest]) {
                for (lowest = 0; lowest < count; ++lowest) {
                    if (has_data[lowest]) {
                        break;
                    }
                }
                /* exit the inner-while() if all records for all
                 * opened files have been processed */
                if (lowest == count) {
                    break;
                }
            }

            /* find the lowest key */
            for (i = 0; i < count; ++i) {
                if (i != lowest
                    && has_data[i]
                    && memcmp(node[i], node[lowest], key_byte_width) < 0)
                {
                    /* record the index of this lowest key */
                    lowest = i;
                }
            }

            /* cache this low key; initialize values from this record */
            memcpy(cached_key, node[lowest], key_byte_width);
            memset(values, 0, sizeof(values));
            addValuesDistincts(&rwrec[lowest], values, distincts,
                               sip_set, dip_set);

            /* replace the value we just processed */
            has_data[lowest] = fillRecordAndKey(rwios[lowest],
                                                &rwrec[lowest],
                                                node[lowest]);

            /* process all entries in all input files that match the
             * current key */
            for (i = 0; i < count; ++i) {
                while (has_data[i]
                       && 0 == memcmp(cached_key, node[i], key_byte_width))
                {
                    /* keys are same, add this entry's values */
                    addValuesDistincts(&rwrec[i], values, distincts,
                                       sip_set, dip_set);

                    /* get next record and key */
                    has_data[i] = fillRecordAndKey(rwios[i], &rwrec[i],
                                                   node[i]);
                }
            }

            /* output this key and its values.  If we opened all input
             * files, we simply print the data.  Otherwise, we need to
             * store the key and the value into the current temp file,
             * but any distinct fields makes this complicated.  We
             * loop through the IPSets and store the key and each sip
             * and dip into the temporary file.  Once all inputs are
             * processed, use mergeFiles() to merge the files. */

            if (temp_file_count) {
                addNodeToTemp(fp_intermediate, cached_key, values,
                              sip_set, dip_set);
            } else {
                /* print the entry and reset IPSets */
                writeAsciiRecord(cached_key, values, distincts);
                if (sip_set) {
                    skIPTreeRemoveAll(sip_set);
                    distincts[sip_idx] = 0;
                }
                if (dip_set) {
                    skIPTreeRemoveAll(dip_set);
                    distincts[dip_idx] = 0;
                }
            }
        } /* inner-while */

        /* Close the input files that we processed this time. */
        for (i = 0; i < count; i++) {
            skStreamDestroy(&rwios[i]);
        }

        /* Close the intermediate temp file. */
        if (fp_intermediate) {
            if (EOF == fclose(fp_intermediate)) {
                skAppPrintErr("Error closing file %s: %s",
                              tempFileGetName(temp_idx),
                              strerror(errno));
                appExit(EXIT_FAILURE);
            }
            fp_intermediate = NULL;
        }

    } while (more_input_files);

    /* We're done with the IPSets */
    if (sip_set) {
        skIPTreeDelete(&sip_set);
    }
    if (dip_set) {
        skIPTreeDelete(&dip_set);
    }

    /* If any temporary files were written, we now have to merge
     * them */
    if (0 != temp_file_count) {
        mergeFiles(0, temp_file_count);
    }
}


int main(int argc, char **argv)
{
    /* Global setup */
    appSetup(argc, argv);

    if (uniq_flags.presorted_input) {
        uniqPresorted();
    } else {
        uniqHash();
    }

    /* Done, do cleanup */
    appTeardown();

    return (distinct_err ? RWUNIQ_NO_MEMORY_EXIT_CODE : 0);
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
