/*
** Copyright (C) 2001-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwuniqsetup.c
**
**  Application setup for rwuniq.  See rwuniq.c for a description.
*/


#include "silk.h"

RCSIDENT("$SiLK: rwuniqsetup.c 13074 2008-11-20 20:28:45Z mthomas $");

#include "rwuniq.h"
#include "rwuniqtemp.h"
#include "sksite.h"


/* TYPEDEFS AND DEFINES */

/* file handle for --help usage message */
#define USAGE_FH stdout

/* Where to write filenames if --print-file specified */
#define PRINT_FILENAMES_FH  stderr

/* Maximum number of dynamic libraries that the application can open */
#define APP_MAX_DYNLIBS 8

/* Internal interface to the plug-in (dynlib) library */
typedef struct app_dynlib_st {
    /* the libdynlib object */
    dynlibInfoStruct_t *dlisp;
    /* subtract this offset from the global field ID to get the field
     * IDs used by this plug-in */
    unsigned int        offset;
    /* function to get output from dynlib */
    uniqf_t             fxn;
    /* whether this plug-in was specified via --dynamic-library */
    unsigned            via_switch :1;
} app_dynlib_t;


/* LOCAL VARIABLES */

/* available key fields */
static sk_stringmap_t *field_map = NULL;

/* the maximum ID in the field_names string-map */
static uint32_t max_field_id = 0;

/* the text the user entered for the --fields switch */
static char *fields_arg = NULL;

/* name of program to run to page output */
static char *pager;

/* the floor of the sTime and/or eTime */
static uint32_t time_bin_size = 0;

/* when time-binning is active and all time fields---sTime, eTime,
 * elapased---are requested, we must adjust 'elapsed' so that it is
 * equal to eTime-sTime.  this is non-0 if we must adjust */
static int adjust_elapsed = 0;

/* input checker */
static iochecksInfoStruct_t *ioISP = NULL;

/* List of dynamic libraries to attempt to open at startup */
static const char *app_dynlib_names[] = {
    SK_PLUGIN_ADD_SUFFIX("addrtype"),
    SK_PLUGIN_ADD_SUFFIX("ccfilter"),
    SK_PLUGIN_ADD_SUFFIX("pmapfilter"),
    SK_PLUGIN_ADD_SUFFIX("silkpython"),
    NULL /* sentinel */
};

/* Dynamic libraries we actually opened */
static app_dynlib_t app_dynlib_list[APP_MAX_DYNLIBS];

/* Number of dynamic libraries actually opened */
static int app_dynlib_count;

/* non-zero if we are shutting down due to a signal; controls whether
 * errors are printed in appTeardown(). */
static int caught_signal = 0;


/* OPTIONS */

typedef enum {
    OPT_FIELDS,
    OPT_PRESORTED_INPUT,
    OPT_SORT_OUTPUT,
    OPT_ALL_COUNTS,
    OPT_BYTES,
    OPT_PACKETS,
    OPT_FLOWS,
    OPT_STIME,
    OPT_ETIME,
    OPT_SIP_DISTINCT,
    OPT_DIP_DISTINCT,
    OPT_DYNAMIC_LIB_VALUE,
    OPT_BIN_TIME,
    OPT_EPOCH_TIME,
    OPT_INTEGER_IPS,
    OPT_ZERO_PAD_IPS,
    OPT_INTEGER_SENSORS,
    OPT_NO_TITLES,
    OPT_NO_COLUMNS,
    OPT_COLUMN_SEPARATOR,
    OPT_NO_FINAL_DELIMITER,
    OPT_DELIMITED,
    OPT_PRINT_FILENAMES,
    OPT_COPY_INPUT,
    OPT_OUTPUT_PATH,
    OPT_PAGER,
    OPT_TEMP_DIRECTORY,
    OPT_LEGACY_TIMESTAMPS
#ifdef ENABLE_BINARY
    , OPT_ENABLE_BINARY
#endif
} appOptionsEnum;


static struct option appOptions[] = {
    {"fields",              REQUIRED_ARG, 0, OPT_FIELDS},
    {"presorted-input",     NO_ARG,       0, OPT_PRESORTED_INPUT},
    {"sort-output",         NO_ARG,       0, OPT_SORT_OUTPUT},
    {"all-counts",          NO_ARG,       0, OPT_ALL_COUNTS},
    {"bytes",               OPTIONAL_ARG, 0, OPT_BYTES},
    {"packets",             OPTIONAL_ARG, 0, OPT_PACKETS},
    {"flows",               OPTIONAL_ARG, 0, OPT_FLOWS},
    {"stime",               NO_ARG,       0, OPT_STIME},
    {"etime",               NO_ARG,       0, OPT_ETIME},
    {"sip-distinct",        OPTIONAL_ARG, 0, OPT_SIP_DISTINCT},
    {"dip-distinct",        OPTIONAL_ARG, 0, OPT_DIP_DISTINCT},
    {OPT_DYNAMIC_LIBRARY,   REQUIRED_ARG, 0, OPT_DYNAMIC_LIB_VALUE},
    {"bin-time",            OPTIONAL_ARG, 0, OPT_BIN_TIME},
    {"epoch-time",          NO_ARG,       0, OPT_EPOCH_TIME},
    {"integer-ips",         NO_ARG,       0, OPT_INTEGER_IPS},
    {"zero-pad-ips",        NO_ARG,       0, OPT_ZERO_PAD_IPS},
    {"integer-sensors",     NO_ARG,       0, OPT_INTEGER_SENSORS},
    {"no-titles",           NO_ARG,       0, OPT_NO_TITLES},
    {"no-columns",          NO_ARG,       0, OPT_NO_COLUMNS},
    {"column-separator",    REQUIRED_ARG, 0, OPT_COLUMN_SEPARATOR},
    {"no-final-delimiter",  NO_ARG,       0, OPT_NO_FINAL_DELIMITER},
    {"delimited",           OPTIONAL_ARG, 0, OPT_DELIMITED},
    {"print-filenames",     NO_ARG,       0, OPT_PRINT_FILENAMES},
    {"copy-input",          REQUIRED_ARG, 0, OPT_COPY_INPUT},
    {"output-path",         REQUIRED_ARG, 0, OPT_OUTPUT_PATH},
    {"pager",               REQUIRED_ARG, 0, OPT_PAGER},
    {"temp-directory",      REQUIRED_ARG, 0, OPT_TEMP_DIRECTORY},
    {"legacy-timestamps",   OPTIONAL_ARG, 0, OPT_LEGACY_TIMESTAMPS},
#ifdef ENABLE_BINARY
    {"binary",              NO_ARG,       0, OPT_ENABLE_BINARY},
#endif
    {0,0,0,0}               /* sentinel entry */
};

static const char *appHelp[] = {
    NULL, /* generated dynamically */
    ("Assume input has been presorted using\n"
     "\trwsort invoked with the exact same --fields value. Def. No"),
    ("Present the output in sorted order. Def. No"),
    ("Enable the next five switches--count everything.  If no\n"
     "\tcount is specified, flows are counted.  Def. No"),
    ("Sum bytes in each bin; optionally choose to print\n"
     "\tbins whose total is in given range; range is MIN or MIN-MAX. Def. No"),
    ("Sum packets in each bin; optionally choose to print\n"
     "\tbins whose total is in given range; range is MIN or MIN-MAX. Def. No"),
    ("Count flow records in each bin; optionally choose to print\n"
     "\tbins whose count is in given range; range is MIN or MIN-MAX. Def. No"),
    "Print earliest time flow was seen in each bin. Def. No",
    "Print latest time flow was seen  in each bin. Def. No",
    ("Count distinct sIPs in each bin; optionally choose to\n"
     "\tprint bins whose count is in range; range is MIN or MIN-MAX. Def. No"),
    ("Count distinct dIPs in each bin; optionally choose to\n"
     "\tprint bins whose count is in range; range is MIN or MIN-MAX. Def. No"),
    "Use given dynamic library to add fields. Def. None",
    ("When using 'sTime' or 'eTime' as a key, adjust time(s) to\n"
     "\tto appear in N-second bins (floor of time is used). Def. No, "
     DEFAULT_TIME_BIN_STR),
    "Print times in UNIX epoch seconds. Def. No",
    "Print IP numbers as integers. Def. Canonical form",
    "Print IP numbers in zero-padded canonical form. Def. No",
    "Print sensor as an integer. Def. Sensor name",
    "Do not print column titles. Def. Print titles",
    "Disable fixed-width columnar output. Def. Columnar",
    "Use specified character between columns. Def. '|'",
    "Suppress column delimiter at end of line. Def. No",
    "Shortcut for --no-columns --no-final-del --column-sep=CHAR",
    "Print names of input files as they are opened. Def. No",
    "Copy all input SiLK Flows to given pipe or file. Def. No",
    "Send output to given file path. Def. stdout",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    NULL, /* generated dynamically */
    ("Timestamp format. Choices:\n"
     "\t0==(new)\"YYYY/MM/DDThh:mm:ss\"; 1==(legacy)\"MM/DD/YYYY hh:mm:ss\".\n"
     "\tDef. 0 when switch not provided; 1 when switch provided with no value"),
#ifdef ENABLE_BINARY
    "Output in binary (hash serialization) format. Def. ASCII",
#endif
    (char *)NULL
};



/* LOCAL FUNCTION PROTOTYPES */

static void appUsageLong(void);
static int  appOptionsHandler(clientData cData, int opt_index, char *opt_arg);
static void appHandleSignal(int sig);

/* Adding key and value fields */
static int  parseKeyFields(const char *field_string);
static key_field_t *addComputedField(
    uint32_t        id,
    size_t          width,
    uniqf_t         uniqf,
    unsigned int    fn_id);
static key_field_t *addSimpleField(
    uint32_t        id,
    size_t          width,
    sk_vector_t    *outputs_vec);
static int setValueFields(void);

static int outStreamsAddAsciiStream(
    sk_vector_t        *outputs_vec,
    rwAsciiStream_t    *a_stream,
    const uint32_t     *field_list,
    uint32_t            field_count);
static int outStreamsAddPlugin(
    sk_vector_t        *outputs_vec,
    uint32_t            field_id);
static void outStreamsSetProperties(void);

/* Functions to handle plug-ins */
static int appAddDynlib(const char *dlpath, int warn_on_error);
static int  appConfigureAllDynlibs(void);
static app_dynlib_t* fieldNum2Dynlib(uint32_t fieldNum);

/* Sometimes key fields must be modified---e.g., sTime and eTime when
 * the --bin-time switch is given.  These functions have the same
 * signature as the uniq() function and do the necessary
 * modification. */
static int setKeyStime(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);
static int setKeyEtime(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);
static int setKeyElapsed(
    unsigned int    id,
    uint8_t        *bin_value,
    char           *text_value,
    size_t          text_len,
    rwRec          *rwrec);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); skOptionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
    FILE *fh = USAGE_FH;
    int i;
    int j;

#define USAGE_MSG                                                             \
    ("--fields=N [SWITCHES] [FILES]\n"                                        \
     "\tSummarize SiLK Flow records into user-defined keyed bins specified\n" \
     "\twith the --fields switch.  For each keyed bin, print byte, packet,\n" \
     "\tand/or flow counts and/or the time window when key was active.\n"     \
     "\tWhen no files are given on command line, flows are read from STDIN.\n")

    /* Load information about the dynlibs fields.  This is delayed to
     * this point to allow options processing to happen in the
     * dynlibs, which can possibly affect the fields they export.  It
     * would be better if we didn't have to do this here, but this
     * would require changing the dynlib API significantly. */
    appConfigureAllDynlibs();

    fprintf(fh, "%s %s", skAppName(), USAGE_MSG);

    fprintf(fh, "\nSWITCHES:\n");
    skOptionsDefaultUsage(fh);
    sksiteOptionsUsage(fh);
    for (i = 0; appOptions[i].name; i++) {
        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch ((appOptionsEnum)i) {
          case OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key:\n");
            rwAsciiFieldMapPrintUsage(fh, field_map);

            /* Print the IPv6 usage next */
            skIPv6PolicyUsage(fh);
            break;
          case OPT_TEMP_DIRECTORY:
            fprintf(fh,
                    ("Store temporary files in this directory.\n"
                     "\tDef. $" TMPDIR_ENVAR1 " or $" TMPDIR_ENVAR2));
#ifdef DEFAULT_TEMP_DIR
            fprintf(fh, " or %s", DEFAULT_TEMP_DIR);
#endif
            fprintf(fh, "\n");
            break;

          default:
            /* Simple help text from the appHelp array */
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibOptionsUsage(app_dynlib_list[j].dlisp, fh);
    }

    fprintf(fh, "\nNOTES:\t* The --fields switch is required.\n");
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
void appSetup(int argc, char **argv)
{
    int using_pager = 0;
    size_t i;
    int j;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

    /* check that NUM_VALUES and NUM_DISTINCTS are correct */
    assert(NUM_VALUES == _NUM_VALUES_);
    assert(NUM_DISTINCTS == _NUM_DISTINCTS_);

#ifdef DEFAULT_TEMP_DIR
    assert(DEFAULT_TEMP_DIR != NULL);
#endif

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    outputs = NULL;
    output_count = 0;
    uniq_io.outFPath = "stdout";
    uniq_io.outF = stdout;
    value_byte_width = 0;
    value_num_fields = 0;
    memset(&uniq_flags, 0, sizeof(uniq_flags));

    /* allow for 0 input and 0 output pipes */
    ioISP = iochecksSetup(1, 0, argc, argv);

    /* initialize string-map of field identifiers: add default fields,
     * then remove millisec fields, since unique-ing over them makes
     * little sense.
     *
     * Note that although we remove the MSEC fields from the available
     * fields here, the remainder of the code still supports MSEC
     * fields---which are mapped onto the non-MSEC versions of the
     * fields. */
    if (rwAsciiFieldMapAddDefaultFields(&field_map)) {
        skAppPrintErr("Unable to setup fields stringmap");
        appExit(EXIT_FAILURE);
    }
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_STIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ETIME_MSEC);
    (void)skStringMapRemoveByID(field_map, RWREC_FIELD_ELAPSED_MSEC);
    max_field_id = RWREC_PRINTABLE_FIELD_COUNT - 1;

    /* register the options */
    if (skOptionsRegister(appOptions, &appOptionsHandler, NULL)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE)
        || skIPv6PolicyOptionsRegister(&ipv6_policy))
    {
        skAppPrintErr("Unable to register options");
        appExit(EXIT_FAILURE);
    }

    /* try to load run-time dynamic libraries */
    for (j = 0; app_dynlib_names[j]; ++j) {
        appAddDynlib(app_dynlib_names[j], 0);
    }

    /* parse options */
    ioISP->firstFile = skOptionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        skAppUsage();/* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names, but we
     * should not consider it a complete failure */
    sksiteConfigure(0);

#if SK_ENABLE_IPV6
    /* we use IPSets to count distinct sIP and dIP for presorted input
     * and to count multiple distinct columns for the hash
     * case. IPSets cannot hold IPv6 addresses, so complain on these
     * conditions */
    if (distinct_num_fields && ipv6_policy >= SK_IPV6POLICY_MIX) {
        if (uniq_flags.presorted_input) {
            skAppPrintErr(("Distinct IPs are not yet supported for"
                           " presorted input and IPv6.\n"
                           "\tSpecify --ipv6-policy=asv4 to ignore IPv6"
                           " addresses in the input"));
            appExit(EXIT_FAILURE);
        }
        /* else we are using the hashlib */
        if (distinct_num_fields > 1) {
            skAppPrintErr(("Only a single distinct IP field is supported for"
                           " IPv6 and unsorted\n"
                           "\tinput.  Specify either a single distinct"
                           " columns or --ipv6-policy=asv4\n"
                           "\tto ignore IPv6 addresses"));
            appExit(EXIT_FAILURE);
        }

        /* must adjust size of distinct fields for IPv6 */
        distinct_fields[DISTINCT_SIP].df_r2k = rwrec_MemGetSIPv6;
        distinct_fields[DISTINCT_SIP].df_byte_width = RWREC_SIZEOF_SIPv6;
        distinct_fields[DISTINCT_DIP].df_r2k = rwrec_MemGetDIPv6;
        distinct_fields[DISTINCT_DIP].df_byte_width = RWREC_SIZEOF_DIPv6;
    }
#endif  /* SK_ENABLE_IPV6 */

    /* verify that the temp directory is valid */
    if (tempFileCheckTempdir()) {
        appExit(EXIT_FAILURE);
    }

    /* Set the value and/or distinct fields and parse the user's
     * minima/maxima */
    if (setValueFields()) {
        appExit(EXIT_FAILURE);
    }

    /* If epoch time requested, reduce width of VALUE_STIME and
     * VALUE_ETIME fields. */
    if (time_flags & SKTIMESTAMP_EPOCH) {
        for (i = 0; i < value_num_fields; ++i) {
            if ((value_fields[i].vf_id == VALUE_STIME)
                || (value_fields[i].vf_id == VALUE_ETIME))
            {
                value_fields[i].vf_text_len = 10;
            }
        }
    }

    /* Make sure the user specified at least one key field, parse the
     * fields */
    if (fields_arg == NULL || fields_arg[0] == '\0') {
        skAppPrintErr("The --%s switch is required",
                      appOptions[OPT_FIELDS].name);
        skAppUsage();         /* never returns */
    }
    if (parseKeyFields(fields_arg)) {
        appExit(EXIT_FAILURE);
    }

    /* Get the size of the nodes in the hash table */
    hash_node_bytes = hash_key_bytes + hash_value_bytes;

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP)
        || iochecksInputs(ioISP, 0))
    {
        skAppUsage();
    }

#ifdef ENABLE_BINARY
    /* if binary is selected, check that uniq_io.outF is not a tty */
    if (uniq_flags.binary_output) {
        if (FILEIsATty(uniq_io.outF)) {
            skAppPrintErr("Will not print binary output to tty");
            appExit(EXIT_FAILURE);
        }
    }
#endif

    /* final check. See if stdout is being used for both --copy-input
     * and as the destination for the uniq data.  */
    if (ioISP->stdoutUsed && uniq_io.outF == stdout) {
        skAppPrintErr("stdout used for both --copy-input and ascii output");
        appExit(EXIT_FAILURE);
    }

    /* Invoke the pager */
    using_pager = skOpenPagerWhenStdoutTty(&(uniq_io.outF), &pager);
    if (using_pager < 0) {
        appExit(EXIT_FAILURE);
    }
    if (using_pager) {
        uniq_io.outFPath = pager;
        uniq_io.outIsPipe = 1;
    }

    /* looks good, open the --copy-input destination */
    if (iochecksOpenCopyDest(ioISP)) {
        appExit(EXIT_FAILURE);
    }

    /* Finally, set the output handle properties */
    outStreamsSetProperties();

    /* We don't need this any more.  Destroy it */
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("Unable to register appTeardown() with atexit()");
        appExit(EXIT_FAILURE);
    }

    /* set signal handler to clean up temp files on SIGINT, SIGTERM, etc */
    if (appSetSignalHandler(&appHandleSignal)) {
        appExit(EXIT_FAILURE);
    }

    return;                       /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
void appTeardown(void)
{
    static int teardownFlag = 0;
    int j;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);

    /* Dynamic library teardown */
    for (j = 0; j < app_dynlib_count; ++j) {
        dynlibTeardown(app_dynlib_list[j].dlisp);
        app_dynlib_list[j].dlisp = NULL;
    }
    app_dynlib_count = 0;

    /* close output */
    if (uniq_io.outF != stdout) {
        if (uniq_io.outIsPipe) {
            pclose(uniq_io.outF);
        } else {
            if (EOF == fclose(uniq_io.outF) && !caught_signal) {
                skAppPrintErr("Error closing output file '%s': %s",
                              uniq_io.outFPath, strerror(errno));
            }
        }
    }

    /* destroy output */
    if (outputs) {
        size_t i;
        out_stream_t *ostream;

        for (i = 0; i < output_count; ++i) {
            ostream = outputs[i];
            if (ostream->a_stream) {
                rwAsciiStreamDestroy(&ostream->a_stream);
            }
            free(ostream);
        }

        free(outputs);
        outputs = NULL;
    }

    /* destroy field map */
    if (field_map != NULL) {
        skStringMapDestroy(field_map);
        field_map = NULL;
    }

    /* destroy key */
    if (key_fields) {
        free(key_fields);
        key_fields = NULL;
    }

    /* remove any temporary files */
    tempFileRemoveAll();

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by skOptionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    skOptionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
static int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    size_t i;
    int rv;

    switch ((appOptionsEnum)opt_index) {
      case OPT_FIELDS:
        if (fields_arg) {
            skAppPrintErr("The --%s switch was specified multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        fields_arg = opt_arg;
        break;

      case OPT_ALL_COUNTS:
        for (i = 0; i < NUM_VALUES; ++i) {
            if (value_fields[i].vf_include_in_all) {
                value_fields[i].vf_iscounted = 1;
            }
        }
        break;

      case OPT_FLOWS:
      case OPT_BYTES:
      case OPT_PACKETS:
      case OPT_STIME:
      case OPT_ETIME:
        i = opt_index - OPT_BYTES;
        assert(i == value_fields[i].vf_id);
        value_fields[i].vf_iscounted = 1;
        value_fields[i].vf_user_param = opt_arg;
        break;

      case OPT_SIP_DISTINCT:
      case OPT_DIP_DISTINCT:
        i = opt_index - OPT_SIP_DISTINCT;
        assert(i == distinct_fields[i].df_id);
        distinct_fields[i].df_iscounted = 1;
        distinct_fields[i].df_user_param = opt_arg;
        /* increment here, but note that we will recompute it as we
         * parse the user's parameters */
        ++distinct_num_fields;
        break;

      case OPT_DYNAMIC_LIB_VALUE:
        /* dynamic library */
        if (appAddDynlib(opt_arg, 1)) {
            return 1;
        }
        break;

      case OPT_BIN_TIME:
        if (opt_arg == NULL || opt_arg[0] == '\0') {
            /* no time given; use default */
            time_bin_size = DEFAULT_TIME_BIN;
        } else {
            /* parse user's time */
            rv = skStringParseUint32(&time_bin_size, opt_arg, 1, 0);
            if (rv) {
                skAppPrintErr("Invalid %s '%s': %s",
                              appOptions[opt_index].name, opt_arg,
                              skStringParseStrerror(rv));
                return 1;
            }
        }
        break;

      case OPT_PRESORTED_INPUT:
        uniq_flags.presorted_input = 1;
        break;

      case OPT_SORT_OUTPUT:
        uniq_flags.sort_output = 1;
        break;

      case OPT_EPOCH_TIME:
        if (time_flags & SKTIMESTAMP_MMDDYYYY) {
            skAppPrintErr("Both --%s and --%s specified. Using --%s",
                          appOptions[OPT_EPOCH_TIME].name,
                          appOptions[OPT_LEGACY_TIMESTAMPS].name,
                          appOptions[OPT_EPOCH_TIME].name);
        }
        time_flags = (time_flags & ~SKTIMESTAMP_MMDDYYYY) | SKTIMESTAMP_EPOCH;
        break;

      case OPT_INTEGER_IPS:
        uniq_flags.integer_ips = 1;
        break;

      case OPT_ZERO_PAD_IPS:
        uniq_flags.zero_pad_ips = 1;
        break;

      case OPT_INTEGER_SENSORS:
        uniq_flags.integer_sensors = 1;
        break;

      case OPT_NO_TITLES:
        uniq_flags.no_titles = 1;
        break;

      case OPT_NO_COLUMNS:
        uniq_flags.no_columns = 1;
        break;

      case OPT_NO_FINAL_DELIMITER:
        uniq_flags.no_final_delimiter = 1;
        break;

      case OPT_COLUMN_SEPARATOR:
        g_delimiter = opt_arg[0];
        break;

      case OPT_DELIMITED:
        uniq_flags.no_columns = 1;
        uniq_flags.no_final_delimiter = 1;
        if (opt_arg) {
            g_delimiter = opt_arg[0];
        }
        break;

      case OPT_PRINT_FILENAMES:
        uniq_flags.print_filenames = 1;
        break;

      case OPT_COPY_INPUT:
        if (iochecksAllDestinations(ioISP, opt_arg)) {
            return 1;
        }
        break;

      case OPT_OUTPUT_PATH:
        uniq_io.outFPath = opt_arg;
        if (skOpenFile(uniq_io.outFPath, 1 /* write */, &uniq_io.outF,
                       &uniq_io.outIsPipe))
        {
            return 1;
        }
        break;

      case OPT_PAGER:
        pager = opt_arg;
        break;

      case OPT_TEMP_DIRECTORY:
        assert(opt_arg);
        tempFileSetTempdir(opt_arg);
        break;

      case OPT_LEGACY_TIMESTAMPS:
        if ((opt_arg == NULL) || (opt_arg[0] == '\0') || (opt_arg[0] == '1')) {
            if (time_flags & SKTIMESTAMP_EPOCH) {
                skAppPrintErr("Both --%s and --%s specified; using --%s",
                              appOptions[OPT_EPOCH_TIME].name,
                              appOptions[OPT_LEGACY_TIMESTAMPS].name,
                              appOptions[OPT_EPOCH_TIME].name);
                /* no change */
            } else {
                time_flags |= SKTIMESTAMP_MMDDYYYY;
            }
        } else if (opt_arg[0] != '0') {
            skAppPrintErr("Invalid %s '%s': Select 0 or 1",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

#ifdef ENABLE_BINARY
      case OPT_ENABLE_BINARY:
        skAppPrintErr("Enabling binary output!");
        uniq_flags.binary_output = 1;
        break;
#endif

    }

    return 0;                     /* OK */
}


/*
 *  appExit(status)
 *
 *  Exit the application with the given status.
 */
void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 *  appHandleSignal(signal_value)
 *
 *    Call appExit() to exit the program.  If signal_value is SIGPIPE,
 *    close cleanly; otherwise print a message that we've caught the
 *    signal and exit with EXIT_FAILURE.
 */
static void appHandleSignal(int sig)
{
    caught_signal = 1;

    if (sig == SIGPIPE) {
        /* we get SIGPIPE if something downstream, like rwcut, exits
         * early, so don't bother to print a warning, and exit
         * successfully */
        appExit(EXIT_SUCCESS);
    } else {
        skAppPrintErr("Caught signal..cleaning up and exiting");
        appExit(EXIT_FAILURE);
    }
}


/*
 *  keyfield = addComputedField(map_id, width, uniq_fn, func_arg_id);
 *
 *    Add a new field to the attributes that will make up the key,
 *    that is, to the global 'key_fields' array, adjusting the
 *    'key_num_fields' and 'key_byte_width' globals as well.
 *
 *    This function differs from addSimpleField() in that here the
 *    field is computed by calling the function specified in
 *    'uniq_fn'.  The field may still be considered a "built-in"
 *    field.  For example, if the user is using --bin-time to adjust
 *    the start-time, a function is required to find the floor of the
 *    time.
 *
 *    'map_id' is the ID that this field has in the global field_map
 *    (e.g., RWREC_FIELD_ELAPSED for the duration field).
 *
 *    'width' is the number of bytes required to hold this field's
 *    value in the key.  This value should be non-zero for built-in
 *    fields on the rwRec, and 0 for fields coming from plug-ins.  If
 *    'width' is 0, the function given in 'uniq_fn' will be invoked
 *    twice: once to get the number of bytes required in the key, and
 *    again to get the width of the textual value to use for columnar
 *    output
 *
 *    The function specified in 'func_ptr' is a function used during
 *    input and output.  During input, it converts from an rwRecord to
 *    a binary value; during output, it converts from that binary
 *    value to a textual value suitable for printing.  If 'width' is
 *    zero, the 'func_ptr' function is also called during set-up.  See
 *    the description of 'width' above.
 *
 *    'func_arg_id' is a value that will be passed to 'func_ptr' to
 *    represent the 'map_id'.
 *
 *    Returns a pointer into the global 'key_fields' array that is a
 *    handle to the newly added key.
 */
static key_field_t *addComputedField(
    uint32_t        map_id,
    size_t          width,
    uniqf_t         uniqf,
    unsigned int    fn_id)
{
    key_field_t *key = &(key_fields[key_num_fields]);

    assert(width || uniqf);

    key->kf_id      = map_id;
    key->kf_offset  = key_byte_width;
    key->kf_width   = width;
    key->kf_fxn     = uniqf;
    key->kf_fxn_id  = fn_id;

    if (uniqf != NULL && width == 0) {
        uint8_t dummy_byte;
        rwRec dummy_rec;

        /* binary width */
        key->kf_width = (size_t)uniqf(fn_id, &dummy_byte, NULL, 0, NULL);

        /* printed width; if requested size is wider than the maximum
         * column width (VALUE_OUT_BUF_SIZE), only use the maximum. */
        key->kf_text_len = uniqf(fn_id, NULL, NULL, 0, &dummy_rec);
        if (key->kf_text_len == 0 || key->kf_text_len > VALUE_OUT_BUF_SIZE) {
            key->kf_text_len = VALUE_OUT_BUF_SIZE;
        }
    }

    /* compute new key width */
    key_byte_width += key->kf_width;
    ++key_num_fields;

    return key;
}


/*
 *  keyfield = addSimpleField(map_id, width, outputs_vec);
 *
 *    Add a new field to the attributes that will make up the key,
 *    that is, to the global 'key_fields' array, adjusting the
 *    'key_num_fields' and 'key_byte_width' globals as well.
 *
 *    This differs from addComputedField() in that the field is a
 *    "simple" built-in field that comes directly from the rwRec
 *    without requiring a function.
 *
 *    Because these fields are "simple", the are processed first; this
 *    allows computed fields that may require these values to be
 *    certain the values are present.
 *
 *    'map_id' is the ID that this field has in the global field_map
 *    (e.g., RWREC_FIELD_ELAPSED for the duration field).
 *
 *    'width' is the number of bytes required to hold this field's
 *    value in the key.
 *
 *    This function may re-order the keys in the 'key_fields' array.
 *    When that happens, the function pointer for plug-ins in the
 *    outputs vector will become invalid.  To keep the function
 *    pointer valid, this function will adjust entries in the
 *    'outputs_vec' if it re-orders the 'key_fields'.
 */
static key_field_t *addSimpleField(
    uint32_t        map_id,
    size_t          width,
    sk_vector_t    *outputs_vec)
{
    static size_t simple_count = 0;
    out_stream_t **ostream;
    key_field_t *key;
    size_t i;

    /* if all fields so far have been simple fields, add this field to
     * the end of the list; otherwise, shift the generated fields
     * forward by one */
    if (simple_count < key_num_fields) {
        /* shift generated fields forward by one */
        memmove(&key_fields[simple_count + 1], &key_fields[simple_count],
                ((key_num_fields - simple_count) * sizeof(key_field_t)));

        /* may need to adjust the output streams to point to the new
         * location */
        for (i = 0;
             NULL != (ostream = skVectorGetValuePointer(outputs_vec, i));
             ++i)
        {
            if ((*ostream)->a_stream) {
                continue;
            }
            ++(*ostream)->k_field;
        }
    }

    /* add the field */
    key = &(key_fields[simple_count]);
    key->kf_id      = map_id;
    key->kf_offset  = key_byte_width;
    key->kf_width   = width;
    key->kf_fxn     = NULL;
    key->kf_fxn_id  = map_id;

    /* compute new key width */
    key_byte_width += key->kf_width;

    /* increment counts */
    ++key_num_fields;
    ++simple_count;

    return key;
}


/*
 *  ok = setValueFields();
 *
 *    Loop over all the entries in the value_field[] array; for those
 *    value fields the user wants to count, parse the parameter to
 *    the switch.  The parameter should be NULL to use the default
 *    limits, a single value ("10") or a single value followed by a
 *    single hyphen ("10-") to be used as the minimum threshold, or
 *    the minimum and maximum limits separated by a hyphen ("10-20").
 *
 *    If the user didn't provide a value field, default to counting
 *    flows.
 *
 *    Figure out all the offsets into the values byte-array for each
 *    of the value fields and the overall length of the byte-array.
 *
 *    Returns 0 on success, or 1 if the opt_arg could not be parsed.
 */
static int setValueFields(void)
{
    int rv;
    const char *sp;
    size_t offset;
    uint8_t p_order;
    uint8_t target_order;
    uint8_t seen_min;
    size_t seen_idx;
    size_t i;
    int opt_index;

    /* process all the 'values' switches, parsing the user's parameter
     * (if any) and removing from the value_fields array all inactive
     * entries. */
    for (i = 0; i < NUM_VALUES; ++i) {
        if (0 == value_fields[i].vf_iscounted) {
            continue;
        }
        opt_index = i + OPT_BYTES;

        /* If an argument is provided, parse it as min-max.  If a
         * single number or a single number followed by a hyphen,
         * treat it as the minimum with no maximum. */
        sp = value_fields[i].vf_user_param;
        if (sp && sp[0]) {
            rv = skStringParseRange64(&value_fields[i].vf_min,
                                      &value_fields[i].vf_max,
                                      sp, 0, 0, SKUTILS_RANGE_SINGLE_OPEN);
            if (rv) {
                skAppPrintErr("Invalid %s '%s': %s",
                              appOptions[opt_index].name,
                              value_fields[i].vf_user_param,
                              skStringParseStrerror(rv));
                return 1;
            }
            /* Treat a single value as a having no max, not as a range
             * of a single value */
            if ((value_fields[i].vf_min == value_fields[i].vf_max)
                && !strchr(sp, '-'))
            {
                value_fields[i].vf_max = UINT64_MAX;
            }
        }

        value_byte_width += value_fields[i].vf_byte_width;

        /* move the values we actually use to the front of the
         * value_fields array */
        if (i != value_num_fields) {
            memcpy(&value_fields[value_num_fields], &value_fields[i],
                   sizeof(value_field_t));
        }
        ++value_num_fields;
    }

    /* reset distinct_num_fields */
    distinct_num_fields = 0;

    /* process all the 'distinct' switches, parsing the user's
     * parameter (if any) and removing from the distinct_fields array
     * all inactive entries. */
    for (i = 0; i < NUM_DISTINCTS; ++i) {
        if (0 == distinct_fields[i].df_iscounted) {
            continue;
        }
        opt_index = i + OPT_SIP_DISTINCT;

        /* If an argument is provided, parse it as min-max.  If a
         * single number or a single number followed by a hyphen,
         * treat it as the minimum with no maximum. */
        sp = distinct_fields[i].df_user_param;
        if (sp && sp[0]) {
            rv = skStringParseRange64(&distinct_fields[i].df_min,
                                      &distinct_fields[i].df_max,
                                      sp, 0, 0, SKUTILS_RANGE_SINGLE_OPEN);
            if (rv) {
                skAppPrintErr("Invalid %s '%s': %s",
                              appOptions[opt_index].name,
                              distinct_fields[i].df_user_param,
                              skStringParseStrerror(rv));
                return 1;
            }
            /* Treat a single value as a having no max, not as a range
             * of a single value */
            if ((distinct_fields[i].df_min == distinct_fields[i].df_max)
                && !strchr(sp, '-'))
            {
                distinct_fields[i].df_max = UINT64_MAX;
            }
        }

        distinct_fields[i].df_offset = distinct_byte_width;
        distinct_byte_width += distinct_fields[i].df_byte_width;

        /* move the values we actually use to the front of the
         * distinct_fields array */
        if (i != distinct_num_fields) {
            memcpy(&distinct_fields[distinct_num_fields], &distinct_fields[i],
                   sizeof(distinct_field_t));
        }
        ++distinct_num_fields;
    }

    /* Make certain the user specified at least one value or distinct
     * field.  If not, count by records (flows) for legacy
     * compatibility. */
    if ((value_num_fields + distinct_num_fields) == 0) {
        assert(value_fields[VALUE_FLOWS].vf_id == VALUE_FLOWS);
        value_byte_width += value_fields[VALUE_FLOWS].vf_byte_width;
        memcpy(&value_fields[value_num_fields], &value_fields[VALUE_FLOWS],
               sizeof(value_field_t));
        ++value_num_fields;
        hash_value_bytes = value_byte_width;
        return 0;
    }

    /* compute all the offsets; since the packing order is different
     * than the printing order, and values are in the value_fields[]
     * array in printing order. */
    offset = 0;
    target_order = 0;
    while (offset != value_byte_width) {
        assert(offset < value_byte_width);
        assert(target_order < NUM_VALUES);
        seen_idx = seen_min = NUM_VALUES;
        for (i = 0; i < value_num_fields; ++i) {
            p_order = value_fields[i].vf_pack_order;
            if ((p_order < seen_min) && (p_order >= target_order)) {
                seen_idx = i;
                seen_min = p_order;
                if (p_order == target_order) {
                    /* there is no need to look further */
                    break;
                }
            }
        }
        assert(seen_min != NUM_VALUES);
        value_fields[seen_idx].vf_offset = offset;
        offset += value_fields[seen_idx].vf_byte_width;
        target_order = seen_min + 1;
    }

    /* compute total length of value used by the hash table;
     * currently, the hash table requires we use *SOMETHING* as the
     * value. */
    hash_value_bytes = (value_byte_width > 0) ? value_byte_width : 1;

    return 0;
}


/*
 *  status = parseKeyFields(field_string);
 *
 *    Parse the string that represents the columns the user wishes to
 *    "uniq" over and fill in the key_fields[] and outputs[] arrays.
 *    Return 0 on success or non-zero on error.
 */
static int parseKeyFields(const char *field_string)
{
    /* the fields (columns) to print and the order to print them; each
     * value is an ID from field_map */
    uint32_t *field_list;

    /* number of fields to print; length of field_list */
    uint32_t field_count;

    /* return value; assume failure */
    int rv = 1;

    /* whether we've added the dPort key.  Both DPORT and
     * ICMP_TYPE_CODE map to the dPort. */
    unsigned int saw_dport = 0;

    /* use bits to determine which of elapsed(1), sTime(2), and
     * eTime(4) were seen */
    unsigned int saw_time = 0;

    /* which of class and type were seen */
    unsigned int saw_class_type = 0;

    /* used to build an array of out_stream_t's that will eventually
     * fill the outputs[] array. */
    sk_vector_t *outputs_vec = NULL;

    /* we can map multiple rwRec fields into a single rwAsciiStream_t,
     * but we must know the range of start and end. This gives start
     * of the range, as an index into the field_list[] array. */
    int first_position = 0;

    rwAsciiStream_t *current_stream = NULL;

    /* whether we need to reorder the key fields to have etime occur
     * last */
    int move_etime = 0;

    size_t j;
    uint32_t i;

    /* Load information about the dynlibs fields.  This is delayed to
     * this point to allow options processing to happen in the
     * dynlibs, which can possibly affect the fields they export.  It
     * would be better if we didn't have to do this here, but this
     * would require changing the dynlib API significantly. */
    appConfigureAllDynlibs();

    /* parse the field-list into an array of fields */
    if (rwAsciiFieldMapParseFields(&field_list, &field_count, field_string,
                                   field_map, RW_ASCII_DUPES_ERROR))
    {
        /* library printed the error */
        return 1;
    }

    /* set the number of key-fields and the key-byte-width to zero.
     * create the array to hold the information about each
     * key-field. */
    key_byte_width = 0;
    key_num_fields = 0;
    key_fields = calloc(field_count, sizeof(key_field_t));
    if (key_fields == NULL) {
        goto END;
    }

    /* create vector to hold output streams */
    outputs = NULL;
    outputs_vec = skVectorNew(sizeof(out_stream_t*));
    if (outputs_vec == NULL) {
        goto END;
    }

    /* Go over the list of fields, looking for anything that we have
     * to handle specially; for example, how we hande eTime depends on
     * whether sTime and/or duration are also specified. */
    for (i = 0; i < field_count; ++i) {
        switch (field_list[i]) {
          case RWREC_FIELD_SIP:
            /* Counting distinct IPs when the IP is part of the
             * key doesn't make sense. */
            for (j = 0; j < distinct_num_fields; ++j) {
                if (distinct_fields[j].df_id == DISTINCT_SIP) {
                    skAppPrintErr(("Error: Cannot count distinct sIPs"
                                   " when sIP is part of key."));
                    goto END;
                }
            }
            break;

          case RWREC_FIELD_DIP:
            /* Counting distinct IPs when the IP is part of the
             * key doesn't make sense. */
            for (j = 0; j < distinct_num_fields; ++j) {
                if (distinct_fields[j].df_id == DISTINCT_DIP) {
                    skAppPrintErr(("Error: Cannot count distinct dIPs"
                                   " when dIP is part of key."));
                    goto END;
                }
            }
            break;

          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            if (saw_time & 4) {
                move_etime = 1;
            }
            saw_time |= 2;

          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            if (saw_time & 4) {
                move_etime = 1;
            }
            saw_time |= 1;
            break;

          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            saw_time |= 4;
            break;

          default:
            break;
        }
    }

    /* Warn if --bin-time requested but time not part of key.  Warn
     * that duration field will be modified when all time fields are
     * given. */
    if (time_bin_size != 0) {
        if (saw_time < 2) {
            if (FILEIsATty(stderr)) {
                skAppPrintErr("No time fields in key; %s switch ignored",
                              appOptions[OPT_BIN_TIME].name);
            }
        } else if (saw_time == 7) {
            /* must adjust elapsed to be eTime-sTime */
            adjust_elapsed = 1;
            if (FILEIsATty(stderr)) {
                skAppPrintErr("Warning: modifying duration field "
                              "to be eTime-sTime");
            }
        }
    }

    /* Go over the fields again, creating output streams and filling
     * in the key_fields[] array. */
    for (i = 0; i < field_count; ++i) {
        if (field_list[i] >= RWREC_PRINTABLE_FIELD_COUNT) {
            /* field comes from a plug-in */
            /* first, close the current stream */
            if (NULL != current_stream) {
                rv = outStreamsAddAsciiStream(outputs_vec, current_stream,
                                              &field_list[first_position],
                                              i - first_position);
                if (rv != 0) {
                    goto END;
                }
                current_stream = NULL;
            }

            /* create out_stream_t for plug-in, initialize plug-in,
             * and add it as a key field */
            if (outStreamsAddPlugin(outputs_vec, field_list[i])) {
                goto END;
            }
        } else {
            /* field is built-in */
            if (NULL == current_stream) {
                /* create new ascii-stream */
                if (0 != rwAsciiStreamCreate(&current_stream)) {
                    skAppPrintErr("Cannot create stream");
                    goto END;
                }
                first_position = i;
            }

            /* add this field to the key_fields[]. */
            switch ((rwrec_printable_fields_t)field_list[i]) {
              case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    addSimpleField(field_list[i], RWREC_SIZEOF_SIPv6,
                                   outputs_vec);
                    break;
                }
#endif  /* SK_ENABLE_IPV6 */
                addSimpleField(field_list[i], RWREC_SIZEOF_SIPv4, outputs_vec);
                break;

              case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    addSimpleField(field_list[i], RWREC_SIZEOF_DIPv6,
                                   outputs_vec);
                    break;
                }
#endif /* SK_ENABLE_IPV6 */
                addSimpleField(field_list[i], RWREC_SIZEOF_DIPv4, outputs_vec);
                break;

              case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
                if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                    addSimpleField(field_list[i], RWREC_SIZEOF_NHIPv6,
                                   outputs_vec);
                    break;
                }
#endif /* SK_ENABLE_IPV6 */
                addSimpleField(field_list[i], RWREC_SIZEOF_NHIPv4,
                               outputs_vec);
                break;

              case RWREC_FIELD_SPORT:
                addSimpleField(field_list[i], RWREC_SIZEOF_SPORT, outputs_vec);
                break;

              case RWREC_FIELD_DPORT:
              case RWREC_FIELD_ICMP_TYPE_CODE:
                if (0 == saw_dport) {
                    saw_dport = 1;
                    addSimpleField(field_list[i], RWREC_SIZEOF_DPORT,
                                   outputs_vec);
                }
                break;

              case RWREC_FIELD_PROTO:
                addSimpleField(field_list[i], RWREC_SIZEOF_PROTO, outputs_vec);
                break;

              case RWREC_FIELD_PKTS:
                addSimpleField(field_list[i], RWREC_SIZEOF_PKTS, outputs_vec);
                break;

              case RWREC_FIELD_BYTES:
                addSimpleField(field_list[i], RWREC_SIZEOF_BYTES, outputs_vec);
                break;

              case RWREC_FIELD_FLAGS:
                addSimpleField(field_list[i], RWREC_SIZEOF_FLAGS, outputs_vec);
                break;

              case RWREC_FIELD_STIME:
              case RWREC_FIELD_STIME_MSEC:
                /* use size of 4 since we convert the sktime_t to
                 * seconds since the UNIX epoch */
                addComputedField(field_list[i], 4, &setKeyStime, 0);
                break;

              case RWREC_FIELD_ELAPSED:
              case RWREC_FIELD_ELAPSED_MSEC:
                addComputedField(field_list[i], RWREC_SIZEOF_ELAPSED,
                                 &setKeyElapsed, 0);

                break;

              case RWREC_FIELD_ETIME:
              case RWREC_FIELD_ETIME_MSEC:
                addComputedField(RWREC_FIELD_ETIME, 4, &setKeyEtime, 0);
                break;

              case RWREC_FIELD_SID:
                addSimpleField(field_list[i], RWREC_SIZEOF_SID, outputs_vec);
                break;

              case RWREC_FIELD_INPUT:
                addSimpleField(field_list[i], RWREC_SIZEOF_INPUT, outputs_vec);
                break;

              case RWREC_FIELD_OUTPUT:
                addSimpleField(field_list[i], RWREC_SIZEOF_OUTPUT,
                               outputs_vec);
                break;

              case RWREC_FIELD_INIT_FLAGS:
                addSimpleField(field_list[i], RWREC_SIZEOF_INIT_FLAGS,
                               outputs_vec);
                break;

              case RWREC_FIELD_REST_FLAGS:
                addSimpleField(field_list[i], RWREC_SIZEOF_REST_FLAGS,
                               outputs_vec);
                break;

              case RWREC_FIELD_TCP_STATE:
                addSimpleField(field_list[i], RWREC_SIZEOF_TCP_STATE,
                               outputs_vec);
                break;

              case RWREC_FIELD_APPLICATION:
                addSimpleField(field_list[i], RWREC_SIZEOF_APPLICATION,
                               outputs_vec);
                break;

              case RWREC_FIELD_FTYPE_CLASS:
              case RWREC_FIELD_FTYPE_TYPE:
                if (0 == saw_class_type) {
                    saw_class_type = 1;
                    addSimpleField(field_list[i], RWREC_SIZEOF_FLOW_TYPE,
                                   outputs_vec);
                }
                break;
            }
        }
    }

    /* all done; close the final_stream */
    if (current_stream != NULL) {
        rv = outStreamsAddAsciiStream(outputs_vec, current_stream,
                                      &field_list[first_position],
                                      i - first_position);
        if (rv != 0) {
            goto END;
        }
        current_stream = NULL;
    }

    /* convert the output_vec to an array */
    output_count = skVectorGetCount(outputs_vec);
    outputs = calloc(output_count, sizeof(out_stream_t*));
    if (outputs == NULL) {
        goto END;
    }
    skVectorToArray(outputs, outputs_vec);
    skVectorDestroy(outputs_vec);
    outputs_vec = NULL;

    /* there is one more bit of madness to handle.  The eTime field
     * must occur after the sTime and elapsed fields when restoring
     * the fields. */
    if (move_etime) {
        size_t etime_idx = UINT32_MAX;
        size_t other_idx = UINT32_MAX;
        key_field_t k;
        for (i = 0; i < key_num_fields; ++i) {
            switch (key_fields[i].kf_id) {
              case RWREC_FIELD_STIME:
              case RWREC_FIELD_STIME_MSEC:
              case RWREC_FIELD_ELAPSED:
              case RWREC_FIELD_ELAPSED_MSEC:
                other_idx = i;
                break;

              case RWREC_FIELD_ETIME:
              case RWREC_FIELD_ETIME_MSEC:
                etime_idx = i;
                break;
            }
        }
        if (other_idx > key_num_fields || etime_idx > key_num_fields) {
            abort();
        }
        /* swap the etime field with the final non-etime-Time field  */
        memcpy(&k, &key_fields[other_idx], sizeof(key_field_t));
        memcpy(&key_fields[other_idx], &key_fields[etime_idx],
               sizeof(key_field_t));
        memcpy(&key_fields[etime_idx], &k, sizeof(key_field_t));
    }

    /* adjust the offset of any distinct fields by the key width */
    for (i = 0; i < distinct_num_fields; ++i) {
        distinct_fields[i].df_offset += key_byte_width;
    }

    /* size of the key used in the hash table */
    hash_key_bytes = key_byte_width + distinct_byte_width;

    /* check size of key */
    if (key_byte_width > MAX_KEY_BYTE_WIDTH) {
        skAppPrintErr("Fields make too large a key: %u bytes > %u max",
                      (unsigned int)key_byte_width, MAX_KEY_BYTE_WIDTH);
        goto END;
    }

    /* successful */
    rv = 0;

  END:
    if (rv != 0) {
        /* something went wrong.  clean up */
        if (key_fields) {
            free(key_fields);
            key_fields = NULL;
        }
        if (outputs_vec != NULL) {
            /* clean up all the out_stream_t's on error */
            output_count = skVectorGetCount(outputs_vec);
            for (i = 0; i < output_count; ++i) {
                out_stream_t *o_stream;
                skVectorGetValue(&o_stream, outputs_vec, i);
                if (NULL != o_stream->a_stream) {
                    rwAsciiStreamDestroy(&(o_stream->a_stream));
                }
                free(o_stream);
            }
        }
        if (outputs != NULL) {
            for (i = 0; i < output_count; ++i) {
                out_stream_t *o_stream = outputs[i];
                if (NULL != o_stream->a_stream) {
                    rwAsciiStreamDestroy(&(o_stream->a_stream));
                }
                free(o_stream);
            }
            free(outputs);
            outputs = NULL;
        }
    }
    /* do standard clean-up */
    if (outputs_vec != NULL) {
        skVectorDestroy(outputs_vec);
    }
    if (field_list != NULL) {
        free(field_list);
    }

    return rv;
}




/*
 *  status = outStreamsAddAsciiStream(outputs_vec, a_stream, list, count);
 *
 *    Create a new out_stream_t and add it to the vector of output
 *    streams, 'outputs_vec'.  The vector must already exist.
 *
 *    The new out_stream_t is a wrapper around the ascii-stream
 *    'a_stream', which must already exist; 'list' is an array of
 *    integers that contains the fields of the 'a_stream' to be
 *    printed; 'count' is the number of elements in 'list'.
 *
 *    Returns 0 on success, or -1 for a memory allocation error.
 */
static int outStreamsAddAsciiStream(
    sk_vector_t        *outputs_vec,
    rwAsciiStream_t    *a_stream,
    const uint32_t     *field_list,
    uint32_t            field_count)
{
    out_stream_t  *ostream;

    ostream = calloc(1, sizeof(out_stream_t));
    if (ostream == NULL) {
        return -1;
    }

    rwAsciiSetFields(a_stream, field_list, field_count);
    ostream->a_stream = a_stream;

    if (0 != skVectorAppendValue(outputs_vec, &ostream)) {
        free(ostream);
        return -1;
    }

    return 0;
}


/*
 *  status = outStreamsAddPlugin(outputs_vec, field_id);
 *
 *    Create a new out_stream_t and add it to the vector of output
 *    streams, 'outputs_vec'.  The vector must already exist.
 *
 *    The new out_stream_t is a wrapper around the 'cut_fn' provided
 *    by the plug-in (run-time dynamic library) that provides the
 *    field numbered 'field_id'.
 *
 *    Returns 0 on success, or -1 for a memory allocation error or if
 *    the plug-in that provides the field numbered 'field_id' cannot
 *    be found.
 */
static int outStreamsAddPlugin(
    sk_vector_t        *outputs_vec,
    uint32_t            field_id)
{
    out_stream_t *ostream;
    app_dynlib_t *app_dl;
    key_field_t *keyfield;
    unsigned int id;

    ostream = calloc(1, sizeof(out_stream_t));
    if (ostream == NULL) {
        return -1;
    }

    if (0 != skVectorAppendValue(outputs_vec, &ostream)) {
        free(ostream);
        return -1;
    }

    /* map the field_id to the dynamic library and initialize the
     * plug-in; find the id that the plug-in uses for this field. */
    app_dl = fieldNum2Dynlib(field_id);
    if (NULL == app_dl) {
        /* could not initialize the plug-in */
        return -1;
    }
    id = field_id - app_dl->offset;

    /* add the key-field */
    keyfield = addComputedField(field_id, 0, app_dl->fxn, id);
    ostream->k_field = keyfield;

    return 0;
}


/*
 *  outStreamsSetProperties();
 *
 *    Use the switches entered by the user to set properties on all
 *    the ascii-streams that appear in the global 'outputs' array.
 */
static void outStreamsSetProperties(void)
{
    uint32_t i;
    rwAsciiStream_t *a_stream;
    int ip_warning_printed = 0;

    for (i = 0; i < output_count; ++i) {
        if (NULL == outputs[i]->a_stream) {
            continue;
        }

        /* only ascii streams make it here */
        a_stream = outputs[i]->a_stream;

        /* properties for every stream */
        rwAsciiSetNoNewline(a_stream);
        rwAsciiSetDelimiter(a_stream, g_delimiter);
        rwAsciiSetOutputHandle(a_stream, uniq_io.outF);
        rwAsciiSetIPv6Policy(a_stream, ipv6_policy);

        /* simple properties */
        rwAsciiSetTimestampFlags(a_stream, time_flags);
        if (uniq_flags.no_titles) {
            rwAsciiSetNoTitles(a_stream);
        }
        if (uniq_flags.no_columns) {
            rwAsciiSetNoColumns(a_stream);
        }
        if (uniq_flags.integer_sensors) {
            rwAsciiSetIntegerSensors(a_stream);
        }

        /* conflicting properties */
        if (uniq_flags.zero_pad_ips && !uniq_flags.integer_ips) {
            rwAsciiSetZeroPadIps(a_stream);
        }
        if (uniq_flags.integer_ips) {
            if ((uniq_flags.zero_pad_ips) && (ip_warning_printed == 0)) {
                ip_warning_printed = 1;
                skAppPrintErr("--integer-ips option overrides"
                              " --zero-pad-ips\n"
                              "\tWill print IPs as integers.");
            }
            rwAsciiSetIntegerIps(a_stream);
        }
#if 0
        if ((uniq_flags.icmpTandC) && (icmp_t_and_c == 0)) {
            icmp_t_and_c = 1;
            /* would prefer to do this in the ascii-stream where the
             * user has used sport/dport columns */
            rwAsciiSetIcmpTypeCode(a_stream);
        }
#endif
    }
}


/*
 *  ok = setKeyStime(id, bin_value, text_value, text_len, rwrec);
 *
 *    When 'text_value' is null (input), copy the start-time value (in
 *    seconds) from 'rwrec' into 'bin_value'.  If the user requested
 *    the floor of the start-time, do that calculation.  The
 *    start-time is always stored in network byte order.
 *
 *    When 'text_value' is non-null (output), copy the value from
 *    'bin_value' to the start-time field in the 'rwrec'.
 *
 *    This function always returns 0.
 *
 *    The 'text_value' parameter is only used to determine whether the
 *    input or output is requested.  The 'id' and 'text_len'
 *    parameters are ignored.
 */
static int setKeyStime(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char           *text_value,
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t sTime;

    if (NULL == text_value) {
        if (time_bin_size <= 1) {
            sTime = htonl(rwRecGetStartSeconds(rwrec));
        } else {
            sTime = rwRecGetStartSeconds(rwrec);
            /* find the floor of the time */
            sTime = htonl(sTime - (sTime % time_bin_size));
        }
        memcpy(bin_value, &sTime, sizeof(sTime));
    } else {
        memcpy(&sTime, bin_value, sizeof(sTime));
        rwRecSetStartTime(rwrec, (sktime_t)1000 * ntohl(sTime));
    }

    return 0;
}


/*
 *  ok = setKeyElapsed(id, bin_value, text_value, text_len, rwrec);
 *
 *    When 'text_value' is NULL (input), get the elapsed field from
 *    the 'rwrec' and adjust it so that the computed start-time and
 *    end-time fields fall into the proper bins with respect to the
 *    global 'time_bin_size'.  Put that computed elapsed value into
 *    'bin_value' in network byte order.
 *
 *    When 'text_value' is non-NULL (output), copy the value from
 *    'bin_value' to the elapsed field in the 'rwrec'.
 *
 *    This function always returns 0.
 *
 *    The 'text_value' parameter is only used to determine whether the
 *    input or output is requested.  The 'id' and 'text_len'
 *    parameters are ignored.
 */
static int setKeyElapsed(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char           *text_value,
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t elapsed;
    uint32_t sTime;

    if (NULL == text_value) {
        if (!adjust_elapsed) {
            elapsed = htonl(rwRecGetElapsedSeconds(rwrec));
        } else {
            /*
             * set elapsed to:
             * (eTime - (eTime % bin_size)) - (sTime - (sTime % bin_size))
             */
            sTime = rwRecGetStartSeconds(rwrec);
            elapsed = rwRecGetElapsedSeconds(rwrec);
            elapsed = htonl(elapsed + (sTime % time_bin_size)
                            - ((sTime + elapsed) % time_bin_size));
        }
        memcpy(bin_value, &elapsed, sizeof(elapsed));
    } else {
        memcpy(&elapsed, bin_value, sizeof(elapsed));
        rwRecSetElapsed(rwrec, 1000 * ntohl(elapsed));
    }

    return 0;
}


/*
 *  len = setKeyEtime(id, bin_value, text_value, text_len, rwrec);
 *
 *    When 'text_value' is null (input), use the start-time and
 *    elapsed fields on the 'rwrec' to compute the end-time and put
 *    the value into 'bin_value'.  Adjust the end-time relative to the
 *    global 'time_bin_size' if required.
 *
 *    When 'text_value' is non-null (output), adjust the start-time
 *    and elapsed fields on the 'rwrec' so that the end-time will have
 *    the value present in 'bin_value'.  This fucntion assumes the
 *    start-time and elapsed values have already been copied from the
 *    byte-array.
 *
 *    This function always returns 0.
 *
 *    The 'text_value' parameter is only used to determine whether the
 *    input or output is requested.  The 'id' and 'text_len'
 *    parameters are ignored.
 */
static int setKeyEtime(
    unsigned int    UNUSED(id),
    uint8_t        *bin_value,
    char           *text_value,
    size_t          UNUSED(text_len),
    rwRec          *rwrec)
{
    uint32_t eTime, sTime, elapsed;

    if (NULL == text_value) {
        /* compute eTime and put it into bin_value */
        if (time_bin_size == 0) {
            eTime = htonl(rwRecGetEndSeconds(rwrec));
        } else {
            /* compute floor */
            eTime = rwRecGetEndSeconds(rwrec);
            eTime = htonl(eTime - (eTime % time_bin_size));
        }
        memcpy(bin_value, &eTime, sizeof(eTime));
        return 0;
    }

    /* given an eTime value, find the sTime and elapsed values to set
     * on the rwrec that will add to give eTime.  This assumes that
     * the rwrec's sTime and elapsed have already been set. */
    memcpy(&eTime, bin_value, sizeof(eTime));
    eTime = ntohl(eTime);

    sTime = rwRecGetStartSeconds(rwrec);
    elapsed = rwRecGetElapsedSeconds(rwrec);
    if (sTime) {
        if (elapsed == 0) {
            /* set elapsed based on start and end times */
            assert(sTime <= eTime);
            rwRecSetElapsed(rwrec, (1000 * (eTime - sTime)));
        }
#ifndef NDEBUG
        else {
            /* nothing to do to the records; verify that the times all
             * make sense. */
            assert(eTime == (sTime + elapsed));
        }
#endif
    } else if (elapsed) {
        /* set start time based on end time and elapsed */
        assert(eTime > elapsed);
        rwRecSetStartTime(rwrec, (sktime_t)1000 * (eTime - elapsed));
    } else {
        /* Just set sTime to eTime and elapsed is already 0 */
        rwRecSetStartTime(rwrec, (sktime_t)1000 * eTime);
    }

    return 0;
}


/*
 *  appConfigureDynlib(dlisp, warn_on_error);
 *
 *    Configure the dynamic library dlisp.  Return 1 if there are any
 *    problems during configuration or loading field information, 0
 *    otherwise.
 */
static int appConfigureDynlib(
    dynlibInfoStruct_t *dlisp,
    int                 warn_on_error)
{
    sk_stringmap_entry_t *entries = NULL;
    int i;
    uint32_t offset;
    int rv = 1; /* return value; assume error */

    /* the IDs for the fields from the plug-in are added above this
     * value.  All fields from the plug-in must have an ID > 0. */
    offset = max_field_id;

    /* get the fields for the application */
    if (dynlibConfigure(dlisp, (void*)(&entries)) != 0) {
        if (warn_on_error) {
            skAppPrintErr("Error getting fields from plug-in '%s'",
                          dynlibGetPath(dlisp));
        }
        goto END;
    }

    if (entries == NULL) {
        /* We didn't get any fields from dynlibConfigure().  That's
         * the only way this application supports fields, so ignore
         * this dynlib. */
        if (warn_on_error) {
            skAppPrintErr(("Did not get fields from configure()"
                           " function in plug-in '%s'"),
                          dynlibGetPath(dlisp));
        }
        goto END;
    }

    /* the dynlibConfigure() routine gave us the fields */
    for (i = 0; entries[i].name != NULL; ++i) {
        /* check that value is not too small or too big */
        if (entries[i].id == 0) {
            skAppPrintErr("Invalid id '0' for field '%s' in plug-in '%s'",
                          entries[i].name, dynlibGetPath(dlisp));
            goto END;
        }
        if (offset >= (UINT32_MAX - entries[i].id)) {
            skAppPrintErr("Invalid id '%u' for field '%s' in plug-in '%s'",
                          entries[i].id, entries[i].name,
                          dynlibGetPath(dlisp));
            goto END;
        }

        /* add it */
        if (SKSTRINGMAP_OK != skStringMapAddID(field_map, entries[i].name,
                                               offset + entries[i].id))
        {
            skAppPrintErr("Cannot add entries to map");
            goto END;
        }
        /* keep track of the maximum field id */
        if (max_field_id < (offset + entries[i].id)) {
            max_field_id = offset + entries[i].id;
        }
    }

    rv = 0;

  END:
    if (rv != 0) {
        sk_link_err_t link_rv;
        sk_link_item_t *node;
        sk_stringmap_entry_t *entry;

        /* roll back the field_map to those whose IDs are <= offset */
        link_rv = skLinkGetHead(&node, field_map);
        while (link_rv == SKLINK_OK) {
            skLinkGetData((void**)&entry, node);
            link_rv = skLinkGetNext(&node, node);
            if (entry->id > offset) {
                /* remove the entry */
                skStringMapRemoveByName(field_map, entry->name);
            }
        }
        max_field_id = offset;
    }

    return rv;
}


/*
 *  appConfigureAllDynlibs();
 *
 *    Configure all known dynlibs.  Return the number of dynlibs that
 *    failed configuration.  This function is idempotent, and will
 *    return 0 on subsequent calls.
 */
static int appConfigureAllDynlibs(void)
{
    static int app_dynlibs_configured = 0;
    int rv;
    int i;
    int failed = 0;

    if (app_dynlibs_configured) {
        return 0;
    }
    app_dynlibs_configured = 1;

    for (i = 0; i < app_dynlib_count; i++) {
        /* Store the field offset for this plug-in */
        app_dynlib_list[i].offset = max_field_id;

        rv = appConfigureDynlib(app_dynlib_list[i].dlisp,
                                app_dynlib_list[i].via_switch);
        if (rv != 0) {
            /* Failed */
            ++failed;
        }
    }

    return failed;
}


/*
 *  appAddDynlib(dlpath, via_switch);
 *
 *    Load the dynamic library at dlpath, store its information in the
 *    app_dynlib_list, and increment the app_dynlib_count.  Does not
 *    get the fields from the dynlib, that happens in
 *    appConfigureDynlib().  Return 0 on success.  Return 1 if the
 *    library cannot be loaded or if too many dynamic libraries have
 *    been loaded.
 */
static int appAddDynlib(const char *dlpath, int via_switch)
{
    dynlibInfoStruct_t *dlisp;
    uniqf_t uniqf;

    /* too many plug-ins? */
    if (app_dynlib_count == APP_MAX_DYNLIBS) {
        skAppPrintErr("Too many dynlibs.  Only %u allowed", APP_MAX_DYNLIBS);
        return 1;
    }

    /* load it */
    dlisp = dynlibCreate(DYNLIB_UNIQ);
    if (dynlibLoad(dlisp, dlpath)) {
        /* failed to find library, missing symbols, or setup failed */
        if (via_switch) {
            skAppPrintErr("Error loading dynamic library: '%s': %s",
                          dlpath, dynlibLastError(dlisp));
        }
        return 1;
    }

    uniqf = (uniqf_t)dynlibGetRWProcessor(dlisp);
    /* dynlibCheck() should have made this check already */
    assert(uniqf != NULL);

    /* add the dynamic library */
    app_dynlib_list[app_dynlib_count].fxn = uniqf;
    app_dynlib_list[app_dynlib_count].dlisp = dlisp;
    app_dynlib_list[app_dynlib_count].via_switch = (via_switch ? 1 : 0);
    ++app_dynlib_count;

    return 0;
}


/*
 *  fieldNum2Dynlib(field_id);
 *
 *    Given a field number 'field_id' from the global 'field_map',
 *    return a pointer to a app_dynlib_t from the app_dynlib_list[]
 *    that points to the dynamic library used to print the field_id
 *    field.  Initializes the dynamic library.
 *
 *    Return NULL if the field_id is not related to any dynamic
 *    library or if the initialization of dynamic library failed.
 */
static app_dynlib_t *fieldNum2Dynlib(uint32_t field_id)
{
    int j;
    app_dynlib_t *app_dl;

    if (!app_dynlib_count || (field_id <= app_dynlib_list[0].offset)) {
        return NULL;
    }

    for (j = app_dynlib_count - 1; j != 0; --j) {
        if (field_id > app_dynlib_list[j].offset) {
            break;
        }
    }
    app_dl = &(app_dynlib_list[j]);

    /* since we will be using the dynamic library, initialize it */
    if (dynlibInitialize(app_dl->dlisp)) {
        /* did not find the field; go through a lot of trouble to
         * print an error message, then return */
        sk_vector_t *entries_vec;
        sk_stringmap_entry_t *entry = NULL;

        entries_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
        if (NULL == entries_vec) {
            skAppPrintErr(("Field not available--cannot initialize %s"),
                          dynlibGetPath(app_dl->dlisp));
            return NULL;
        }
        skStringMapGetNames(entries_vec, field_map, field_id);
        skVectorGetValue(&entry, entries_vec, 0);
        if (NULL != entry) {
            skAppPrintErr(("Field '%s' not available--cannot initialize %s"),
                          entry->name, dynlibGetPath(app_dl->dlisp));
        }
        skVectorClear(entries_vec);
        return NULL;
    }

    return app_dl;
}


/*
 *  int = appNextInput(&rwios);
 *
 *    Fill 'rwios' with the next input file to read.  Return 0 if
 *    'rwios' was successfully opened or 1 if there are no more input
 *    files.
 *
 *    When an input file cannot be opened, the return value is
 *    dependent on the error.  If the error is due to being out of
 *    file handles or memory (EMFILE or ENOMEM), return -2; otherwise
 *    return -1.
 */
int appNextInput(skstream_t **rwios)
{
    static int counter = -1;
    int rv;

    if (counter < 0) {
        counter = ioISP->firstFile;
    } else {
        ++counter;
        if (counter == ioISP->fileCount) {
            /* no more input */
            return 1;
        }
    }

    /* create rwios and open file */
    errno = 0;
    if ((rv = skStreamCreate(rwios, SK_IO_READ, SK_CONTENT_SILK_FLOW))
        || (rv = skStreamBind(*rwios, ioISP->fnArray[counter]))
        || (rv = skStreamOpen(*rwios))
        || (rv = skStreamReadSilkHeader(*rwios, NULL)))
    {
        if (errno == EMFILE || errno == ENOMEM) {
            /* decrement counter so we try to open this file again */
            --counter;
            rv = -2;
            PRINTDEBUG((stderr, "unable to open %s because of %s\n",
                        ioISP->fnArray[counter], strerror(errno)));
        } else {
            if (uniq_flags.print_filenames) {
                fprintf(PRINT_FILENAMES_FH, "%s\n", ioISP->fnArray[counter]);
            }
            skStreamPrintLastErr(*rwios, rv, &skAppPrintErr);
            rv = -1;
        }
        skStreamDestroy(rwios);
        return rv;
    }

    if (uniq_flags.print_filenames) {
        fprintf(PRINT_FILENAMES_FH, "%s\n", skStreamGetPathname(*rwios));
    }
    skStreamSetCopyInput(*rwios, ioISP->inputCopyFD);
    skStreamSetIPv6Policy(*rwios, ipv6_policy);

    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
