/*
** Copyright (C) 2001-2008 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwuniqtemp.c
**
**    Functions to handle temp file creation and access.
**
**  TO DO:
**
**    -- Should better handle the case where we reach the
**    MAX_TMP_FILES limit.
*/

#include "silk.h"

RCSIDENT("$SiLK: rwuniqtemp.c 13074 2008-11-20 20:28:45Z mthomas $");

#include "rwuniqtemp.h"
#include "utils.h"


/* LOCAL VARIABLES */

/* temporary directory */
static const char *user_temp_dir = NULL;

/* template used to make temporary files. */
static char temp_file_template[PATH_MAX];

/* names of temporary files */
static char *tmp_names[MAX_TMP_FILES];


/* FUNCTION DEFINITIONS */


/*
 *  ok = appSetSignalHandler(sig_fn)
 *
 *    Call sig_fn to clean up any temporary files whenever we get a
 *    signal.
 */
int appSetSignalHandler(sig_fn_t sig_fn)
{
    /* list of handled signals */
    int sigs[] = {SIGINT, SIGPIPE, SIGQUIT, SIGTERM};
    struct sigaction act;
    size_t i;

    act.sa_handler = sig_fn;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
#ifdef SA_INTERRUPT
    act.sa_flags |= SA_INTERRUPT;
#endif

    for (i = 0; i < (sizeof(sigs)/sizeof(int)); ++i) {
        if (sigaction(sigs[i], &act, NULL) < 0) {
            skAppPrintErr("Cannot register handler for signal #%d",
                          sigs[i]);
            return -1;
        }
    }
    return 0;
}


/*
 *  tempFileSetTempdir(dir);
 *
 *    Set the directory to use for temporary files to 'dir'.
 */
void tempFileSetTempdir(const char *temp_dir)
{
    user_temp_dir = temp_dir;
}


/*
 *  ok = tempFileCheckTempdir();
 *
 *    Fill in the C string 'template' of size 'len' with the temp-file
 *    template.  Attempts to find a temporary directory, either from
 *    the user's input, from the TMPDIR environment variable, or from
 *    the DEFAULT_TEMP_DIR macro.  Also uses the application name.
 *    'template' is suitable for passing to mkstemp().  Returns 0 on
 *    success, or -1 if no directory could be found.
 */
int tempFileCheckTempdir(void)
{
    const char *tmp_dir = NULL;
    int rv;

    /* initialize tmp_names */
    memset(tmp_names, 0, sizeof(tmp_names));

    /* First, figure out the directory to use.  Use the user's option
     * if given, or SILK_TMPDIR, TMPDIR, or the default */
    if (!tmp_dir) {
        tmp_dir = user_temp_dir;
    }
    if (!tmp_dir) {
        tmp_dir = getenv(TMPDIR_ENVAR1);
    }
    if (!tmp_dir) {
        tmp_dir = getenv(TMPDIR_ENVAR2);
    }
#ifdef DEFAULT_TEMP_DIR
    if (!tmp_dir) {
        tmp_dir = DEFAULT_TEMP_DIR;
    }
#endif /* DEFAULT_TEMP_DIR */
    if (!tmp_dir) {
        skAppPrintErr("Cannot find a value for the temporary directory.");
        return -1;
    }
    if ( !skDirExists(tmp_dir)) {
        skAppPrintErr("Temporary directory '%s' does not exist", tmp_dir);
        return -1;
    }

    rv = snprintf(temp_file_template, sizeof(temp_file_template),
                  "%s/%s_tmp.XXXXXXXX", tmp_dir, skAppName());
    if ((size_t)rv >= sizeof(temp_file_template) || rv < 0) {
        skAppPrintErr("snprintf error");
        return -1;
    }

    return 0;
}


const char *tempFileGetName(int tmp_idx)
{
    return (tmp_names[tmp_idx] ? tmp_names[tmp_idx] : "NULL");
}


/*
 *  filep = tempFileCreate(tmp_idx);
 *
 *    Create and open a new temporary file indexed by 'tmp_idx.'  The
 *    name of the file can be found by tempFileGetName(), and the file
 *    can be re-opened with the tempFileOpen() function.  Returns a
 *    file pointer to the new file, or NULL on error: 'tmp_idx' out of
 *    range, mkstemp() fails.
 *
 */
FILE *tempFileCreate(int tmp_idx)
{
    FILE *fp = NULL;
    int fd;

    /* check that tmp_idx is in range */
    if (tmp_idx < 0 || tmp_idx >= MAX_TMP_FILES) {
        skAppPrintErr("tempFileCreate index %d out of range 0 < n < %d",
                      tmp_idx, MAX_TMP_FILES);
        return NULL;
    }

    /* copy template name */
    tmp_names[tmp_idx] = strdup(temp_file_template);
    if (NULL == tmp_names[tmp_idx]) {
        skAppPrintErr("Out of memory!");
        return NULL;
    }

    /* open file */
    fd = mkstemp(tmp_names[tmp_idx]);
    if (fd == -1) {
        skAppPrintErr("Cannot create temp file %s: %s",
                      tmp_names[tmp_idx], strerror(errno));
        free(tmp_names[tmp_idx]);
        tmp_names[tmp_idx] = NULL;
        return NULL;
    }

    /* convert descriptor to pointer */
    fp = fdopen(fd, "w");
    if (fp == NULL) {
        skAppPrintErr("Cannot convert descriptor to pointer for %s: %s",
                      tmp_names[tmp_idx], strerror(errno));
        close(fd);
        return NULL;
    }

    PRINTDEBUG((stderr, "new tmp %d => '%s'\n", tmp_idx, tmp_names[tmp_idx]));

    return fp;
}


/*
 *  fp = tempFileOpen(tmp_idx);
 *
 *    Returns a FILE* pointing to a unique temporary filename
 *    associated with 'tmp_idx' that has been previously created with
 *    tempFileCreate().  Returns NULL if file cannot be opened.
 */
FILE *tempFileOpen(int tmp_idx)
{
    const char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);
    PRINTDEBUG((stderr, "existing tmp %d => '%s'\n", tmp_idx, fn));

    return fopen(fn, "r");
}


/*
 *  tempFileRemove(tmp_idx);
 *
 *    Remove the temporary file indexed by 'tmp_idx'.
 */
void tempFileRemove(int tmp_idx)
{
    const char *fn;

    assert (0 <= tmp_idx && tmp_idx < MAX_TMP_FILES);
    assert (tmp_names[tmp_idx]);

    fn = tempFileGetName(tmp_idx);

    PRINTDEBUG((stderr, "Removing file %s of size %lu\n",
                fn, (unsigned long)skFileSize(fn)));
    if (-1 == remove(fn)) {
        if (skFileExists(fn)) {
            skAppPrintErr("Cannot remove file '%s': %s",
                          fn, strerror(errno));
        }
    }

    tmp_names[tmp_idx][0] = '\0';
    free(tmp_names[tmp_idx]);
    tmp_names[tmp_idx] = NULL;
}


/*
 *  tempFileRemoveAll();
 *
 *    Remove all temporary files.
 */
void tempFileRemoveAll(void)
{
    int i;

    for (i = 0; i < MAX_TMP_FILES; ++i) {
        if (tmp_names[i]) {
            tempFileRemove(i);
        }
    }

}


/*
 *  ok = tempFileDumpRecords(tmp_idx, buffer, size, count);
 *
 *    Print 'count' records having size 'size' bytes from 'buffer'
 *    into a newly created temporary file.  Index the temp file by
 *    'tmp_idx'.  Returns 0 on success, or -1 if a temp file cannot be
 *    created, or the write or close calls fail.
 */
int tempFileDumpRecords(
    int             tmp_idx,
    const void     *rec_buffer,
    uint32_t        rec_size,
    uint32_t        rec_count)
{
    FILE* temp_filep = NULL;
    int rv = -1; /* return value */

    temp_filep = tempFileCreate(tmp_idx);
    if (temp_filep == NULL) {
        goto END;
    }

    PRINTDEBUG((stderr, "Writing %lu records to '%s'...",
                (unsigned long)rec_count, tempFileGetName(tmp_idx)));
    if (rec_count != fwrite(rec_buffer, rec_size, rec_count, temp_filep)) {
        /* error writing, errno may or may not be set */
        PRINTDEBUG((stderr, "ERROR!\n"));
        skAppPrintErr("Error writing records to temp file %s: %s",
                      tempFileGetName(tmp_idx), strerror(errno));
        goto END;
    }

    /* Success so far */
    PRINTDEBUG((stderr, "done\n"));
    rv = 0;

  END:
    /* close the file if open */
    if (temp_filep) {
        if (fclose(temp_filep) == EOF) {
            /* error closing file; only report error if status so far
             * is good. */
            if (rv == 0) {
                skAppPrintErr("Error closing temp file %s: %s",
                              tempFileGetName(tmp_idx), strerror(errno));
                rv = -1;
            }
        }
    }

    return rv;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
